Package: tourr
Title: Tour Methods for Multivariate Data Visualisation
Version: 1.2.7
Authors@R: c(
    person("Hadley", "Wickham", email = "h.wickham@gmail.com", role = c("aut", "ctb"), comment = c(ORCID = "0000-0003-4757-117X")),
    person("Dianne", "Cook", email = "dicook@monash.edu", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-3813-7155")),
    person("Nick", "Spyrison", email = "nicholas.spyrison@monash.edu", role = "ctb", comment = c(ORCID = "0000-0002-8417-0212")),
    person("Ursula", "Laa", email = "ursula.laa@boku.ac.at", role = "ctb", comment = c(ORCID = "0000-0002-0249-6439")), 
    person("H. Sherry","Zhang", email = "huizezhangsh@gmail.com", role = "ctb", comment = c(ORCID = "0000-0002-7122-1463")),
    person("Stuart", "Lee", email = "stuart.lee1@monash.edu", role = "ctb", comment = c(ORCID = "0000-0003-1179-8436")))
Description: Implements geodesic interpolation and basis
    generation functions that allow you to create new tour
    methods from R.
Depends: R (>= 4.1.0)
Imports: MASS, tibble, dplyr, stats, utils, grDevices, ash, energy,
        mgcv, geozoo, cassowaryr, minerva
Suggests: aplpack, testthat (>= 3.0.0), colorspace, ggplot2, gifski,
        knitr, rmarkdown, tidyr, covr, plotly
License: MIT + file LICENSE
LazyData: true
URL: https://ggobi.github.io/tourr/
BugReports: https://github.com/ggobi/tourr/issues
Encoding: UTF-8
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-15 07:42:31 UTC; cookd
Author: Hadley Wickham [aut, ctb] (ORCID:
    <https://orcid.org/0000-0003-4757-117X>),
  Dianne Cook [aut, cre] (ORCID: <https://orcid.org/0000-0002-3813-7155>),
  Nick Spyrison [ctb] (ORCID: <https://orcid.org/0000-0002-8417-0212>),
  Ursula Laa [ctb] (ORCID: <https://orcid.org/0000-0002-0249-6439>),
  H. Sherry Zhang [ctb] (ORCID: <https://orcid.org/0000-0002-7122-1463>),
  Stuart Lee [ctb] (ORCID: <https://orcid.org/0000-0003-1179-8436>)
Maintainer: Dianne Cook <dicook@monash.edu>
Repository: CRAN
Date/Publication: 2026-05-15 08:10:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("[",history_array)
S3method("[[",history_array)
S3method(as.array,history_array)
S3method(as.array,history_list)
S3method(as.list,history_array)
S3method(as.list,history_list)
S3method(length,history_array)
S3method(plot,path_curve)
S3method(plot,path_index)
S3method(print,history_array)
S3method(print,tour_path)
S3method(str,history_array)
export(MIC)
export(TIC)
export(anchored_orthogonal_distance)
export(andrews)
export(angular_breaks)
export(animate)
export(animate_andrews)
export(animate_density2d)
export(animate_depth)
export(animate_dist)
export(animate_faces)
export(animate_groupxy)
export(animate_idx)
export(animate_image)
export(animate_pca)
export(animate_pcp)
export(animate_sage)
export(animate_scatmat)
export(animate_slice)
export(animate_stars)
export(animate_stereo)
export(animate_trails)
export(animate_xy)
export(anomaly_index)
export(areColors)
export(basis_init)
export(basis_random)
export(center)
export(check_dup)
export(cmass)
export(cumulative_radial)
export(dcor2d)
export(dcor2d_2)
export(dependence_tour)
export(display_andrews)
export(display_density2d)
export(display_depth)
export(display_dist)
export(display_faces)
export(display_groupxy)
export(display_idx)
export(display_image)
export(display_pca)
export(display_pcp)
export(display_sage)
export(display_scatmat)
export(display_slice)
export(display_stars)
export(display_stereo)
export(display_trails)
export(display_xy)
export(draw_tour_axes)
export(estimate_eps)
export(find_platform)
export(freeze)
export(frozen_guided_tour)
export(frozen_tour)
export(geodesic_path)
export(grand_tour)
export(guided_anomaly_tour)
export(guided_section_tour)
export(guided_tour)
export(holes)
export(interpolate)
export(is_orthonormal)
export(lda_pp)
export(linear_breaks)
export(little_tour)
export(local_tour)
export(loess2d)
export(mahal_dist)
export(manual_slice)
export(mapColors)
export(mapShapes)
export(new_geodesic_path)
export(new_tour)
export(norm_bin)
export(norm_kol)
export(normalise)
export(orthonormalise)
export(orthonormalise_by)
export(path_curves)
export(path_dist)
export(path_index)
export(paths_index)
export(pda_pp)
export(planned2_tour)
export(planned_tour)
export(proj_dist)
export(radial_tour)
export(render)
export(render_anim)
export(render_gif)
export(render_proj)
export(rescale)
export(save_history)
export(search_better)
export(search_better_random)
export(search_geodesic)
export(search_jellyfish)
export(search_polish)
export(search_posse)
export(skewness)
export(slice_index)
export(sphere_data)
export(splines2d)
export(stringy)
export(thaw)
importFrom(grDevices,dev.cur)
importFrom(grDevices,dev.flush)
importFrom(grDevices,dev.hold)
importFrom(grDevices,dev.off)
importFrom(grDevices,hcl)
importFrom(grDevices,rgb)
importFrom(graphics,abline)
importFrom(graphics,arrows)
importFrom(graphics,axis)
importFrom(graphics,box)
importFrom(graphics,contour)
importFrom(graphics,hist)
importFrom(graphics,image)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,pairs)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(graphics,rect)
importFrom(graphics,segments)
importFrom(graphics,stars)
importFrom(graphics,text)
importFrom(stats,loess)
importFrom(stats,mahalanobis)
importFrom(stats,na.omit)
importFrom(stats,qchisq)
importFrom(stats,quantile)
importFrom(stats,residuals)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,var)
importFrom(utils,globalVariables)
importFrom(utils,tail)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  eval = FALSE,
  comment = "#>",
  rmarkdown.html_vignette.check_title = FALSE
)

## ----setup--------------------------------------------------------------------
# library(tourr)

## ----cube3--------------------------------------------------------------------
# # generate 3D cube vertices
# cube <- geozoo::cube.iterate(3)
# # data is stored points, edges contains the needed two column matrix for connecting points
# cube$points
# cube$edges
# # call grand tour with the scatterplot (xy) display, turning off axes
# animate_xy(cube$points, edges = cube$edges, axes = "off")

## ----cube5--------------------------------------------------------------------
# cube5 <- geozoo::cube.iterate(5)
# animate_xy(cube5$points, edges = cube5$edges, axes = "off")

## -----------------------------------------------------------------------------
# # get centroids of 3 clusters of the flea data
# n <- nrow(flea)
# set.seed(1019)
# flea_centroids <- stats::kmeans(flea[,1:6], 3)$centers
# flea_aug <- rbind(flea[,1:6], flea_centroids)
# col <- c(rep("black", n), rep("orange", 3))
# flea_edges <- matrix(c(n+1, n+2, n+1, n+3, n+2, n+3), ncol=2, byrow = TRUE)
# animate_xy(flea_aug, edges = flea_edges,
#            col = col, edges.col = "orange",
#            edges.width = 3)

#' Display tour path with ggobi
#'
#' Use ggobi to display a 2D tour.
#'
#' @param ...  other arguments passed on to \code{\link{animate}}
#' @export
#' @examples
#' \dontrun{
#' animate_ggobi(flea[, 1:6])
#' # Spelling things out explicitly
#' animate(flea[, 1:6], grand_tour(), display_ggobi())
#' }
display_ggobi <- function() {
  if(!require("rggobi", quietly = TRUE)) {
    stop("rggobi required for ggobi based tour")
  }
  
  gd <- NULL  
  init <- function(data) {
    message("Pause the tour in GGobi to allow R control to begin")
    g <- ggobi(data)
    gd <<- displays(g)[[1]]
    pmode(gd) <- "2D Tour"
  }
  
  render_data <- function(data, proj, geodesic) {
    ggobi_display_set_tour_projection(gd, proj)
  }
  
  list(
    init = init,
    render_frame = nul,
    render_transition = nul,
    render_data = render_data,
    render_target = nul
  )
}

#' @rdname display_ggobi
#' @inheritParams animate
#' @export
animate_ggobi <- function(data, tour_path = grand_tour(), ...) {
  animate(data, tour_path, display_ggobi(...), ...)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = TRUE
)
library(tourr)

## ----basic, eval = FALSE------------------------------------------------------
# animate_xy(flea[,1:6],
#   obs_labels = as.character(1:nrow(flea)),
#   axes = "off"
# )

## ----subset, eval = FALSE-----------------------------------------------------
# # Find the five points furthest from the multivariate centre
# d2    <- mahalanobis(flea[,1:6], colMeans(flea[,1:6]), cov(flea[,1:6]))
# top5  <- order(d2, decreasing = TRUE)[1:5]
# 
# # Build a label vector: empty for most points, row number for the top 5
# lbls <- rep("", nrow(flea))
# lbls[top5] <- as.character(top5)
# 
# animate_xy(flea[,1:6], obs_labels = lbls, axes="bottomleft")

## ----colour, eval = FALSE-----------------------------------------------------
# # Colour by species; label only row 1 and row 22 (arbitrary examples)
# lbls <- rep("", nrow(flea))
# lbls[c(1, 22)] <- rownames(flea)[c(1, 22)]
# 
# animate_xy(flea[,1:6],
#   col       = flea$species,
#   obs_labels = lbls,
#   axes      = "bottomleft"
# )

## ----gif, eval = FALSE--------------------------------------------------------
# lbls <- rep("", nrow(f))
# lbls[c(1, 22)] <- rownames(flea)[c(1, 22)]
# 
# render_gif(
#   flea[,1:6],
#   tour_path = grand_tour(),
#   display   = display_xy(
#     col        = flea$species,
#     obs_labels = labs,
#     axes       = "bottomleft"
#   ),
#   gif_file = "labelled_tour.gif",
#   frames   = 60,
#   width    = 400,
#   height   = 400
# )

## ----cex, eval = FALSE--------------------------------------------------------
# animate_xy(flea[,1:6],
#   obs_labels = as.character(1:nrow(f)),
#   axes = "off",
#   cex  = 0.6
# )

# f1 <- save_history(flea[, 1:6],grand_tour(d=1), max_bases = 10)
# x <- path1d_ggobi(f1)
# ggobi(x) 
path1d_ggobi <- function(basis, interp=TRUE, support=TRUE) {
  nc <- dim(basis)[1]

  if (interp) 
    basis_interp <- interpolate(basis)
  else
    basis_interp <- basis

  x <- matrix(f1interp[,1,],ncol=nc,byrow=T) 
  colnames(x) <- paste("a", 1:nc, sep="")

  if (support) {
    y <- NULL
    for (i in 1:1000) 
      y <- rbind(y, t(basis_random(nc, 1)))
    colnames(y) <- paste("a", 1:nc, sep="")
    x <- rbind(x, y)
    x <- cbind(x, c(rep(1, nrow(x)-1000), rep(0, 1000)))
    colnames(x)[nc+1] <- "PathorSupport"
  }
  x
}

hcl <- expand.grid(h = seq(0, 360, by = 10), l = 80, c = seq(0, 100, by = 10))
hcl <- transform(hcl, 
  angle = h * pi / 180,
  radius = c / diff(range(c))
)
hcl <- transform(hcl,
  x = radius * sin(angle),
  y = radius * cos(angle)
)
hcl$colour <- hcl(hcl$h, hcl$c, hcl$l, fixup=FALSE)

par(pty="s", mar=c(0,0,0,0), bg="grey85")
# with(hcl, plot(x, y, col=colour, pch=20, cex=3 , axes=FALSE, xlab="", ylab=""))


rgb_vals <- expand.grid(b = seq(0, 0.2, length=10), g = 0, r = seq(0.8, 1, length = 10))
rgb_vals$colour <- do.call(rgb, rgb_vals)

with(rgb_vals, plot(r, b, col=colour, pch=20, cex=8, axes=FALSE, xlab="", ylab=""))
with(rgb_vals, text(r, b, cex = 0.5, label = paste(format(r, TRUE, 2), format(g, TRUE, 2), format(b, TRUE, 2))))


# Blue: rgb(0, 0.91, 0.89)
# Red: rgb(0.98, 0.052, 0)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = TRUE
)
library(tourr)

## ----grand, eval = FALSE------------------------------------------------------
# # 2D projections — the default
# f <- flea[,1:6] # For convenience
# animate_xy(f)
# 
# # 1D projections shown as a density
# animate_dist(f)

## ----guided, eval = FALSE-----------------------------------------------------
# # Find cluster structure without class labels
# animate_xy(f, tour_path = guided_tour(holes()))
# 
# # Same search, colour by known species once structure is found
# animate_xy(f,
#   tour_path = guided_tour(holes()),
#   col = flea$species
# )
# 
# # Use class labels directly to separate groups
# animate_xy(f,
#   tour_path = guided_tour(lda_pp(flea$species)),
#   col = flea$species
# )

## ----little, eval = FALSE-----------------------------------------------------
# animate_xy(f, tour_path = little_tour())

## ----planned, eval = FALSE----------------------------------------------------
# # Save a grand tour path
# set.seed(42)
# t1 <- save_history(f, max = 10)
# 
# # Replay it in a scatterplot
# animate_xy(f, tour_path = planned_tour(t1))
# 
# # Replay the same path with species colour added
# animate_xy(f,
#   tour_path = planned_tour(t1),
#   col = flea$species
# )
# 
# # Cycle continuously through the saved bases
# animate_xy(f, tour_path = planned_tour(t1, cycle = TRUE))

## ----local, eval = FALSE------------------------------------------------------
# # Start from a specific projection (e.g. one saved from a guided tour)
# start <- basis_random(6, 2)
# 
# # Explore a small neighbourhood of angle pi/4
# animate_xy(f,
#   tour_path = local_tour(start, angle = pi / 4),
#   col = flea$species
# )

## ----radial, eval = FALSE-----------------------------------------------------
# # Use a saved projection as the starting point — here we take the
# # end-point of a guided tour run with holes()
# set.seed(42)
# guided_path  <- save_history(f, guided_tour(holes()), sphere = TRUE)
# start        <- matrix(guided_path[, , dim(guided_path)[3]], nrow = 6)
# 
# # Rotate variable 4 (elytra width) out and back in
# animate_xy(f,
#   tour_path = radial_tour(start, mvar = 4),
#   col = flea$species,
#   rescale = TRUE
# )
# 
# # Rotate two variables simultaneously to see their joint contribution
# animate_xy(f,
#   tour_path = radial_tour(start, mvar = c(3, 4)),
#   col = flea$species,
#   rescale = TRUE
# )

## ----radial-1d, eval = FALSE--------------------------------------------------
# start1d <- basis_random(6, 1)
# animate_dist(f, radial_tour(start1d, mvar = 2), rescale = TRUE)

## ----frozen, eval = FALSE-----------------------------------------------------
# # Fix variable 3 to contribute equally to both axes
# frozen      <- matrix(NA, nrow = 6, ncol = 2)
# frozen[3, ] <- 0.5
# 
# animate_xy(f, tour_path = frozen_tour(2, frozen))

## ----gif, eval = FALSE--------------------------------------------------------
# render_gif(
#   f,
#   tour_path = guided_tour(holes()),
#   display   = display_xy(col = flea$species),
#   gif_file  = "guided_flea.gif",
#   frames    = 60
# )

#' Image tour path animation.
#'
#' Animate a 1d tour path with an image plot.  This animation requires a
#' different input data structure, a 3d array.  The first two dimensions are
#' locations on a grid, and the 3rd dimension gives the observations to be
#' mixed with the tour.
#'
#' @param xs x limit that is used in making the size of the plot
#' @param ys y limit that is used in making the size of the plot
#' @param ... other arguments passed on to \code{\link{animate}}
#' @seealso \code{\link{animate}} for options that apply to all animations
#' @keywords hplot
#' @export
#' @examples
#' str(ozone)
#' animate_image(ozone)
display_image <- function(xs, ys, ...) {
  render_frame <- function() {
    blank_plot(xlim = c(1, xs), ylim = c(1, xs))
  }

  render_data <- function(data, proj, geodesic) {
    z <- data %*% proj
    dim(z) <- c(xs, ys)
    image(
      x = seq_len(xs), y = seq_len(ys),
      z = t(z),
      zlim = c(-2, 2), add = TRUE
    )
  }

  list(
    init = nul,
    render_frame = render_frame,
    render_transition = nul,
    render_data = render_data,
    render_target = nul
  )
}


#' @rdname display_image
#' @inheritParams animate
#' @export
animate_image <- function(data, tour_path = grand_tour(1), ...) {
  xs <- dim(data)[1]
  ys <- dim(data)[2]
  zs <- dim(data)[3]
  dim(data) <- c(xs * ys, zs)

  animate(
    data = data, tour_path = tour_path,
    display = display_image(xs, ys, ...), ...
  )
}

#' Display tour path with a sliced scatterplot
#'
#' Animate a 2D tour path with a sliced scatterplot.
#'
#' @param axes position of the axes: center, bottomleft or off
#' @param center if TRUE, centers projected data to (0,0).  This pins the
#'  center of data cloud and make it easier to focus on the changing shape
#'  rather than position.
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param edges A two column integer matrix giving indices of ends of lines.
#' @param edges.col colour of edges to be plotted, Defaults to "black.
#' @param col color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".
#' @param pch_slice marker for plotting points inside the slice.
#'   Defaults to 20.
#' @param pch_other marker for plotting points outside the slice.
#'   Defaults to 46.
#' @param cex_slice size of the points inside the slice. Defaults to 2.
#' @param cex_other size if the points outside the slice. Defaults to 1.
#' @param v_rel relative volume of the slice. If not set, suggested value
#'   is calculated and printed to the screen.
#' @param anchor A vector specifying the reference point to anchor the slice.
#'   If NULL (default) the slice will be anchored at the data center.
#' @param anchor_nav position of the anchor: center, topright or off
#' @param rescale Default FALSE. If TRUE, rescale all variables to range [0,1].
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @param axislablong text labels only for the long axes in a projection, default FALSE
#' @param ...  other arguments passed on to \code{\link{animate}} and
#'   \code{\link{display_slice}}
#' @export
#' @examples
#' # Generate samples on a 3d and 5d hollow sphere using the geozoo package
#' sphere3 <- geozoo::sphere.hollow(3)$points
#' sphere5 <- geozoo::sphere.hollow(5)$points
#'
#' # Columns need to be named before launching the tour
#' colnames(sphere3) <- c("x1", "x2", "x3")
#' colnames(sphere5) <- c("x1", "x2", "x3", "x4", "x5")
#'
#' # Animate with the slice display using the default parameters
#' animate_slice(sphere3)
#' animate_slice(sphere5)
#'
#' # Animate with off-center anchoring
#' anchor3 <- matrix(rep(0.7, 3), ncol=3)
#' anchor5 <- matrix(rep(0.3, 5), ncol=5)
#' animate_slice(sphere3, anchor = anchor3)
#' # Animate with thicker slice to capture more points in each view
#' animate_slice(sphere5, anchor = anchor5, v_rel = 0.02)
display_slice <- function(center = TRUE, axes = "center", half_range = NULL,
                          col = "black", pch_slice = 20, pch_other = 46,
                          cex_slice = 2, cex_other = 1, v_rel = NULL,
                          anchor = NULL, anchor_nav = "off",
                          edges = NULL, edges.col = "black",
                          palette = "Zissou 1",
                          axislablong = FALSE, ...) {
  labels <- NULL
  h <- NULL

  # If colors are a variable, convert to colors
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    col <- mapColors(col, palette)
  }
  if (is.factor(edges.col) | !areColors(edges.col)) {
    edges.gps <- edges.col
    edges.col <- mapColors(edges.col, palette)
  }

  init <- function(data) {
    half_range <<- compute_half_range(half_range, data, center)
    labels <<- abbreviate(colnames(data), 3)
    v_rel <<- compute_v_rel(v_rel, half_range, ncol(data))
    # Translate volume v_rel to cutoff h
    h <<- v_rel^(1 / (ncol(data) - 2))
    message("Using v_rel=", format(v_rel, digits = 2), ", corresponding to a cutoff h=", format(h, digits = 2))
  }

  if (!is.null(edges)) {
    if (!is.matrix(edges) && ncol(edges) == 2) {
      stop("Edges matrix needs two columns, from and to, only.")
    }
  }

  render_frame <- function() {
    par(pty = "s", mar = rep(0.1, 4))
    blank_plot(xlim = c(-1, 1), ylim = c(-1, 1))
  }
  render_transition <- function() {
    rect(-1, -1, 1, 1, col = "#FFFFFFE6", border = NA)
  }

  render_data <- function(data, proj, geodesic, with_anchor = anchor) {
    draw_tour_axes(proj, labels, limits = 1, axes, longlabels=axislablong)
    if (!is.null(with_anchor)) {
      rng <- apply(data, 2, range)
      colnames(with_anchor) <- colnames(data)
      draw_slice_center(with_anchor, rng, limits = 1, anchor_nav = anchor_nav)
    }

    # Render projected points
    x <- data %*% proj
    d <- anchored_orthogonal_distance(proj, data, with_anchor)
    pch <- rep(pch_other, nrow(x))
    pch[d < h] <- pch_slice
    cex <- rep(cex_other, nrow(x))
    cex[d < h] <- cex_slice
    if (center) x <- center(x)
    x <- x / half_range
    points(x, col = col, pch = pch, cex = cex)

    if (!is.null(edges)) {
      segments(x[edges[, 1], 1], x[edges[, 1], 2],
               x[edges[, 2], 1], x[edges[, 2], 2],
               col = edges.col
      )
    }
  }


  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}

#' @rdname display_slice
#' @inheritParams animate
#' @export
animate_slice <- function(data, tour_path = grand_tour(), rescale = FALSE, ...) {
  animate(data, tour_path, display_slice(...), rescale = rescale)
}

#' Draw slice center guide with base graphics
#' @keywords internal
draw_slice_center <- function(anchor, rng, limits, anchor_nav, ...) {
  anchor_nav <- match.arg(anchor_nav, c("center", "topright", "off"))
  if (anchor_nav == "off") {
    return()
  }

  if (anchor_nav == "center") {
    axis_scale <- 2 * limits / 3
    axis_pos <- 0
  } else if (anchor_nav == "topright") {
    axis_scale <- limits / 6
    axis_pos <- 2 / 3 * limits
  }

  adj <- function(x) axis_pos + x * axis_scale

  n <- ncol(anchor)
  theta <- seq(90, 450, length = n + 1) * pi/180
  theta <- theta[1:n]

  xx <- cos(theta)
  yy <- sin(theta)
  cgap <- 1 # shift of min out from middle
  seg <- 4
  cglty <- 3
  cglwd <- 1
  cglcol <- "black"
  for (i in 0:seg) {
    polygon(adj(xx * (i + cgap)/(seg + cgap)),
            adj(yy * (i + cgap)/(seg + cgap)),
            lty = cglty, lwd = cglwd, border = cglcol)
    arrows(adj(xx/(seg + cgap)),
           adj(yy/(seg + cgap)),
           adj(xx * 1),
           adj(yy * 1),
           lwd = cglwd, lty = cglty,
           length = 0, col = cglcol)
  }
  VLABELS <- colnames(anchor)
  text(adj(xx * 1.2), adj(yy * 1.2), VLABELS, cex=0.8)
  xxs <- xx
  yys <- yy
  scale <- cgap/(seg + cgap) +
    (anchor[1, ] - rng[1, ])/
    (rng[2,] - rng[1, ]) * seg/(seg + cgap)
  for (j in 1:n) {
    xxs[j] <- adj(xx[j] * cgap/(seg + cgap) + xx[j] *
      (anchor[1, j] - rng[1, j])/(rng[2, j] - rng[1, j]) *
      seg/(seg + cgap))
    yys[j] <- adj(yy[j] * cgap/(seg + cgap) + yy[j] *
      (anchor[1, j] - rng[1, j])/(rng[2, j] - rng[1, j]) *
      seg/(seg + cgap))
  }
  polygon(xxs, yys, lty = 1, lwd = 2,
          border = "black",
          col = NULL)
  points(adj(xx * scale), adj(yy * scale), pch = 16,
         col = "black")

}

#' A planned tour path.
#'
#' The planned tour takes you from one basis to the next in a
#' set order.  Once you have visited all the planned bases, you either stop
#' or start from the beginning once more (if \code{cycle = TRUE}).
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param basis_set the set of bases as a list of projection matrices
#'   or a 3d array
#' @param cycle cycle through continuously (\code{TRUE}) or stop after
#'   first pass (\code{FALSE})
#' @keywords hplot dynamic
#' @seealso The \code{\link{little_tour}}, a special type of planned tour
#'   which cycles between all axis parallel projections.
#' @rdname planned-tour
#' @export
#' @examples
#' twod <- save_history(flea[, 1:3], max = 5)
#' str(twod)
#' animate_xy(flea[, 1:3], planned_tour(twod))
#' animate_xy(flea[, 1:3], planned_tour(twod, TRUE))
#'
#' oned <- save_history(flea[, 1:6], grand_tour(1), max = 3)
#' animate_dist(flea[, 1:6], planned_tour(oned))
planned_tour <- function(basis_set, cycle = FALSE) {
  index <- 1
  basis_set <- as.list(basis_set)

  n <- length(basis_set)
  if (cycle) {
    generator <- function(current, data, ...) {
      if (is.null(current)) {
        return(basis_set[[1]])
      }

      index <<- (index %% n) + 1
      target <- basis_set[[index]]
      list(target = target)
    }
  } else {
    generator <- function(current, data, ...) {
      if (is.null(current)) {
        return(basis_set[[1]])
      }
      index <<- index + 1
      if (index > n) {
        return(NULL)
      }
      target <- basis_set[[index]]
      list(target = target)
    }
  }

  new_geodesic_path("planned", generator)
}

#' @rdname planned-tour
#' @export
planned2_tour <- function(basis_set) {
  index <- 1
  generator <- function(current, data, ...) {
    if (is.null(current)) {return(as.matrix(basis_set[,,1]))}

    index <<- index + 1
    if (index > dim(basis_set)[3]) {
      return(NULL)
    }
    target <- as.matrix(basis_set[,,index])
    list(target = target)

  }

  new_geodesic_path("planned2", generator)
}

#' A guided anomaly tour path.
#'
#' The guided anomaly tour is a variation of the guided tour that is
#' using an ellipse to determine anomalies on which to select target planes.
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate_xy}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param index_f the section pursuit index function to optimise. The function
#'   needs to take two arguments, the projected data, indexes of anomalies.
#' @param d target dimensionality
#' @param alpha the initial size of the search window, in radians
#' @param cooling the amount the size of the search window should be adjusted
#'   by after each step
#' @param search_f the search strategy to use
#' @param max.tries the maximum number of unsuccessful attempts to find
#'   a better projection before giving up
#' @param max.i the maximum index value, stop search if a larger value is found
#' @param ellipse pxp variance-covariance matrix defining ellipse, default NULL.
#'        Useful for comparing data with some hypothesized null.
#' @param ellc This can be considered the equivalent of a critical value, used to
#'        scale the ellipse larger or smaller to capture more or fewer anomalies. Default 3.
#' @param ellmu This is the centre of the ellipse corresponding to the mean of the
#'        normal population. Default vector of 0's
#' @param ... arguments sent to the search_f
#' @seealso \code{\link{slice_index}} for an example of an index functions.
#' \code{\link{search_geodesic}}, \code{\link{search_better}},
#'   \code{\link{search_better_random}} for different search strategies
#' @importFrom stats mahalanobis qchisq
#' @export
#' @examples
#' animate_xy(flea[, 1:6], guided_anomaly_tour(anomaly_index(),
#'   ellipse=cov(flea[,1:6])), ellipse=cov(flea[,1:6]), axes="off")
guided_anomaly_tour <- function(index_f, d = 2, alpha = 0.5, cooling = 0.99,
                                max.tries = 25, max.i = Inf,
                                ellipse, ellc=NULL, ellmu=NULL,
                                search_f = search_geodesic, ...) {
  h <- NULL

  generator <- function(current, data, tries, ...) {
    if (is.null(current)) {
      return(basis_random(ncol(data), d))
    }

    if (is.null(h)) {
      half_range <- compute_half_range(NULL, data, FALSE)
    }

    index <- function(proj) {
      if (nrow(ellipse) == nrow(proj)) {

        if (is.null(ellc))
          ellc <<- qchisq(0.95, nrow(proj))
        else
          stopifnot(ellc > 0) # Needs to be positive
        if (is.null(ellmu))
          ellmu <<- rep(0, nrow(proj))
        else
          stopifnot(length(ellmu) == nrow(proj)) # Right dimension
        #message("Using ellc = ", format(ellc, digits = 2))

        # Check which observations are outside pD ellipse
        mdst <- mahalanobis(data,
                            center=ellmu,
                            cov=ellipse)
      #mdst <- mahal_dist(data, ellipse)
        anomalies <- which(mdst > ellc)
        stopifnot(length(anomalies) > 0)
        #cat(length(anomalies), "\n")

        # Project ellipse into 2D
        evc <- eigen(ellipse) #
        ellinv <- (evc$vectors) %*% as.matrix(diag(evc$values)) %*% t(evc$vectors)
        e2 <- t(proj) %*% ellipse %*% proj
        evc2 <- eigen(e2)
        ell2d <- as.matrix(evc2$vectors) %*% diag(sqrt(evc2$values*ellc)) %*% t(as.matrix(evc2$vectors))

        ell2dinv <- (evc2$vectors) %*% diag(evc2$values*ellc) %*% t(evc2$vectors)
        ellmu2d <- t(as.matrix(ellmu)) %*% proj
        #evc <- eigen(ellipse)
        #ellinv <- (evc$vectors) %*% diag(evc$values) %*% t(evc$vectors)
        #e2 <- t(proj) %*% ellinv %*% proj
        #evc2 <- eigen(e2)
        #ell2d <- (evc2$vectors) %*% diag(sqrt(evc2$values)) %*% t(evc2$vectors)
        #e3 <- eigen(ell2d)
        #ell2dinv <- (e3$vectors) %*% diag(e3$values) %*% t(e3$vectors)
        index_f(as.matrix(data[anomalies,]) %*% proj, e2, ellmu2d)
      }
    }

    cur_index <- index(current)

    if (cur_index > max.i) {
      cat("Found index ", cur_index, ", larger than selected maximum ", max.i, ". Stopping search.\n",
        sep = ""
      )
      cat("Final projection: \n")
      if (ncol(current) == 1) {
        for (i in 1:length(current)) {
          cat(sprintf("%.3f", current[i]), " ")
        }
        cat("\n")
      }
      else {
        for (i in 1:nrow(current)) {
          for (j in 1:ncol(current)) {
            cat(sprintf("%.3f", current[i, j]), " ")
          }
          cat("\n")
        }
      }
      return(NULL)
    }

    basis <- search_f(current, alpha, index, tries, max.tries, cur_index = cur_index, ...)
    alpha <<- alpha * cooling

    list(target = basis$target, index = index)
  }

  new_geodesic_path("guided", generator)
}

#' Render frames of animation to disk
#'
#' @param data matrix, or data frame containing numeric columns
#' @param tour_path tour path generator
#' @param display the method used to render the projected data,
#'   e.g. \code{\link{display_xy}}, \code{\link{display_pcp}}
#' @param dev name of output device to use (e.g. \code{\link{png}},
#'   \code{\link{pdf}})
#' @param ... other options passed to output device
#' @param apf angle (in radians) per frame
#' @param frames number of frames in output
#' @param rescale default FALSE. If TRUE, rescale all variables to range [0,1]
#' @param sphere if true, sphere all variables
#' @param start starting projection.  If \code{NULL}, uses path default.
#' @keywords hplot
#' @export
#' @examples
#' \donttest{
#' tmp_path <- tempdir()
#' render(flea[, 1:6], grand_tour(), display_xy(), "pdf",
#'   frames = 3,
#'   file.path(tmp_path, "test.pdf")
#' )
#' render(flea[, 1:6], grand_tour(), display_xy(), "png",
#'   frames = 3,
#'   file.path(tmp_path, "test-%03d.png")
#' )
#' }
render <- function(data, tour_path, display, dev, ..., apf = 1 / 10, frames = 50, rescale = FALSE, sphere = FALSE, start = NULL) {
  if (!is.matrix(data)) {
    message("Converting input data to the required matrix format.")
    data <- as.matrix(data)
  }
  if (rescale) data <- rescale(data)
  if (sphere) data <- sphere_data(data)

  dev <- match.fun(dev)
  dev(...)
  on.exit(dev.off())

  record <-
    tibble::tibble(
      basis = list(),
      index_val = numeric(),
      info = character(),
      method = character(),
      alpha = numeric(),
      tries = numeric(),
      loop = numeric()
    )

  tour <- new_tour(data, tour_path, start, ...)
  step <- tour(0, ...)

  display$init(data)

  i <- 0
  stop_next <- FALSE
  while (i < frames) {
    display$render_frame()
    display$render_data(data, step$proj, step$target)


    if (stop_next) {
      return(invisible())
    }
    i <- i + 1
    step <- tour(apf)
    if (step$step < 0) stop_next <- TRUE
  }
  invisible()
}


#' Render frames of animation to a gif file
#'
#' @param data matrix, or data frame containing numeric columns
#' @param tour_path tour path generator
#' @param display the method used to render the projected data,
#'   e.g. \code{\link{display_xy}}, \code{\link{display_pcp}}
#' @param gif_file Name of gif file (default = "animation.gif")
#' @param ... other options passed to \code{\link{png}}
#' @param apf angle (in radians) per frame
#' @param frames number of frames in output
#' @param rescale default FALSE. If TRUE, rescale all variables to range [0,1]
#' @param sphere if true, sphere all variables
#' @param start starting projection.  If \code{NULL}, uses path default.
#' @param loop Logical for gifski to loop or not, default=TRUE
#'
#' @examples
#' \dontrun{
#' # gifski needs to be installed to render a gif
#' if (requireNamespace("gifski", quietly = TRUE)) {
#'   gif_file <- file.path(tempdir(), "test.gif")
#'   render_gif(flea[, 1:6], grand_tour(), display_xy(), gif_file)
#'   utils::browseURL(gif_file)
#'   unlink(gif_file)
#' }
#' }
#' @export
render_gif <- function(data, tour_path, display, gif_file = "animation.gif", ..., apf = 1 / 10, frames = 50, rescale = FALSE, sphere = FALSE, start = NULL, loop = TRUE) {
  if (!requireNamespace("gifski", quietly = TRUE)) {
    stop("To use this function please install the 'gifski' package",
      call. = FALSE
    )
  }

  # temp png files
  dir <- tempdir()
  png_path <- file.path(dir, "frame%03d.png")

  render(
    data = data,
    tour_path = tour_path,
    display = display,
    dev = "png",
    png_path,
    ...,
    apf = apf,
    frames = frames,
    rescale = rescale,
    sphere = sphere,
    start = start
  )

  png_files <- sprintf(png_path, 1:frames)
  on.exit(unlink(png_files))

  # Handle smaller number of frames than requested
  # as happens with the guided tour
  png_exist <- sapply(png_files, file.exists)
  if (!all(png_exist)){
    n_png <- sum(png_exist)
    warning(paste0("Note: only ", n_png, " frames generated, argument frames = ", frames, " is ignored."))
  }
  png_files <- png_files[png_exist]

  gifski::gifski(png_files, gif_file, delay = apf, progress = TRUE, loop = loop, ...)
}

#' Render plotly animation frame
#'
#' This function takes a projection matrix as produced by
#' save_history(), and draws it on the projected data
#' like a biplot. This will product the data objects needed
#' in order for the user to plot with base or ggplot2.
#' Note that for now this only works for 2D projections.
#'
#' @param data matrix, or data frame containing numeric columns,
#'   should be standardised to have mean 0, sd 1
#' @param prj projection matrix
#' @param axis_labels of the axes to be displayed
#' @param obs_labels labels of the observations to be available for interactive mouseover
#' @param limits value setting the lower and upper limits of
#'   projected data, default 1
#' @param position position of the axes: center (default),
#'   bottomleft or off
#'
#' @return list containing projected data, circle and segments for axes
#' @export
#' @examples
#' data(flea)
#' flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
#' prj <- basis_random(ncol(flea[,1:6]), 2)
#' p <- render_proj(flea_std, prj)
#' if (require("ggplot2")) {
#'   ggplot() +
#'     geom_path(data=p$circle, aes(x=c1, y=c2)) +
#'     geom_segment(data=p$axes, aes(x=x1, y=y1, xend=x2, yend=y2)) +
#'     geom_text(data=p$axes, aes(x=x2, y=y2, label=rownames(p$axes))) +
#'     geom_point(data=p$data_prj, aes(x=P1, y=P2)) +
#'     xlim(-1,1) + ylim(-1, 1) +
#'     theme_bw() +
#'     theme(aspect.ratio=1,
#'        axis.text=element_blank(),
#'        axis.title=element_blank(),
#'        axis.ticks=element_blank(),
#'        panel.grid=element_blank())
#' }
render_proj <- function(data, prj, axis_labels=NULL, obs_labels=NULL, limits=1, position="center"){
  # Check dimensions ok
  try(if (ncol(data) != nrow(prj))
           stop("Number of columns of data don't match number of rows of prj"))
  try(if(ncol(prj) != 2)
           stop("Number of columns of prj needs to be 2"))

  # Project data and scale into unit box
  data_prj <- as.matrix(data) %*% as.matrix(prj)
  rng <- range(data_prj)
  data_prj <- data_prj/max(abs(rng))
  colnames(data_prj) <- c("P1", "P2")
  data_prj <- data.frame(data_prj)

  # Add observation labels
  if (is.null(obs_labels))
    obs_labels <- as.character(1:nrow(data))
  data_prj$obs_labels <- obs_labels

  # Axis scale
  if (position == "center") {
    axis_scale <- 2 * limits / 3
    axis_pos <- 0
  } else if (position == "bottomleft") {
    axis_scale <- limits / 6
    axis_pos <- -2 / 3 * limits
  }
  adj <- function(x) axis_pos + x * axis_scale

  # Compute segments
  axes <- data.frame(x1=adj(0), y1=adj(0),
                     x2=adj(prj[, 1]), y2=adj(prj[, 2]))
  # Make labels if missing
  if (is.null(axis_labels))
    axis_labels <- colnames(data)
  rownames(axes) <- axis_labels

  # Compute circle
  theta <- seq(0, 2 * pi, length = 50)
  circle <- data.frame(c1 = adj(cos(theta)), c2=adj(sin(theta)))

  return(list(data_prj=data_prj, axes=axes, circle=circle))
}

#' Render a set of animation frames
#'
#' This function takes a set of frames as produced by
#' save_history(), and creates the projected data and axes
#' in for format needed to create the animation using plotly.
#' It will be useful for showing a tour where mouseover can
#' be used to identify points.
#' Note that for now this only works for 2D projections.
#'
#' @param data matrix, or data frame containing numeric columns,
#'   should be standardised to have mean 0, sd 1
#' @param vars numeric columns of data to be projected, as a vector, eg 1:4
#' @param frames array of projection matrices, should be interpolated already
#' @param edges to and from of row id's to connect with an line
#' @param axis_labels labels of the axes to be displayed
#' @param obs_labels labels of the observations to be available for interactive mouseover
#' @param limits value setting the lower and upper limits of
#'   projected data, default 1
#' @param position position of the axes: center (default),
#'   left of data or off
#'
#' @return list containing indexed projected data, edges, circle and segments for axes
#' @export
#' @examples
#' data(flea)
#' flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
#' t1 <- save_history(flea_std, max=2)
#' t1i <- tourr::interpolate(t1, 0.1)
#' p <- render_anim(data=flea_std, frames=t1i)
#' if (require(ggplot2)) {
#'   pg <- ggplot() +
#'     geom_path(data=p$circle, aes(x=c1, y=c2, frame=frame)) +
#'     geom_segment(data=p$axes, aes(x=x1, y=y1, xend=x2, yend=y2, frame=frame)) +
#'     geom_text(data=p$axes, aes(x=x2, y=y2, frame=frame, label=axis_labels)) +
#'     geom_point(data=p$frames, aes(x=P1, y=P2, frame=frame, label=obs_labels)) +
#'     coord_equal() +
#'     theme_bw() +
#'     theme(axis.text=element_blank(),
#'         axis.title=element_blank(),
#'         axis.ticks=element_blank(),
#'         panel.grid=element_blank())
#'   if (interactive()) {
#'     require(plotly)
#'     ggplotly(pg, width=500, height=500) |>
#'       animation_button(label="Go") |>
#'       animation_slider(len=0.8, x=0.5, xanchor="center") |>
#'       animation_opts(easing="linear", transition=0, redraw=FALSE)
#'   }
#'}
render_anim <- function(data, vars=NULL, frames, edges=NULL, axis_labels=NULL, obs_labels=NULL, limits=1, position="center") {
  # Check dimensions ok
  try(if (length(dim(frames)) != 3)
    stop("The frames object needs to be an array with three dimensions"))
  try(if(ncol(frames[,,1]) != 2)
    stop("Number of columns of a frame needs to be 2"))

  # Axis scale
  if (position == "center") {
    axis_scale <- 2 * limits / 3
    axis_pos <- 0
  } else if (position == "bottomleft") {
    axis_scale <- limits / 6
    axis_pos <- -2 / 3 * limits
  }
  adj <- function(x) axis_pos + x * axis_scale

  # Get numeric columns
  if (is.null(vars)) {
    num_vars <- 1:ncol(data)
    extra_vars <- NULL
  } else {
    num_vars <- c(1:ncol(data))[vars]
    extra_vars <- c(1:ncol(data))[-vars]
  }
  try(if (length(num_vars) != nrow(frames[,,1]))
    stop("Number of columns of data don't match number of rows of prj"))

  # Project data into all frames
  # Loop is better than purrr because only working with a small
  # number because object will get too large.
  nf <- dim(frames)[3]
  frames_m <- NULL
  axes_m <- NULL
  for (i in 1:nf) {
    # Data projection
    ftmp <- as.matrix(data[,num_vars]) %*% matrix(frames[,,i], ncol=2)
    ftmp <- cbind(ftmp, data[,extra_vars], rep(i+100, nrow(data)))
    frames_m <- rbind(frames_m, ftmp)
    # Axis display
    axes <- data.frame(x1=adj(0), y1=adj(0),
                       x2=adj(matrix(frames[,,i][,1])),
                       y2=adj(matrix(frames[,,i][,2])),
                       frame=i+100)
    rownames(axes) <- colnames(data[,num_vars])
    axes_m <- rbind(axes_m, axes)
  }

  # Scale into unit box
  rng <- range(frames_m[,1:2])
  frames_m[,1:2] <- frames_m[,1:2]/max(abs(rng))
  colnames(frames_m) <- c("P1", "P2", colnames(data[,extra_vars]), "frame")
  frames_m <- data.frame(frames_m)

  # Add observation labels
  if (is.null(obs_labels))
    obs_labels <- as.character(1:nrow(data))
  frames_m$obs_labels <- rep(obs_labels, nf)

  # Add edge set
  edges_m <- NULL
  if (!is.null(edges)) {
    for (i in 1:nf) {
      x <- frames_m[frames_m$frame == i,]
      e <- data.frame(x=x[edges[,1],1],
                      xend=x[edges[,2],1],
                      y=x[edges[,1],2],
                      yend=x[edges[,2],2],
                      frame=i+100)
      edges_m <- rbind(edges_m, e)
    }
  }

  # Make labels if missing and add to axes object
  if (is.null(axis_labels))
    axis_labels <- colnames(data[,num_vars])
  axes_m$axis_labels <- rep(axis_labels, nf)

  # Compute circle
  theta <- seq(0, 2 * pi, length = 50)
  circle <- data.frame(c1 = adj(cos(theta)),
                       c2 = adj(sin(theta)))
  circle_m <- matrix(rep(t(as.matrix(circle)), nf), ncol=2, byrow=TRUE)
  circle_m <- data.frame(c1 = circle_m[,1],
                         c2 = circle_m[,2],
                         frame=rep(1:nf+100, rep(length(theta), nf)))

  return(list(frames=frames_m, edges=edges_m, axes=axes_m, circle=circle_m))
}

#' Save tour history.
#'
#' Save a tour path so it can later be displayed in many different ways.
#'
#' @param data matrix, or data frame containing numeric columns
#' @param tour_path tour path generator
#' @param max_bases maximum number of new bases to generate.  Some tour paths
#'  (like the guided tour) may generate less than the maximum.
#' @param start starting projection, if you want to specify one
#' @param rescale Default FALSE. If TRUE, rescale all variables to range [0,1]?
#' @param sphere if true, sphere all variables
#' @param step_size distance between each step - defaults to \code{Inf} which
#'   forces new basis generation at each step.
#' @param ... additional arguments passed to tour path
#' @export
#' @examples
#' # You can use a saved history to replay tours with different visualisations
#'
#' t1 <- save_history(flea[, 1:6], max = 3)
#' animate_xy(flea[, 1:6], planned_tour(t1))
#' ## andrews_history(t1)
#' ## andrews_history(interpolate(t1))
#'
#' ## t1 <- save_history(flea[, 1:6], grand_tour(4), max = 3)
#' ## animate_pcp(flea[, 1:6], planned_tour(t1))
#' ## animate_scatmat(flea[, 1:6], planned_tour(t1))
#'
#' ## t1 <- save_history(flea[, 1:6], grand_tour(1), max = 3)
#' ## animate_dist(flea[, 1:6], planned_tour(t1))
#'
#' testdata <- matrix(rnorm(100 * 3), ncol = 3)
#' testdata[1:50, 1] <- testdata[1:50, 1] + 10
#' testdata <- sphere_data(testdata)
#' t2 <- save_history(testdata, guided_tour(holes()),
#'   max = 5
#' )
#' animate_xy(testdata, planned_tour(t2))
#'
#' # Or you can use saved histories to visualise the path that the tour took.
#' plot(path_index(interpolate(t2), holes()))
#'
#' # And you can plot any individual frame using
#' best_prj <- matrix(t2[,,3], ncol=2)
#' p <- render_proj(testdata, best_prj)
#' # which creates a data frame with the elements
#' # to make the plot, see render_proj() for plotting code
#' # OR see draw_tour_axes() for similar code in base graphics
save_history <- function(data, tour_path = grand_tour(), max_bases = 100, start = NULL,
                         rescale = FALSE, sphere = FALSE, step_size = Inf, ...) {
  if (!is.matrix(data)) {
    message("Converting input data to the required matrix format.")
    data <- as.matrix(data)
  }
  if (rescale) data <- rescale(data)
  if (sphere) data <- sphere_data(data)

  record <-
    tibble::tibble(
      basis = list(),
      index_val = numeric(),
      info = character(),
      method = character(),
      alpha = numeric(),
      tries = numeric(),
      loop = numeric()
    )

  if (is.null(start)) {
    start <- tour_path(NULL, data, ...)
  }
  tour <- new_tour(data, tour_path, start, ...)

  projs <- array(NA, c(ncol(data), ncol(start), max_bases + 1))
  projs[,,1] <- start
  princ_dirs <- projs

  i <- 1
  while (i < max_bases) {
    i <- i + 1
    # An infinite step size forces the tour path to generate a new basis
    # every time, so no interpolation occurs.
    step <- tour(step_size)
    if (is.null(step$target)) break # this can happen if no better basis is found in guided tour

    projs[, , i] <- step$target
    if (step$step < 0) break # already appended final projection, break directly
  }

  # Remove empty matrices for tours that terminated early
  # (e.g. guided tour)
  empty <- apply(projs, 3, function(x) all(is.na(x)))
  projs <- projs[, , !empty, drop = FALSE]
  if (length(projs) == 0) {
    return(NULL)
  }

  attr(projs, "data") <- data
  structure(projs, class = "history_array")
}

#' @export
#"[.history_array" <- function(x, i = TRUE, j = TRUE, k = TRUE, ...) {
#  piece <- .subset(x, i, j, k, drop = FALSE)
#  structure(piece,
#    data = attr(x, "data"),
#    class = class(x)
#  )
#}
"[.history_array" <- function(x, i = TRUE, j = TRUE, k = TRUE, ...) {
  # Use unclass() to bypass this method when calling base [,
  # then re-apply drop=FALSE safely via array indexing
  piece <- unclass(x)[i, j, k, drop = FALSE]
  structure(piece,
            data = attr(x, "data"),
            class = class(x)
  )
}
#' @export
"[[.history_array" <- function(x, i, ...) {
  as.matrix(.subset(x, TRUE, TRUE, i, drop = FALSE))
}

#' @export
length.history_array <- function(x) dim(x)[3]

#' @export
str.history_array <- function(object, ...) utils::str(unclass(object))

#' @export
print.history_array <- function(x, ...) {
  attr(x, "data") <- NULL
  NextMethod()
}

#' @export
as.list.history_list <- function(x, ...) x

#' @export
as.list.history_array <- function(x, ...) {
  projs <- lapply(seq_len(dim(x)[3]), function(i) {
    matrix(x[, , i], nrow = dim(x)[1], ncol = dim(x)[2])
  })
  structure(projs, class = "history_list", data = attr(x, "data"))
}
#as.list.history_array <- function(x, ...) {
#  projs <- do.call("c", apply(x, 3, list))
#  structure(projs, class = "history_list", data = attr(x, "data"))
#}

#' @export
as.array.history_array <- function(x, ...) x

#' @export
as.array.history_list <- function(x, ...) {
  dims <- c(nrow(x[[1]]), ncol(x[[1]]), length(x))
  projs <- array(NA, dims)
  for (i in seq_along(x)) {
    projs[, , i] <- x[[i]]
  }
  structure(projs, class = "history_array", data = attr(x, "data"))
}

test_that("interpolate works for 1D tours (grand_tour(1))", {
  # Regression test for https://github.com/ggobi/tourr/issues/XXX
  # On Windows/Linux, apply() drops singleton dimensions in a p x 1 x n
  # history_array, causing [.history_array to be called with mismatched
  # indices and .subset() to throw "incorrect number of dimensions".

  t1 <- save_history(flea[, 1:6], grand_tour(1), max = 3)

  # Confirm the array has the expected singleton-column shape
  expect_equal(dim(t1), c(6L, 1L, 3L))

  # This is the line that failed on Windows/Linux: interpolate() calls
  # as.list.history_array() and indexes slices of the 3D array.
  # It must not error regardless of platform apply() dimension-drop behaviour.
  result <- interpolate(t1, 0.01)

  # Result should be a 3D history_array with correct row/col dimensions
  expect_equal(dim(result)[1], 6L)   # p (number of variables)
  expect_equal(dim(result)[2], 1L)   # d (tour dimension, must stay 1)
  expect_gt(dim(result)[3], 3L)      # more frames after interpolation

  # Each interpolated basis should be a valid unit-norm column
  for (k in seq_len(dim(result)[3])) {
    basis <- matrix(unclass(result)[, , k], nrow = 6)
    expect_equal(
      crossprod(basis)[1, 1], 1,
      tolerance = 1e-6,
      label = sprintf("basis at frame %d has unit norm", k)
    )
  }
})

test_that("as.list.history_array preserves matrix shape for 1D and 2D tours", {
  # as.list() is called internally by interpolate() via planned_tour().
  # For a p x 1 x n array, apply(x, 3, list) can silently drop the
  # singleton second dimension on some platforms, returning vectors
  # instead of matrices. Every element of the list must be a p x d matrix.

  for (d in c(1L, 2L)) {
    t1 <- save_history(flea[, 1:6], grand_tour(d), max = 4)
    lst <- as.list(t1)

    expect_equal(
      length(lst), dim(t1)[3],
      label = sprintf("list length matches n_frames for d=%d", d)
    )

    for (i in seq_along(lst)) {
      m <- lst[[i]]
      expect_true(
        is.matrix(m),
        label = sprintf("element %d is a matrix for d=%d", i, d)
      )
      expect_equal(
        dim(m), c(6L, d),
        label = sprintf("element %d has dim c(6,%d)", i, d)
      )
    }
  }
})

#' Print out the final projection basis
#' @param basis the projection basis to print, used in search_* functions
#' @keywords internal
print_final_proj <- function(basis){
  message("Final projection: ")

  if (ncol(basis) == 1) {
    for (i in 1:length(basis)) {
      message(sprintf("% .3f", basis[i]), " ")
    }
  } else {
    for (i in 1:nrow(basis)) {
      message(paste0(sprintf("% .3f", basis[i, ]), sep = "  "))
    }
  }
}


#' Rescale a matrix or data frame
#'
#' Standardise each column to have range [0, 1]
#'
#' @param df data frame or matrix
#' @keywords manip
#' @export
rescale <- function(df) {
  apply(df, 2, function(x) (x - min(x)) / diff(range(x)))
}

#' Center a numeric vector by subtracting off its mean.
#'
#' @param x numeric vector
#' @export
center <- function(x) {
  scale(x, center = TRUE, scale = FALSE)
}

#' Sphere a matrix (or data frame) by transforming variables to
#' principal components.
#'
#' Sphering is often useful in conjunction with the guided tour, as it
#' removes simpler patterns that may conceal more interesting findings.
#'
#' @param df   data frame or matrix
#' @keywords manip
#' @export
sphere_data <- function(df) {
  apply(stats::predict(stats::prcomp(df)), 2, scale)
}


#' A null function
#'
#' This function does nothing, and is a useful default callback function
#'
#' @param ... all arguments to \code{...} are ignore
#' @name xnul
#' @keywords internal
nul <- function(...) {}


#' Set up a blank plot to display data projections
#' @keywords internal
blank_plot <- function(...) {
  old <- par(mar = c(0, 0, 0, 0))
  # on.exit(par(old))
  plot(
    x = NA, y = NA, xlab = "", ylab = "",
    axes = FALSE, frame = TRUE, xaxs = "i", yaxs = "i",
    ...
  )
}


#' Find the platform
#' Find the platform being used by the user
#'
#' @keywords internal
#' @export
find_platform <- function() {
  os <- R.Version()$os
  gui <- .Platform$GUI

  if (length(grep("linux", os)) == 1) {
    osType <- "lin"
  } else if (length(grep("darwin", os)) == 1) {
    osType <- "mac"
  } else {
    osType <- "win"
  }

  if (gui == "RStudio") {
    type <- "rstudio"
  } else if (osType %in% c("lin", "mac") && gui != "X11") {
    type <- "gui"
  } else if (osType == "win" && gui == "Rgui") {
    type <- "gui"
  } else {
    type <- "cli"
  }

  list(os = osType, iface = type)
}

#' Prints information on how to stop the output
#'
#' This function prints the corresponding information on how to stop the plotting.
#'
#' @keywords internal
to_stop <- function() {
  plat <- find_platform()
  if (plat$os == "win" || plat$iface == "rstudio") {
    key <- "Esc"
  } else if (plat$os == "mac" && plat$iface == "gui") {
    key <- "Esc"
  } else {
    key <- "Ctrl + C"
  }
  message("Press ", key, " to stop tour running")
}

#' Test if all entries are colors
#'
#' @param x vector
#' @export
areColors <- function(x) {
  all(
    sapply(x, function(X) {
      tryCatch(is.matrix(grDevices::col2rgb(X)),
        error = function(e) FALSE
      )
    })
  )
}

#' Map vector of factors to color
#'
#' @param x vector
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @export
mapColors <- function(x, palette) {
  n <- length(unique(x))
  # Handle manual colour setting
  if (length(palette) > 1) {
    stopifnot(length(palette) == n)
    pal <- palette
  }
  else {
    pal <- grDevices::hcl.colors(n, palette=palette)
  }
  pal[as.numeric(as.factor(x))]
}

#' Map vector of factors to pch
#'
#' @param x vector
#' @param shapeset vector of integers indicating point shapes
#' @export
mapShapes <- function(x, shapeset) {
  n <- length(unique(x))
  stopifnot(length(shapeset) >= n)
  shapes <- shapeset
  shapes[as.numeric(x)]
}

#' Create a new tour.
#'
#' The tour function provides the common machinery behind all tour methods:
#' interpolating from basis to basis, and generating new bases when necessary.
#' You should not have to call this function.
#'
#' @param data the data matrix to be projected
#' @param tour_path basis generator, a function that generates a new basis,
#'   called with the previous projection and the data set.  For more
#'   complicated tour paths, this will need to be a closure with local
#'   variables.  Should return NULL if the tour should terminate
#' @param start starting projection, if omitted will use default projection
#'   from generator
#' @seealso \code{\link{save_history}}, \code{\link{render}} and
#'   \code{\link{animate}} for examples of functions that use this function
#'   to run dynamic tours.
#' @keywords hplot dynamic internal
#' @return a function with single argument, step_size.  This function returns
#'  a list containing the new projection, the current target and the number
#'  of steps taken towards the target.
#' @details
#'
#' If you are intended to call \code{new_tour()} from the global environment, try \code{save_history()} and then animate with a \code{planned_tour()}. See \code{\link[tourr]{save_history}} for examples on this.
#'
#' @export
new_tour <- function(data, tour_path, start = NULL, ...) {
  stopifnot(inherits(tour_path, "tour_path"))

  if (is.null(start)) {
    start <- tour_path(NULL, data, ...)
  }

  proj <- list()
  proj[[1]] <- start

  # Initialise first step
  target <- NULL
  step <- 0

  cur_dist <- 0
  target_dist <- 0
  geodesic <- NULL

  function(step_size, ...) {

    step <<- step + 1
    cur_dist <<- cur_dist + step_size

    multi_bases_stop <- inherits(start, "multi-bases") && is.null(proj[[1]])
    if ((target_dist == 0 && step > 1 )|| multi_bases_stop) { # should only happen for guided tour when no better basis is found (relative to starting plane)
      return(list(proj = tail(proj, 1)[[1]], target = target, step = -1)) # use negative step size to signal that we have reached the final target
    }
    # We're at (or past) the target, so generate a new one and reset counters
    if (step_size > 0 && is.finite(step_size) && cur_dist >= target_dist && !inherits(start, "multi-bases")) {

      ## interrupt
      if (attr(tour_path, "name") == "guided") {
        rcd_env <- parent.frame()
        if ("new_basis" %in% rcd_env[["record"]]$info & !rcd_env[["record"]]$method[2] %in% c("search_geodesic", "search_polish")) {
          last_two <- tail(dplyr::filter(rcd_env[["record"]], info == "new_basis"), 2)

          if (last_two$index_val[1] > last_two$index_val[2]) {
            # search_better_random may give probabilistic acceptance, leave it as it is
          } else {
            interp <- dplyr::filter(rcd_env[["record"]], tries == max(tries), info == "interpolation")
            interp <- dplyr::filter(interp, index_val == max(index_val))

            target <- dplyr::filter(rcd_env[["record"]], tries == max(tries), info == "new_basis")

            # deem the target basis as the new current basis if the interpolation doesn't reach the target basis
            # used when the index_f is not smooth
            if (target$index_val > interp$index_val) {
              proj[[length(proj) + 1]] <<- geodesic$ingred$interpolate(1.) # make sure next starting plane is previous target
              target <- dplyr::mutate(target, info = "interpolation", loop = step + 1, alpha = NA)
              rcd_env[["record"]] <- dplyr::add_row(rcd_env[["record"]], target)
            } else if (target$index_val < interp$index_val & nrow(interp) != 0) {
              # the interrupt
              proj[[length(proj) + 1]] <<- interp$basis[[1]]

              rcd_env[["record"]] <- dplyr::filter(
                rcd_env[["record"]],
                id <= which(rcd_env[["record"]]$index_val == interp$index_val)
              )
            }
          }
        } else {
          proj[[length(proj) + 1]] <<- geodesic$ingred$interpolate(1.)
          if (nrow(rcd_env[["record"]]) != 0){
            rcd_env[["record"]] <- dplyr::add_row(
              rcd_env[["record"]],
              basis = list(proj[[length(proj)]]),
              index_val = geodesic$index(proj[[length(proj)]]),
              info = "interpolation",
              tries = geodesic$tries,
              method = dplyr::last(rcd_env[["record"]]$method),
              loop = step + 1
            )

            rcd_env[["record"]] <- dplyr::mutate(
              rcd_env[["record"]],
              id = dplyr::row_number()
            )
          }
        }
      }
      else {
        proj[[length(proj) + 1]] <<- geodesic$ingred$interpolate(1)
      }
    }

    if (cur_dist >= target_dist) {
      if (inherits(proj[[1]], "multi-bases")){
        geodesic <<- tour_path(proj[[1]], data, ...)
        proj <<- list(geodesic$target)
        target_dist <<- 10
        cur_dist <<-10
        return(list(proj = proj[[1]], target = NULL, step = -1)) # use negative step size to signal that we have reached the final target
      }

      geodesic <<- tour_path(proj[[length(proj)]], data, ...)
      if (is.null(geodesic$ingred)) {
        return(list(proj = proj[[length(proj)]], target = target, step = -1)) # use negative step size to signal that we have reached the final target
      }

      target_dist <<- geodesic$ingred$dist
      target <<- geodesic$ingred$Fz
      cur_dist <<- 0
      # Only exception is if the step_size is infinite - we want to jump
      # to the target straight away
      if (!is.finite(step_size)) {
        cur_dist <<- target_dist
      }

      step <<- 0
      proj <<- list()
      proj[[1]] <<- start
    }

    proj[[step + 2]] <<- geodesic$ingred$interpolate(cur_dist / target_dist)


    if (attr(tour_path, "name") == "guided") {
      rcd_env <- parent.frame()
      rcd_env[["record"]] <- dplyr::add_row(
        rcd_env[["record"]],
        basis = list(proj[[step + 2]]),
        index_val = geodesic$index(proj[[step + 2]]),
        info = "interpolation",
        tries = geodesic$tries,
        method = dplyr::last(rcd_env[["record"]]$method),
        loop = step + 1
      )
      rcd_env[["record"]] <- dplyr::mutate(
        rcd_env[["record"]],
        id = dplyr::row_number()
      )
    }


    list(proj = proj[[length(proj)]], target = target, step = step)
  }
}
# globalVariables(c("basis", "id", "index", "index_val"))

#' @importFrom grDevices dev.cur dev.flush dev.hold dev.off hcl rgb
#' @importFrom graphics abline axis box hist image lines pairs par plot points polygon rect segments stars text arrows
NULL

#' Scatterplot matrix tour path animation.
#'
#' Animate a nD tour path with a scatterplot matrix.
#'
#' The lines show the observations, and the points, the values of the
#' projection matrix.
#'
#' @param ... other arguments passed on to \code{\link{animate}}
#' @seealso \code{\link{animate}} for options that apply to all animations
#' @keywords hplot
#' @export
#' @examples
#' animate_scatmat(flea[, 1:6], grand_tour(2))
#' animate_scatmat(flea[, 1:6], grand_tour(6))
display_scatmat <- function(...) {
  render_data <- function(data, proj, geodesic) {
    pairs(data %*% proj, pch = 20, ...)
  }

  list(
    init = nul,
    render_frame = nul,
    render_transition = nul,
    render_data = render_data,
    render_target = nul
  )
}


#' @rdname display_scatmat
#' @inheritParams animate
#' @export
animate_scatmat <- function(data, tour_path = grand_tour(3), ...) {
  animate(
    data = data, tour_path = tour_path,
    display = display_scatmat(...), ...
  )
}

#' Display tour path with trails
#'
#' Animate a 2D tour path with a point trails
#'
#' @param axes position of the axes: center, bottomleft or off
#' @param center if TRUE, centers projected data to (0,0).  This pins the
#'  center of data cloud and make it easier to focus on the changing shape
#'  rather than position.
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param col color to be plotted.  Defaults to "black"
#' @param pch shape of the point to be plotted.  Defaults to 20.
#' @param cex size of the point to be plotted.  Defaults to 1.
#' @param past draw line between current projection and projection \code{past}
#'   steps ago
#' @param cex magnification of plotting text relative to default. Defaults to 1.
#' @param axislablong text labels only for the long axes in a projection, default FALSE
#' @param ...  other arguments passed on to \code{\link{animate}} and
#'   \code{\link{display_xy}}
#' @export
#' @examples
#' animate_trails(flea[,1:6], col=flea$species)
#'
display_trails <- function(center = TRUE, axes = "center", half_range = NULL, col = "black", pch = 20, cex = 1, past = 3, axislablong = FALSE, ...) {

  # Inherit most behaviour from display_xy.  This is a little hacky, but
  # the only way until tourr switch to a proper object system.
  xy <- display_xy(
    center = center, axes = axes, half_range = half_range,
    col = col, pch = pch, cex = cex, ...
  )
  xy_env <- environment(xy$init)

  xy_env$past <- past
  xy_env$past_x <- vector("list", past)

  # Only difference is the display method
  render_data <- function(data, proj, geodesic) {
    draw_tour_axes(proj, labels, 1, axes, longlabels=axislablong)

    x <- data %*% proj
    if (center) x <- center(x)
    x <- x / half_range

    # Render projected points
    last_x <- past_x[[1]]
    if (!is.null(last_x)) {
      segments(last_x[, 1], last_x[, 2], x[, 1], x[, 2], col = col, lwd = cex*2)
    }
    points(x, col = col, pch = pch, cex = cex)

    past_x <<- c(past_x[2:past], list(x))
  }
  environment(render_data) <- xy_env
  # Quiet R CMD check note:
  past_x <- NULL

  xy$render_data <- render_data
  xy
}

# globalVariables("past_x")

#' @rdname display_trails
#' @inheritParams animate
#' @export
animate_trails <- function(data, tour_path = grand_tour(), ...) {
  animate(data, tour_path, display_trails(...), ...)
}

#' Display tour path with a scatterplot
#'
#' Animate a 2D tour path with a scatterplot.
#'
#' @param axes position of the axes: center, bottomleft or off
#' @param center if TRUE, centers projected data to (0,0).  This pins the
#'  center of data cloud and make it easier to focus on the changing shape
#'  rather than position.
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param edges A two column integer matrix giving indices of ends of lines.
#' @param col color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".
#' @param pch shape of the point to be plotted, can be a factor or integer.  Defaults to 20.
#' @param cex size of the point to be plotted.  Defaults to 1.
#' @param edges.col colour of edges to be plotted, Defaults to "black"
#' @param edges.width line width for edges, default 1
#' @param obs_labels vector of text labels to display
#' @param ellipse pxp variance-covariance matrix defining ellipse, default NULL. Useful for
#'        comparing data with some null hypothesis
#' @param ellc This can be considered the equivalent of a critical value, used to
#'        scale the ellipse larger or smaller to capture more or fewer anomalies. Default 3.
#' @param ellmu This is the centre of the ellipse corresponding to the mean of the
#'        normal population. Default vector of 0's
#' @param ellmarks mark the extreme points with red crosses, default TRUE
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @param shapeset numbers corresponding to shapes in base R points, to use for mapping
#'        categorical variable to shapes, default=c(15:17, 23:25)
#' @param axislablong text labels only for the long axes in a projection, default FALSE
#' @param ...  other arguments passed on to \code{\link{animate}} and
#'   \code{\link{display_xy}}
#' @importFrom graphics legend
#' @importFrom stats mahalanobis qchisq
#' @export
#' @examples
#' animate_xy(flea[, 1:6])
#' animate(flea[, 1:6], tour_path = grand_tour(), display = display_xy())
#' # This won't do anything because the flea data is standardised
#' # but use rescale option to force scaling before displaying
#' animate(flea[, 1:6],
#'   tour_path = grand_tour(),
#'   display = display_xy(),
#'   rescale = TRUE
#' )
#' animate(flea[, 1:6],
#'   tour_path = grand_tour(),
#'   display = display_xy(half_range = 0.5)
#' )
#' animate_xy(flea[, 1:6], tour_path = little_tour())
#' animate_xy(flea[, 1:3], tour_path = guided_tour(holes()), sphere = TRUE)
#' animate_xy(flea[, 1:6], center = FALSE)
#'
#' # The default axes are centered, like a biplot, but there are other options
#' animate_xy(flea[, 1:6], axes = "bottomleft")
#' animate_xy(flea[, 1:6], axes = "off")
#' animate_xy(flea[, 1:6], dependence_tour(c(1, 2, 1, 2, 1, 2)),
#'   axes = "bottomleft"
#' )
#'
#' animate_xy(flea[, -7], col = flea$species)
#' animate_xy(flea[, -7], col = flea$species,
#'              pch = flea$species)
#'
#' animate_xy(flea[, -7], col = flea$species,
#'   obs_labels=as.character(1:nrow(flea)), axes="off")
#'
#' # You can also draw lines
#' edges <- matrix(c(1:5, 2:6), ncol = 2)
#' animate(
#'   flea[, 1:6], grand_tour(),
#'   display_xy(axes = "bottomleft", edges = edges)
#' )
#' # An ellipse can be drawn on the data using a specified var-cov
#' animate_xy(flea[, 1:6], axes = "off", ellipse=cov(flea[,1:6]))
display_xy <- function(center = TRUE, axes = "center", half_range = NULL,
                       col = "black", pch = 20, cex = 1,
                       edges = NULL, edges.col = "black", edges.width=1,
                       obs_labels = NULL,
                       ellipse = NULL, ellc = NULL, ellmu = NULL,
                       ellmarks = TRUE,
                       palette="Zissou 1", shapeset=c(15:17, 23:25),
                       axislablong = FALSE, ...) {
  # Needed for CRAN checks
  labels <- NULL
  gps <- NULL
  shapes <- NULL

  # If colors are a variable, convert to colors
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    col <- mapColors(col, palette)
  }
  if (is.factor(edges.col) | !areColors(edges.col)) {
    edges.gps <- edges.col
    edges.col <- mapColors(edges.col, palette)
  }
  # If shapes are a variable, convert shapes
  if (is.factor(pch)) {
    shapes <- mapShapes(pch, shapeset)
  } else {
    shapes <- pch
  }

  init <- function(data) {
    half_range <<- compute_half_range(half_range, data, center)
    labels <<- abbreviate(colnames(data), 3)

    if (!is.null(ellipse)) {
      if (nrow(ellipse) == ncol(data)) {

        if (is.null(ellc))
          ellc <<- qchisq(0.95, ncol(data))
        else
          stopifnot(ellc > 0) # Needs to be positive
        if (is.null(ellmu))
          ellmu <<- rep(0, ncol(data))
        else
          stopifnot(length(ellmu) == ncol(data)) # Right dimension
        message("Using ellc = ", format(ellc, digits = 2))
      }
    }
  }

  if (!is.null(edges)) {
    if (!is.matrix(edges) && ncol(edges) == 2) {
      stop("Edges matrix needs two columns, from and to, only.")
    }
  }

  render_frame <- function() {
    par(pty = "s", mar = rep(0.1, 4))
    blank_plot(xlim = c(-1, 1), ylim = c(-1, 1))
  }
  render_transition <- function() {
    rect(-1, -1, 1, 1, col = "#FFFFFFE6", border = NA)
  }
  render_data <- function(data, proj, geodesic) {
    draw_tour_axes(proj, labels, limits = 1, axes, longlabels=axislablong, ...)

    # Render projected points
    x <- data %*% proj
    if (center) x <- center(x)
    x <- x / half_range
    points(x, col = col, pch = shapes, cex = cex)

    # Render labels for obs, if provided
    if (!is.null(obs_labels)) {
      text(x, labels=obs_labels, col=col, pos=4, offset=0.1)
    }
    # Draw segments between points, if provided
    if (!is.null(edges)) {
      segments(x[edges[, 1], 1], x[edges[, 1], 2],
        x[edges[, 2], 1], x[edges[, 2], 2],
        col = edges.col,
        lwd = edges.width
      )
    }

    # add a legend, only if a variable was used
    if (is.factor(gps)) {
      numcol <- unique(col)
      if (length(numcol) > 1) {
        legend("topright", legend=unique(gps),
             col=numcol, pch=15)
      }
    }
    if (is.factor(pch)) {
      numpch <- unique(shapes)
      if (length(numpch) > 1) {
        legend("bottomright", legend=unique(pch),
             col="black", pch=unique(shapes))
      }
    }

    # Add index value if using guided tour
    #if (!is.na(cur_index))
    #  text(0, 0, labels=round(cur_index, 2))
    # Draw a pre-determined ellipse on the data
    if (!is.null(ellipse)) {
      if (nrow(ellipse) == nrow(proj)) {

      #  if (is.null(ellc))
      #    ellc <- qchisq(0.95, nrow(proj))
      #  else
      #    stopifnot(ellc > 0) # Needs to be positive
      #  if (is.null(ellmu))
      #    ellmu <- rep(0, nrow(proj))
      #  else
      #    stopifnot(length(ellmu) == nrow(proj)) # Right dimension
      #  message("Using ellc = ", format(ellc, digits = 2))

        # Project ellipse into 2D
        # Notation in paper: ellipse=A, ellinv=A^(-1),
        #                    e2=P^TA^(-1)P, ell2d=B
        # ellipse=var-cov of normal pop
        # ellinv for defining pD ellipse, for dist calc
        # e2 is projected var-cov
        # ell2d is B, used to project ellipse points
        evc <- eigen(ellipse) #
        ellinv <- (evc$vectors) %*% diag(evc$values) %*% t(evc$vectors)
        e2 <- t(proj) %*% ellipse %*% proj
        evc2 <- eigen(e2)
        ell2d <- as.matrix((evc2$vectors)) %*% diag(sqrt(evc2$values*ellc)) %*% t(as.matrix(evc2$vectors))
        #e3 <- eigen(ell2d)
        #ell2dinv <- (e3$vectors) %*% diag(e3$values) %*% t(e3$vectors)

        # Compute the points on an ellipse
        # Generate points on a circle
        sph <- geozoo::sphere.hollow(2, 200)$points
        # Organise so lines connecting consecutive
        # points creates the circle
        sph <- sph[order(sph[,2]),]
        sph1 <- sph[sph[,2]>=0,]
        sph2 <- sph[sph[,2]<0,]
        sph1 <- sph1[order(sph1[,1]),]
        sph2 <- sph2[order(sph2[,1], decreasing=T),]
        sph <- rbind(sph1, sph2)
        sph <- rbind(sph, sph[1,])

        # Transform circle points into an ellipse
        sph2d <- sph%*%ell2d
        # Centre on the given mean
        ellmu2d <- t(as.matrix(ellmu)) %*% proj
        sph2d <- sweep(sph2d, 2, ellmu2d, `+`)
        # Scale ellipse into plot space
        sph2d <- sph2d/half_range

        lines(sph2d)

        # Colour points outside the pD ellipse
        mdst <- mahalanobis(data,
                            center=ellmu,
                            cov=ellipse)
        #mdst <- mahal_dist(data, ellipse)
        anomalies <- which(mdst > ellc)
        #cat("1 ", length(anomalies), "\n")
        if (length(anomalies) > 0 & ellmarks) {
          points(x[anomalies,],
               col = "red",
               pch = 4,
               cex = 2)
        }
      }
      else
        message("Check the variance-covariance matrix generating the ellipse\n")
    }
  }

  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}

#' @rdname display_xy
#' @inheritParams animate
#' @export
animate_xy <- function(data, tour_path = grand_tour(), ...) {
  animate(data, tour_path, display_xy(...), ...)
}

#' Draw tour axes on the projected data with base graphics
#'
#' @param proj matrix of projection coefficients
#' @param labels variable names for the axes, of length the same
#'   as the number of rows of proj
#' @param limits value setting the lower and upper limits of
#'   projected data, default 1
#' @param position position of the axes: center (default),
#'   bottomleft or off
#' @param axis.col colour of axes, default "grey50"
#' @param axis.lwd linewidth of axes, default 1
#' @param axis.text.col colour of axes text, default "grey50"
#' @param longlabels text labels only for the long axes in a projection, default FALSE
#' @param ...  other arguments passed
#' @export
#' @examples
#' data(flea)
#' flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
#' prj <- basis_random(ncol(flea[,1:6]), 2)
#' flea_prj <- as.data.frame(as.matrix(flea_std) %*% prj)
#' par(pty = "s", mar = rep(0.1, 4))
#' plot(flea_prj$V1, flea_prj$V2,
#'      xlim = c(-3, 3), ylim = c(-3, 3),
#'      xlab="P1", ylab="P2")
#' draw_tour_axes(prj, colnames(flea)[1:6], limits=3)
#'
#' plot(flea_prj$V1, flea_prj$V2,
#'      xlim = c(-3, 3), ylim = c(-3, 3),
#'      xlab="P1", ylab="P2")
#' draw_tour_axes(prj, colnames(flea)[1:6], limits=3, position="bottomleft")
#' draw_tour_axes(prj, colnames(flea)[1:6], longlabels=TRUE)
draw_tour_axes <- function(proj, labels, limits=1, position="center",
                           axis.col="grey50", axis.lwd=1, axis.text.col="grey50",
                           longlabels=FALSE, ...) {
  position <- match.arg(position, c("center", "bottomleft", "off"))
  if (position == "off") {
    return()
  }

  if (position == "center") {
    axis_scale <- 2 * limits / 3
    axis_pos <- 0
  } else if (position == "bottomleft") {
    axis_scale <- limits / 6
    axis_pos <- -2 / 3 * limits
  }

  adj <- function(x) axis_pos + x * axis_scale

  segments(adj(0), adj(0), adj(proj[, 1]), adj(proj[, 2]),
           col = axis.col, lwd = axis.lwd)
#  if (!is.null(mvar)) { # colour manip var
#    if ((mvar < (nrow(proj)+1)) & (mvar > 0)) {
#      segments(adj(0), adj(0), adj(proj[, 1]), adj(proj[, 2]), col = "orange")
#    }
#  }
  theta <- seq(0, 2 * pi, length = 50)
  lines(adj(cos(theta)), adj(sin(theta)),
        col = axis.col, lwd = axis.lwd)
  if (longlabels) {
    for (i in 1:length(labels)) {
      if ((proj[i, 1]^2 + proj[i, 2]^2) < 0.15)
        labels[i] <- ""
    }
  }
  text(adj(proj[, 1]), adj(proj[, 2]), label = labels,
       col = axis.text.col)
}

# Tests for display_* functions
#
# Strategy: display functions return a list of closures (init, render_frame,
# render_transition, render_data, render_target). We test them by:
#   1. Checking the returned object has the right structure.
#   2. Calling each closure with realistic inputs inside a null PDF device
#      so we exercise the drawing code without opening a screen window.
#
# We do NOT call animate() in these tests — that would open an interactive
# graphics device and run indefinitely. Instead we drive the closures directly,
# which is both faster and more precise about what is being tested.

# ---------------------------------------------------------------------------
# Shared test fixtures
# ---------------------------------------------------------------------------

dat   <- as.matrix(flea[, 1:6])          # 74 x 6 numeric matrix
proj2 <- basis_random(6, 2)              # 6 x 2 orthonormal basis  (2D tour)
proj1 <- basis_random(6, 1)              # 6 x 1 orthonormal basis  (1D tour)

# Helper: run `expr` inside a null PDF device so base-graphics calls succeed
# but nothing is drawn to screen or disk.
#
# pdf(NULL) opens the device but does not create a plot frame. Drawing
# primitives like points(), text(), and segments() require plot.new() to have
# been called first, which is normally done by the display's render_frame()
# closure. We call plot.new() here so that tests which call render_data()
# directly (without going through render_frame() first) do not fail with
# "plot.new has not been called yet".
with_null_device <- function(expr) {
  pdf(NULL)
  on.exit(dev.off(), add = TRUE)
  plot.new()
  force(expr)
}


# ===========================================================================
# display_xy
# ===========================================================================

test_that("display_xy returns a correctly structured display object", {
  d <- display_xy()
  expect_named(d, c("init", "render_frame", "render_transition",
                    "render_data", "render_target"))
  expect_true(all(vapply(d, is.function, logical(1))))
})

test_that("display_xy init sets half_range from data", {
  d <- display_xy()
  with_null_device({
    d$init(dat)
    # render_data should now work without error, confirming init ran correctly
    expect_no_error(d$render_data(dat, proj2, geodesic = FALSE))
  })
})

test_that("display_xy render_data accepts col and pch vectors", {
  col_vec <- ifelse(flea$species == "Heikert'", "steelblue", "tomato")
  d <- display_xy(col = col_vec, pch = 20)
  with_null_device({
    d$init(dat)
    expect_no_error(d$render_data(dat, proj2, geodesic = FALSE))
  })
})

test_that("display_xy respects a fixed half_range", {
  # When half_range is fixed, the projected coordinates should be scaled by
  # exactly that value. We verify render_data runs without error and that
  # init does not override the supplied half_range.
  d <- display_xy(half_range = 1)
  with_null_device({
    d$init(dat)
    expect_no_error(d$render_data(dat, proj2, geodesic = FALSE))
  })
})

test_that("display_xy works with axes = 'off'", {
  d <- display_xy(axes = "off")
  with_null_device({
    d$init(dat)
    expect_no_error(d$render_frame())
    expect_no_error(d$render_data(dat, proj2, geodesic = FALSE))
  })
})

test_that("display_xy works with edges argument", {
  # edges is an n x 2 integer matrix of point-index pairs to connect
  edges <- matrix(c(1L, 2L, 3L, 4L), ncol = 2)
  d <- display_xy(edges = edges)
  with_null_device({
    d$init(dat)
    expect_no_error(d$render_data(dat, proj2, geodesic = FALSE))
  })
})


# ===========================================================================
# display_dist  (1D distribution display)
# ===========================================================================

test_that("display_dist returns a correctly structured display object", {
  d <- display_dist()
  expect_named(d, c("init", "render_frame", "render_transition",
                    "render_data", "render_target"))
  expect_true(all(vapply(d, is.function, logical(1))))
})

test_that("display_dist render_data works with a 1D projection", {
  d <- display_dist()
  with_null_device({
    d$init(dat)
    expect_no_error(d$render_data(dat, proj1, geodesic = FALSE))
  })
})

test_that("display_dist render_frame does not error", {
  d <- display_dist()
  with_null_device({
    d$init(dat)
    expect_no_error(d$render_frame())
  })
})


# ===========================================================================
# display_pcp  (parallel coordinates)
# ===========================================================================

test_that("display_pcp returns a correctly structured display object", {
  d <- display_pcp()
  expect_named(d, c("init", "render_frame", "render_transition",
                    "render_data", "render_target"))
  expect_true(all(vapply(d, is.function, logical(1))))
})

test_that("display_pcp render_data works with a 2D projection", {
  d <- display_pcp()
  with_null_device({
    d$init(dat)
    expect_no_error(d$render_data(dat, proj2, geodesic = FALSE))
  })
})

test_that("display_pcp accepts col argument", {
  col_vec <- rep(c("red", "blue"), length.out = nrow(dat))
  d <- display_pcp(col = col_vec)
  with_null_device({
    d$init(dat)
    expect_no_error(d$render_data(dat, proj2, geodesic = FALSE))
  })
})


# ===========================================================================
# display_density2d
# ===========================================================================

test_that("display_density2d returns a correctly structured display object", {
  d <- display_density2d()
  expect_named(d, c("init", "render_frame", "render_transition",
                    "render_data", "render_target"))
  expect_true(all(vapply(d, is.function, logical(1))))
})

test_that("display_density2d render_data works with a 2D projection", {
  d <- display_density2d()
  with_null_device({
    d$init(dat)
    expect_no_error(d$render_data(dat, proj2, geodesic = FALSE))
  })
})


# ===========================================================================
# Cross-cutting: render_target is NULL for all standard displays
# (render_target is only non-NULL for guided-tour displays)
# ===========================================================================

test_that("standard display render_target closures are nul/NULL", {
  displays <- list(
    display_xy(),
    display_dist(),
    display_pcp(),
    display_density2d()
  )
  for (d in displays) {
    # render_target should either be the package's `nul` function or NULL
    rt <- d$render_target
    expect_true(
      is.null(rt) || identical(rt, tourr:::nul),
      label = "render_target is nul or NULL"
    )
  }
})

#' Calculate the squared Euclidean norm of a vector x
#'
#' @param x numeric vector
#' @keywords internal
eucl_norm_sq <- function(x) {
  sum(x^2)
}

#' Calculate orthogonal distances
#'
#' For each datapoint this function calculates the orthogonal distance from the
#' anchored projection plane.
#'
#' @param plane matrix specifying the projection plane
#' @param data data frame or matrix
#' @param anchor A vector specifying the reference point to anchor the plane
#'   If NULL (default) the slice will be anchored at the origin.
#' @return distance vector
#' @export
anchored_orthogonal_distance <- function(plane, data, anchor = NULL) {
  # by default anchor is at the origin
  if (is.null(anchor)) {
    anchor <- matrix(colMeans(data), ncol = length(plane[, 1]))
  } else {
    anchor <- matrix(anchor, ncol = length(plane[, 1]))
  }

  # this is the squared length of alpha (orthogonal component of anchor on plane wrt the origin)
  alpha_sq <- sum(anchor^2) - sum((anchor %*% plane)^2)

  # for each datapoint calculate the squarred length of orthogonal distance from plane (anchored at zero)
  full_norm_sq <- apply(data, 1, eucl_norm_sq)
  subtr1 <- as.matrix(data) %*% plane[, 1]
  subtr2 <- as.matrix(data) %*% plane[, 2]
  dist_sq <- full_norm_sq - subtr1^2 - subtr2^2

  # finally there is a cross term that we need to take into account
  xterm <- 2 * (as.matrix(data) %*% t(anchor) -
    subtr1 * drop(anchor %*% plane[, 1]) -
    subtr2 * drop(anchor %*% plane[, 2]))

  # putting all three components together we can calculate the anchored distance
  sqrt(dist_sq + alpha_sq - xterm)
}

#' If not set, compute epsilon based on half_range
#'
#' @keywords internal
compute_v_rel <- function(v_rel, half_range, n) {
  if (!is.null(v_rel)) {
    return(v_rel)
  }

  (half_range^(n - 2)) / 10.
}

#' Display a 1D linear aggregation index
#'
#' Animate a 1D tour path for data where individuals are ranked
#' by a multivariate index. Allows one to examine the sensitivity
#' of the ranking on the linear combination. Variables should be
#' scaled to be between 0-1. This is only designed to work with
#' a local tour, or a radial tour.
#'
#' @param center should 1d projection be centered to have mean zero (default: TRUE).
#'   This pins the centre of distribution to the same place, and makes it
#'   easier to focus on the shape of the distribution.
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param abb_vars logical, whether to abbreviate the variable name, if long
#' @param col the color used for points, can be a vector or hexcolors or a
#'   factor, default to "red".
#' @param cex the size used for points, default to 0.5
#' @param panel_height_ratio input to the height argument in
#'   [graphics::layout()] for the height of data and axis panel.
#' @param add_ref_line whether to add a horizontal reference line for each
#' observation, logical default to TRUE
#' @param label the text label, a vector
#' @param label_cex the size for text labels
#' @param label_col the color for text labels
#' @param label_x_pos the x position of text label, currently labels are
#'   positioned at a fixed x value for each observation
#' @param axis_label_cex_upper the size of the axis label in the upper panel
#' @param axis_label_cex_lower the size of the axis label in the lower panel
#' @param axis_bar_col the color of the axis bar
#' @param axis_bar_lwd the width of the axis bar
#' @param axis_bar_label_cex the size of the axis label
#' @param axis_bar_label_col the color of the axis label
#' @param axis_var_cex the size of the variable name
#'   to the right of the axis panel
#' @param axis_var_col the color of the variable name
#'   to the right of the axis panel
#' @param palette name of color palette for point colour, used by
#'   \code{\link{hcl.colors}}, default "Zissou 1"
#' @export
#' @rdname display_idx
#' @examples
#' data(places)
#' places_01 <- apply(places[1:10,1:9], 2, function(x) (x-min(x))/(max(x)-min(x)))
#' b <- matrix(rep(1/sqrt(9), 9), ncol=1)
#' places_init <- cbind(places_01, idx = as.vector(as.matrix(places_01) %*% b))
#' places_sorted <- places_init[order(places_init[,10]), 1:9]
#' animate_idx(places_sorted, tour_path = local_tour(b, angle=pi/8),
#'             label=as.character(places$stnum[1:9]),
#'             label_x_pos = 0)
display_idx <- function(center = FALSE, half_range = NULL, abb_vars = TRUE,
                        col = "red", cex = 3, panel_height_ratio = c(3, 2),
                        label_x_pos = 0.7, label = NULL,
                        label_cex = 1, label_col = "grey80", add_ref_line = TRUE,
                        axis_bar_col = "#000000", axis_bar_lwd = 3,
                        axis_label_cex_upper = 1, axis_label_cex_lower = 1,
                        axis_bar_label_cex = 1, axis_bar_label_col = "#000000",
                        axis_var_cex = 1, axis_var_col = "#000000",
                        palette = "Zissou 1", ...) {
  labels <- NULL
  init <- function(data) {
    half_range <<- compute_half_range(half_range, data, center)
    if (abb_vars) {
      labels <<- abbreviate(colnames(data), 2)
    } else {
      labels <<- colnames(data)
    }
  }
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    lgps <- levels(gps)
    col <- mapColors(col, palette)
  }
  colrs <- unique(col)

  render_frame <- function() {
    # define the plot to have an upper & lower panel, height as 6:1
    layout_m <- matrix(c(1, 2), nrow = 2, ncol = 1)
    graphics::layout(mat = layout_m, heights = panel_height_ratio)
    par(pty = "m", mar = c(1, 4, 1, 1))

  }
  render_transition <- function() {
    rect(-0.1, -1.1, 1.2, 3, col = "#FFFFFFE6", border = NA)
  }
  render_data <- function(data, proj, geodesic) {

    x <- data %*% proj
    if (center) x <- center(x)
    x <- x / half_range
    df <- cbind(x, .y = (3 - 0.1)/nrow(x) * (1:nrow(x)))
    # upper panel: define the axis margin (mgp) and panel margin (mar)
    par(mgp=c(2, 1, 0), mar = c(2, 4, 0.5, 1))
    # initialise the plot box
    plot(
      x = NA, y = NA, xlim = c(-0.1, 1.2), ylim = c(0, 3.05),
      xlab="", ylab = "", xaxs = "i", yaxs = "i", cex = 2,
      xaxt = "n", yaxt = "n"
    )
    axis(1, seq(0, 1, 0.2), cex.axis = axis_label_cex_upper)
    if (add_ref_line) abline(h = df[,".y"], col = "grey60")
    text(x=label_x_pos, y=df[,".y"], labels= label,
         col = label_col, cex = label_cex)
    points(df, col = col, pch = 20, cex = cex)

    # lower panel: define panel margin (mar)
    par(mgp=c(2, 1, 0), mar = c(2.5, 4, 1, 1))
    plot(
      x = NA, y = NA, xlim = c(0, 1.3), ylim = c(-1, 0),
      xlab = "", ylab = "", xaxt = "n", yaxt = "n"
    )
    lines(c(0, 0), c(-1.1, 0), col = "grey60")
    lines(c(1, 1), c(-1.1, 0), col = "grey60")
    axis(1, seq(0, 1, by = 0.2), cex.axis = axis_label_cex_lower)
    box(col = "grey60")

    # Render tour axes
    ax <- (seq_along(proj) - 0.5) / length(proj)
    idx_w <- proj/sum(proj)
    segments(0, -ax, idx_w, -ax, col = axis_bar_col, lwd = axis_bar_lwd)
    text(1.0, -ax, labels, pos = 4, cex = axis_var_cex, col = axis_var_col)
    text(idx_w + 0.01, -ax, format(round(idx_w, 2), nsmall = 2), pos = 4,
         cex = axis_bar_label_cex, col = axis_bar_label_col)
  }

  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}

#' @rdname display_idx
#' @inheritParams animate
#' @export
animate_idx <- function(data, tour_path = grand_tour(1), ...) {
  animate(
    data = data, tour_path = tour_path,
    display = display_idx(...), ...
  )
}

#' Draw the path that the geodesics took.
#'
#' This computes the projected values of each observation at each step, and
#' allows you to recreate static views of the animated plots.
#'
#' @param history list of bases produced by \code{\link{save_history}}
#'   (or otherwise)
#' @param data dataset to be projected on to bases
#' @export
#' @examples
#' path1d <- save_history(flea[, 1:6], grand_tour(1), 3)
#' path2d <- save_history(flea[, 1:6], grand_tour(2), 3)
#'
#' if (require("ggplot2")) {
#'   plot(path_curves(path1d))
#'   plot(path_curves(interpolate(path1d)))
#'
#'   plot(path_curves(path2d))
#'   plot(path_curves(interpolate(path2d)))
#'
#'   # Instead of relying on the built in plot method, you might want to
#'   # generate your own.  Here are few examples of alternative displays:
#'
#'   df <- path_curves(path2d)
#'   ggplot(data = df, aes(x = step, y = value, group = obs:var, colour = var)) +
#'     geom_line() +
#'     facet_wrap(~obs)
#'
#'   library(tidyr)
#'   ggplot(
#'     data = pivot_wider(df,
#'       id_cols = c(obs, step),
#'       names_from = var, names_prefix = "Var",
#'       values_from = value
#'     ),
#'     aes(x = Var1, y = Var2)
#'   ) +
#'     geom_point() +
#'     facet_wrap(~step) +
#'     coord_equal()
#' }
path_curves <- function(history, data = attr(history, "data")) {
  history <- as.list(history)
  n <- length(history)

  project <- function(basis) {
    proj <- data %*% basis
    data.frame(
      obs = factor(row(proj)),
      var = factor(col(proj)),
      value = as.vector(proj)
    )
  }
  projections <- do.call("rbind", lapply(history, project))
  projections$step <- rep(seq_len(n), each = nrow(data) * ncol(history[[1]]))
  class(projections) <- c("path_curve", class(projections))

  projections
}

#' Plot history curves.
#'
#' The default plot method is a line plot with step on the x axis and
#' value on the y axis.  Each observation is drawn with a different line
#' line and the plot is facetted by variable.  This is rather similar in
#' spirit to a parallel coordinates plot or Andrews curves.
#'
#' For alternative ways of plotting this data, see
#'  \code{\link{path_curves}}
#' @keywords internal
#' @export
plot.path_curve <- function(x, ...) {
  step <- NULL # quiet R CMD check warning

  ggplot2::ggplot(data = x, ggplot2::aes(x = step, y = value, group = obs)) +
    ggplot2::geom_line() +
    ggplot2::facet_grid(var ~ .)
}
# globalVariables(c("value", "obs"))

#' Chernoff faces tour path animation.
#'
#' Animate a nD tour path with Chernoff's faces.  Can display up to 18
#' dimensions.
#'
#' This function requires the \code{aplpack} package to draw the
#' Chernoff faces.  See \code{\link[aplpack]{faces}} for more details.
#'
#' @param ... other arguments passed on to \code{\link{animate}}
#' @seealso \code{\link{animate}} for options that apply to all animations
#' @keywords hplot
#' @export
#' @examples
#' # The drawing code is fairly slow, so this animation works best with a
#' # limited number of cases
#' flea_s <- rescale(flea[,1:6])
#' animate_faces(flea_s[19:24, 1:6])
#'
#' animate_faces(flea_s[19:24, 1:6], grand_tour(5))
display_faces <- function(...) {
  if (!requireNamespace("aplpack", quietly = TRUE)) {
    stop("Please install the aplpack package", call. = FALSE)
  }

  render_data <- function(data, proj, geodesic) {
    x <- data %*% proj
    x <- (x + 2) / 4
    aplpack::faces(x, scale = TRUE, face.type = 0,
                   cex = 0.1, print.info = FALSE)
  }

  list(
    init = nul,
    render_frame = nul,
    render_transition = nul,
    render_data = render_data,
    render_target = nul
  )
}


#' @rdname display_faces
#' @inheritParams animate
#' @export
animate_faces <- function(data, tour_path = grand_tour(3), ...) {
  animate(data, tour_path, display = display_faces(...), ...)
}

#' A frozen tour path.
#'
#' A frozen tour fixes some of the values of the orthonormal projection
#' matrix and allows the others to vary freely according to any of the
#' other tour methods.  This frozen tour is a frozen grand tour.  See
#' \code{\link{frozen_guided_tour}} for a frozen guided tour.
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param d target dimensionality
#' @param frozen matrix of frozen variables, as described in
#'   \code{\link{freeze}}
#' @export
#' @examples
#' frozen <- matrix(NA, nrow = 4, ncol = 2)
#' frozen[3, ] <- .5
#' animate_xy(flea[, 1:4], frozen_tour(2, frozen))
#'
#' frozen <- matrix(NA, nrow = 4, ncol = 2)
#' frozen[1, 1] <- 0.5
#' animate_xy(flea[, 1:4], frozen_tour(2, frozen))
#'
#' # Doesn't work - a bug?
#' frozen <- matrix(NA, nrow = 4, ncol = 2)
#' frozen[1:2, 1] <- 1 / 4
#' animate_xy(flea[, 1:4], frozen_tour(2, frozen))
#'
#' \dontrun{
#' # This freezes one entire direction which causes a problem,
#' # and is caught by error handling.
#' # If you want to do this it would be best with a dependence
#' # tour, with one variable set one axis, eg 3rd variable to
#' # x axis would be indicated from the code below
#' frozen <- matrix(NA, nrow = 4, ncol = 2)
#' frozen[3, ] <- c(0, 1)
#' animate_xy(flea[, 1:4], frozen_tour(2, frozen))
#' }
#'
#' # Two frozen variables in five.
#' frozen <- matrix(NA, nrow = 5, ncol = 2)
#' frozen[3, ] <- .5
#' frozen[4, ] <- c(-.2, .2)
#' animate_xy(flea[, 1:5], frozen_tour(2, frozen))
frozen_tour <- function(d = 2, frozen) {
  generator <- function(current, data, ...) {
    if (is.null(current)) {
      # Here is a problem if the frozen part is in the
      # initialised basis, because it gets zero'd out by freeze.
      notfrozen <- rep(TRUE, ncol(data))
      for (i in 1:ncol(data))
        if (sum(frozen[i,]^2, na.rm=TRUE)>0)
          notfrozen[i] <- FALSE
      current <- matrix(0, nrow = ncol(data), ncol = d)
      for (i in 1:d) {
        current[c(1:ncol(data))[notfrozen][i],i] <- 1
      }
      # This will still throw an error if number of frozen is bigger than d

      return(current)
    }


    target <- basis_random(ncol(data), d)
    list(target = target)
  }

  check_freezer_safe(frozen)
  new_geodesic_path("frozen", generator, frozen = frozen)
}

#' The frozen guided tour
#'
#' @param frozen matrix of frozen variables, as described in
#'   \code{\link{freeze}}
#' @param index_f the index function to optimise.
#' @param d target dimensionality
#' @param max.tries the maximum number of unsuccessful attempts to find
#'   a better projection before giving up
#' @seealso \code{\link{cmass}}, \code{\link{holes}} and \code{\link{lda_pp}}
#'   for examples of index functions.  The function should take a numeric
#'   matrix and return a single number, preferrably between 0 and 1.
#' @export
#' @examples
#' frozen <- matrix(NA, nrow = 4, ncol = 2)
#' frozen[3, ] <- .5
#' animate_xy(flea[, 1:4], frozen_guided_tour(frozen, holes()))
frozen_guided_tour <- function(frozen, index_f, d = 2, max.tries = 25) {
  generator <- function(current, data, tries, ...) {
    if (is.null(current)) {
      return(basis_init(ncol(data), d))
    }

    index <- function(proj) {
      index_f(as.matrix(data) %*% proj)
    }

    basis <- search_frozen_geodesic(current, index, tries, max.tries, frozen = frozen)
    list(target = basis$target, index = index)
  }

  check_freezer_safe(frozen)
  new_geodesic_path("frozen-guided", generator, frozen = frozen)
}

#' Check matrix is a valid frozen matrix
#'
#' @keywords internal
#' @param frozen matrix to check for freezability
check_freezer_safe <- function(frozen) {
  stopifnot(is.matrix(frozen))

  lengths <- colSums(frozen^2, na.rm = TRUE)
  if (any(lengths >= 1)) {
    stop("Columns of frozen matrix must have squared norm < 1", call. = FALSE)
  }
}

#' Freeze and thaw matrices
#'
#' Some terminology:
#'   * frozen variables: the variables that have fixed values
#'   * warm variables: the remaining variables that vary freely
#'
#' A frozen matrix specifies which variables to fix in a projection matrix.
#' Warm variables should be missing (\code{NA}) while frozen variables should
#' be set to their fixed values.
#'
#' @keywords internal
#' @export
#' @examples
#' frozen <- matrix(NA, nrow = 4, ncol = 2)
#' frozen[3, ] <- .5
#'
#' input <- basis_random(4, 2)
#' freeze(input, frozen)
#' thaw(input, frozen)
#' freeze(basis_random(4, 2), frozen)
freeze <- function(input, frozen) {
  fixed <- !is.na(frozen)
  input[fixed] <- 0
  input
}

#' @export
#' @rdname freeze
thaw <- function(input, frozen) {
  fixed <- !is.na(frozen)

  input <- normalise(input)
  frozen_lengths <- colSums(frozen^2, na.rm = TRUE)

  input <- sweep(input, 2, sqrt(1 - frozen_lengths), "*")
  input[fixed] <- frozen[fixed]

  input
}

#' Animate a tour path.
#'
#' This is the function that powers all of the tour animations.  If you want
#' to write your own tour animation method, the best place to
#' start is by looking at the code for animation methods that have already
#' implemented in the package.
#'
#' See \code{\link{render}} to render animations to disk.
#'
#' @param data matrix, or data frame containing numeric columns
#' @param tour_path tour path generator, defaults to 2d grand tour
#' @param start projection to start at, if not specified, uses default
#'   associated with tour path
#' @param display takes the display that is suppose to be used, defaults to
#'   the xy display
#' @param aps target angular velocity (in radians per second)
#' @param fps target frames per second (defaults to 15, to accommodate RStudio graphics device)
#' @param max_frames the maximum number of bases to generate.  Defaults to
#'   Inf for interactive use (must use Ctrl + C to terminate), and 1 for
#'   non-interactive use.
#' @param rescale Default FALSE. If TRUE, rescale all variables to range [0,1]?
#' @param sphere if true, sphere all variables
#' @param ... ignored
#' @return an (invisible) list of bases visited during this tour
#' @export
#' @examples
#' f <- flea[, 1:6]
#' animate(f, grand_tour(), display_xy())
#' # or in short
#' animate(f)
#' animate(f, max_frames = 30)
#' \donttest{
#' animate(f, max_frames = 10, fps = 1, aps = 0.1)
#' }
animate <- function(data, tour_path = grand_tour(), display = display_xy(),
                    start = NULL, aps = 1, fps = 10, max_frames = Inf,
                    rescale = FALSE, sphere = FALSE, ...) {
  record <-
    tibble::tibble(
      basis = list(),
      index_val = numeric(),
      info = character(),
      method = character(),
      alpha = numeric(),
      tries = numeric(),
      loop = numeric()
    )
  if (!is.matrix(data)) {
    message("Converting input data to the required matrix format.")
    data <- as.matrix(data)
  }
  if (is.null(colnames(data))){
    message("Adding column names.")
    colnames(data) <- paste0("x", 1:ncol(data))
  }

  if (rescale) data <- rescale(data)
  if (sphere) data <- sphere_data(data)

  # By default, only take single step if not interactive
  # This is useful for the automated tests run by R CMD check
  if (!interactive() && missing(max_frames)) {
    max_frames <- 1
  }
  if (max_frames == Inf) {
    to_stop()
  }
  plat <- find_platform()
  if ((rstudio_gd() && fps > 19) || (positron_gd() && fps > 19)) {
    warning("Rstudio graphics device supports maximum fps of 19", call. = FALSE)
    fps <- 19
  }

  tour <- new_tour(data, tour_path, start, ...)
  start <- tour(0, ...)
  bs <- 1
  bases <- array(NA, c(ncol(data), ncol(start$target), bs))

  # Initialise display
  if (!is.null(start$target)){
    display$init(data)
    display$render_frame()
    display$render_data(data, start$proj)
  }

  b <- 0
  i <- 0

  tryCatch(
    {
      while (i < max_frames) {
        i <- i + 1
        step <- tour(aps / fps, ...)
        if (step$step == 1) {
          b <- b + 1
          if (b > bs) {
            bases <- c(bases, rep(NA, bs * dim(bases)[1] * dim(bases)[2]))
            dim(bases) <- c(ncol(data), ncol(start$target), 2 * bs)
            bs <- 2 * bs
          }
          bases[, , b] <- step$target
        } else if (is.null(step$proj)){
          break
        }

        if (!is.null(start$target)){
          dev.hold()
          on.exit(dev.flush())
          if (plat$os == "win" || plat$iface == "rstudio" || plat$iface == "cli") {
            display$render_frame()
          } else {
            display$render_transition()
          }
          display$render_data(data, step$proj, step$target)
          dev.flush()
          if (step$step < 0) break # break after rendering final projection
          Sys.sleep(1 / fps)
        }
      }
    },
    interrupt = function(cond) {
      dev.flush()
      return()
    }
  )

  if (b != 0) {
    invisible(bases[, , seq_len(b)])
  }

  invisible(record)

  # Need a better way to clean up global variables than this
  # suppressWarnings(rm(tries, cur_index, current, t0, record, envir = globalenv()))
}

rstudio_gd <- function() identical(names(dev.cur()), "RStudioGD")
positron_gd <- function() identical(names(dev.cur()), "Positron Graphics Device")

# globalVariables("record")

#' Generate nearby bases, e.g. for simulated annealing.
#' @keywords internal
basis_nearby <- function(current, alpha = 0.5, method = "linear") {
  method <- match.arg(method, c("linear", "geodesic"))
  new <- basis_random(nrow(current), ncol(current))

  switch(method,
    linear = orthonormalise((1 - alpha) * current + alpha * new),
    geodesic = step_fraction(geodesic_info(current, new), alpha)
  )
}

#' check if the current and target bases are of the same orientation
#' @keywords internal
correct_orientation <- function(current, target){

  for (i in ncol(current)){
    if (det(t(current[,i]) %*% target[,i]) < 0){
      target[,i] <- -target[,i]
    }
  }

  return(target)
}


#' Search for a better projection near the current projection.
#'
#' @param current starting projection
#' @param alpha the angle used to search the target basis from the current basis
#' @param index index function
#' @param tries the counter of the outer loop of the opotimiser
#' @param max.tries maximum number of iteration before giving up
#' @param ... other arguments being passed into the \code{search_better()}
#' @param method whether the nearby bases are found by a linear/ geodesic formulation
#' @param cur_index the index value of the current basis
#' @keywords optimize
#' @importFrom utils tail globalVariables
#' @export
#' @examples
#' animate_xy(flea[, 1:6], guided_tour(holes(), search_f = search_better))
search_better <- function(current, alpha = 0.5, index, tries, max.tries = Inf,...,
                          method = "linear", cur_index = NA) {

  if (is.na(cur_index)) cur_index <- index(current)

  if (cur_index == 0) {
    warning("cur_index is zero!")
  }

  try <- 1

  while (try < max.tries) {
    new_basis <- basis_nearby(current, alpha, method)
    new_index <- index(new_basis)

    rcd_env <- parent.frame(n = 4)
    if (is.null(rcd_env[["record"]])) rcd_env <- parent.frame(n = 1)
    rcd_env[["record"]] <- dplyr::add_row(
      rcd_env[["record"]],
      basis = list(new_basis),
      index_val = new_index,
      info = "random_search",
      tries = tries,
      loop = try,
      method = "search_better",
      alpha = round(alpha, 4)
    )


    if (new_index > cur_index) {
      message("Target: ", sprintf("%.3f", new_index), ", try: ", try)

      nr <- nrow(rcd_env[["record"]])
      rcd_env[["record"]][nr, "info"] <- "new_basis"

      new_basis <- correct_orientation(current, new_basis)
      rcd_env[["record"]][[nr, "basis"]] <- list(new_basis)

      return(list(target = new_basis, alpha = alpha))
    }
    try <- try + 1
  }

  message("No better bases found after ", max.tries, " tries.  Giving up.")
  print_final_proj(current)

  rcd_env[["record"]] <- dplyr::mutate(
    rcd_env[["record"]],
    id = dplyr::row_number()
  )
  NULL
}

#' Search for a better projection using simulated annealing
#'
#' Given an initial \eqn{t0}, the cooling scheme updates temperature at \deqn{T = t0 /\log(i + 1)}
#' The candidate basis is sampled via \deqn{B_j = (1 - \alpha) * B_i + \alpha * B} where alpha defines the neighbourhood, \eqn{B_i} is the current basis, B is a randomly generated basis
#' The acceptance probability is calculated as \deqn{prob = \exp{-abs(I(B_i) - I(B_j))/ T}}
#' For more information, see
#' \url{https://projecteuclid.org/download/pdf_1/euclid.ss/1177011077}
#'
#' @param current starting projection
#' @param alpha the angle used to search the target basis from the current basis
#' @param index index function
#' @param tries the counter of the outer loop of the opotimiser
#' @param max.tries maximum number of iteration before giving up
#' @param method whether the nearby bases are found by a linear/ geodesic formulation
#' @param cur_index the index value of the current basis
#' @param t0 initial decrease in temperature
#' @param ... other arguments being passed into the \code{search_better_random()}
#'
#' @keywords optimize
#' @export
#' @examples
#' animate_xy(flea[, 1:6], guided_tour(holes(), search_f = search_better_random))
search_better_random <- function(current, alpha = 0.5, index, tries,
                                 max.tries = Inf, method = "linear", cur_index = NA, t0 = 0.01,
                                 ...) {
  if (is.na(cur_index)) cur_index <- index(current)

  if (cur_index == 0) {
    warning("cur_index is zero!")
  }
  try <- 1
  while (try < max.tries) {
    new_basis <- basis_nearby(current, alpha, method)
    new_index <- index(new_basis)
    temperature <- t0 / log(try + 1)

    rcd_env <- parent.frame(n = 4)
    rcd_env[["record"]] <- dplyr::add_row(
      rcd_env[["record"]],
      basis = list(new_basis),
      index_val = new_index,
      info = "random_search",
      tries = tries,
      loop = try,
      method = "search_better_random",
      alpha = round(alpha, 4)
    )

    if (new_index > cur_index) {
      message("Target: ", sprintf("%.3f", new_index), ", try: ", try, ", accept")

      nr <- nrow(rcd_env[["record"]])
      rcd_env[["record"]][nr, "info"] <- "new_basis"

      new_basis <- correct_orientation(current, new_basis)
      rcd_env[["record"]][[nr, "basis"]] <- list(new_basis)

      return(list(target = new_basis, alpha = alpha))
    }
    else {
      prob <- min(exp(-abs(cur_index - new_index) / temperature), 1)
      rand <- stats::runif(1)

      if (prob > rand) {
        message("Target: ", sprintf("%.3f", new_index), ", try: ", try,
                ", probabilistic accept p = ", sprintf("%.3f", prob))

        nr <- nrow(rcd_env[["record"]])
        rcd_env[["record"]][nr, "info"] <- "new_basis"

        rcd_env[["record"]] <- dplyr::mutate(
          rcd_env[["record"]],
          id = dplyr::row_number()
        )

        return(list(target = new_basis, alpha = alpha))
      }
    }
    try <- try + 1
  }

  message("No better bases found after ", max.tries, " tries.  Giving up.")
  print_final_proj(current)

  rcd_env[["record"]] <- dplyr::mutate(
    rcd_env[["record"]],
    id = dplyr::row_number()
  )
  NULL
}

# globalVariables(c("t0","tries", "info", "runif"))

#' A radial tour path.
#'
#' The radial tour rotates a chosen variable axis out of the current
#' projection.
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param start initial projection matrix
#' @param mvar variable(s) chosen to rotate out
#' @param ... additional arguments for drawing
#' @export
#' @examples
#' animate_xy(flea[, 1:6], radial_tour(basis_random(6, 2), mvar = 4), rescale=TRUE)
#' animate_xy(flea[, 1:6], radial_tour(basis_random(6, 2), mvar = c(3,4)), rescale=TRUE)
#' animate_dist(flea[, 1:6], radial_tour(basis_random(6, 1), mvar = 4), rescale=TRUE)
#' animate_scatmat(flea[, 1:6], radial_tour(basis_random(6, 3), mvar = 4), rescale=TRUE)
radial_tour <- function(start, mvar = 1, ...) {
  first <- TRUE
  out <- TRUE

  generator <- function(current, data, ...) {
    if (first) {
      new_basis <- start
      first <<- FALSE
      return(new_basis)
    }
    if (out) {
      new <- start
      new[mvar,] <- rep(0, ncol(start))
      new <- orthonormalise(new)
      new_basis <- new
      out <<- !out
    }
    else {
      new_basis <- start
      out <<- !out
    }

    target <- new_basis
    list(target = target)
  }

  new_geodesic_path("radial", generator)
}

#' A dependence tour path.
#'
#' The dependence tour combines a set of independent 1d tours to produce
#' a nd tour.  For the special case of 2d, this is known as a correlation
#' tour.  This tour corresponds to the multivariate method known as
#' generalised canonical correlation, and is used to investigate dependence
#' between groups of variables.
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param pos a numeric vector describing which variables are mapped to
#'   which dimensions: 1 corresponds to first, 2 to second etc.
#' @export
#' @examples
#' animate_xy(flea[, 1:3], dependence_tour(c(1, 2, 2)))
#' animate_xy(flea[, 1:4], dependence_tour(c(1, 2, 1, 2)))
#' animate_pcp(flea[, 1:6], dependence_tour(c(1, 2, 3, 2, 1, 3)))
dependence_tour <- function(pos) {
  stopifnot(is.numeric(pos))
  stopifnot(all.equal(pos, trunc(pos)))

  d <- max(pos)
  generator <- function(current, data, ...) {
    if (is.null(current)) {
      return(basis_init(ncol(data), d))
    }

    mat <- matrix(0, ncol = d, nrow = length(pos))
    for (i in seq_len(d)) {
      mat[pos == i, i] <- basis_random(sum(pos == i), 1)
    }

    target <- mat
    list(target = target)
  }

  new_geodesic_path("independent", generator)
}

#' Anomaly index.
#'
#' Calculates an index that looks for the best projection of
#' observations that are outside a pre-determined p-D ellipse.
#'
#' @export
anomaly_index <- function() {

  function(mat, ell2d, ellmu2d) {

    mat_tab <- #mean(mahal_dist(mat, ell2d))
      mean(mahalanobis(mat, center=ellmu2d, cov=ell2d))
  }
}

#' Search for a better projection based on Poss, 1995
#' @param current starting projection
#' @param alpha the angle used to search the target basis from the current basis
#' @param index index function
#' @param tries the counter of the outer loop of the opotimiser
#' @param max.tries maximum number of iteration before giving up
#' @param cur_index the index value of the current basis
#' @param ... other arguments being passed into the \code{search_better()}
#' @keywords optimize
#' @export
search_posse <- function(current, alpha = 0.5, index, tries, max.tries = 300, cur_index = NA, ...) {
  if (is.na(cur_index)) cur_index <- index(current)

  try <- 1
  h <- 0

  while (try < max.tries) {
    new_basis <- orthonormalise(current + alpha * basis_random(nrow(current), ncol(current)))
    new_index <- index(new_basis)

    rcd_env <- parent.frame(n = 4)
    rcd_env[["record"]] <- dplyr::add_row(
      rcd_env[["record"]],
      basis = list(new_basis),
      index_val = new_index,
      info = "random_search",
      tries = tries,
      loop = try,
      method = "search_posse",
      alpha = round(alpha, 4)
    )

    if (new_index > cur_index) {
      message("Target: ", sprintf("%.3f", new_index), ", try: ", try)

      nr <- nrow(rcd_env[["record"]])
      rcd_env[["record"]][nr, "info"] <- "new_basis"

      return(list(target = new_basis, h = h))
    } else {
      h <- h + 1
    }

    try <- try + 1
  }

  message("No better bases found after ", max.tries, " tries.  Giving up.")
  print_final_proj(current)

  NULL
}

#' Manually slice along a variable axis.
#'
#' The manual slice tour takes the current projection, with display_slice,
#' and changes the slice center.
#'
#' @param data numeric matrix, with n rows and p columns
#' @param proj projection from which slices are constructed
#' @param var variable axis to run the center along: 1, ..., p
#' @param nsteps number of changes in center to make
#' @param v_rel relative volume of the slice. If not set, suggested value
#'   is calculated and printed to the screen.
#' @param rescale Default FALSE. If TRUE, rescale all variables to range [0,1]?
#' @param col color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".
#' @param sphere if true, sphere all variables
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param anchor_nav position of the anchor: center, topright or off
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @param ... other options passed to output device
#' @export
#' @examples
#' # Note that you might need to use the quartz()
#' # on OSX to see the animation
#' sphere5 <- data.frame(geozoo::sphere.hollow(5)$points)
#' proj <- basis_random(5, 2)
#' manual_slice(sphere5, proj, var=3, nsteps=10, rescale=TRUE, half_range=1.5)
manual_slice <- function(data, proj, var=1, nsteps=20,
                         v_rel = 0.01, rescale = FALSE,
                         sphere = FALSE, col = "black",
                         half_range = NULL,
                         anchor_nav = "topright",
                         palette = "Zissou 1", ...) {
  # Standard checks, and scaling
  if (!is.matrix(data)) {
    message("Converting input data to the required matrix format.")
    data <- as.matrix(data)
  }
  if (rescale) data <- rescale(data)
  if (sphere) data <- sphere_data(data)

  # If colors are a variable, convert to colors
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    col <- mapColors(col, palette)
  }

  # Going to use nsteps twice, to go out to edge, come back to 0
  # go out to negative edge and back in to stop at 0
  rng_var <- range(data[,var])
  anchor <- matrix(colMeans(data), ncol = length(proj[, 1]))
  anchor_steps <- c(anchor[,var] + 0.8*(rng_var[2]-anchor[,var])*((0:nsteps)/nsteps),
                    anchor[,var] + 0.8*(rng_var[2]-anchor[,var])*((nsteps:0)/nsteps),
                    anchor[,var] - 0.8*(anchor[,var]-rng_var[1])*((0:nsteps)/nsteps),
                    anchor[,var] - 0.8*(anchor[,var]-rng_var[1])*((nsteps:0)/nsteps))

  # Initialise tour
  tour <- new_tour(data, grand_tour(), proj, ...)

  # Always doing slices
  display <- display_slice(axes="bottomleft", v_rel = v_rel,
                           half_range = half_range,
                           anchor_nav = anchor_nav)

  # Initialise display
  display$init(data)
  display$render_frame()
  display$render_data(data, proj, proj, with_anchor=anchor)

  plat <- find_platform()

  for (j in anchor_steps) {
    anchor[,var] <- j
    dev.hold()
    on.exit(dev.flush())
    if (plat$os == "win" || plat$iface == "rstudio") {
      display$render_frame()
    } else {
      display$render_transition()
    }
    display$render_data(data, proj, proj, with_anchor = anchor)
    dev.flush()
    Sys.sleep(1 / 25)
  }
  invisible()
}

#' Display tour path with a density and scatterplot
#'
#' Animate a 2D tour path with density contour(s) and a scatterplot.
#'
#' @param axes position of the axes: center, bottomleft or off
#' @param center if TRUE, centers projected data to (0,0).  This pins the
#'  center of data cloud and make it easier to focus on the changing shape
#'  rather than position.
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param edges A two column integer matrix giving indices of ends of lines.
#' @param col color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".
#' @param pch shape of the point to be plotted.  Defaults to 20.
#' @param cex size of the point to be plotted.  Defaults to 1.
#' @param contour_quartile Vector of quartiles to plot the contours at. Defaults to 5.
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @param axislablong text labels only for the long axes in a projection, default FALSE
#' @param ...  other arguments passed on to \code{\link{animate}} and
#'   \code{\link{display_density2d}}
#' @importFrom graphics contour
#' @importFrom stats quantile
#' @export
#' @examples
#' animate_density2d(flea[, 1:6])
#' animate(flea[, 1:6], tour_path = grand_tour(), display = display_density2d())
#' animate(flea[, 1:6],
#'   tour_path = grand_tour(),
#'   display = display_density2d(axes = "bottomleft")
#' )
#' animate(flea[, 1:6],
#'   tour_path = grand_tour(),
#'   display = display_density2d(half_range = 0.5)
#' )
#' animate_density2d(flea[, 1:6], tour_path = little_tour())
#'
#' animate_density2d(flea[, 1:3], tour_path = guided_tour(holes()), sphere = TRUE)
#' animate_density2d(flea[, 1:6], center = FALSE)
#'
#' # The default axes are centered, like a biplot, but there are other options
#' animate_density2d(flea[, 1:6], axes = "bottomleft")
#' animate_density2d(flea[, 1:6], axes = "off")
#' animate_density2d(flea[, 1:6], dependence_tour(c(1, 2, 1, 2, 1, 2)),
#'   axes = "bottomleft"
#' )
#'
#' animate_density2d(flea[, -7], col = flea$species)
#'
#' # You can also draw lines
#' edges <- matrix(c(1:5, 2:6), ncol = 2)
#' animate(
#'   flea[, 1:6], grand_tour(),
#'   display_density2d(axes = "bottomleft", edges = edges)
#' )
display_density2d <- function(center = TRUE, axes = "center", half_range = NULL,
                              col = "black", pch = 20, cex = 1,
                              contour_quartile = c(.25, .5, .75), edges = NULL,
                              palette = "Zissou 1",
                              axislablong = FALSE, ...) {
  # If colors are a variable, convert to colors
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    col <- mapColors(col, palette)
  }

  labels <- NULL
  init <- function(data) {
    half_range <<- compute_half_range(half_range, data, center)
    labels <<- abbreviate(colnames(data), 3)
  }

  if (!is.null(edges)) {
    if (!is.matrix(edges) && ncol(edges) == 2) {
      stop("Edges matrix needs two columns, from and to, only.")
    }
  }

  render_frame <- function() {
    par(pty = "s", mar = rep(0.1, 4))
    blank_plot(xlim = c(-1, 1), ylim = c(-1, 1))
  }
  render_transition <- function() {
    rect(-1, -1, 1, 1, col = "#FFFFFFE6", border = NA)
  }
  render_data <- function(data, proj, geodesic) {
    draw_tour_axes(proj, labels, limits = 1, axes, longlabels=axislablong)

    # Render projected points
    x <- data %*% proj
    if (center) x <- center(x)
    x <- x / half_range

    colrs <- unique(col)
    ngps <- length(colrs)
    if (ngps == 1) {
      xd <- MASS::kde2d(x[, 1], x[, 2])

      contour(xd,
        col = col,
        levels = quantile(xd$z, probs = contour_quartile),
        axes = FALSE
      )
    }
    else {
      for (i in 1:ngps) {
        x.sub <- x[col == colrs[i], ]
        xd <- MASS::kde2d(x.sub[, 1], x.sub[, 2])

        contour(xd,
          col = colrs[i],
          levels = quantile(xd$z, probs = contour_quartile),
          axes = FALSE, add = TRUE
        )
      }
    }
    points(x, col = col, pch = pch, cex = cex)

    if (!is.null(edges)) {
      segments(
        x[edges[, 1], 1], x[edges[, 1], 2],
        x[edges[, 2], 1], x[edges[, 2], 2]
      )
    }
  }

  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}

#' @rdname display_density2d
#' @inheritParams animate
#' @export
animate_density2d <- function(data, tour_path = grand_tour(), ...) {
  animate(data, tour_path, display_density2d(...))
}

#' Display 3d projection with depth cues
#'
#' Suggestion to use gray background and colour saturation (instead of
#' gray shading) by Graham Wills.
#
#' @param center should projected data be centered to have mean zero (default:
#'   TRUE). This pins the centre of the data to the same place, and makes it
#'   easier to focus on the shape.
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param ... other arguments passed on to \code{\link{animate}}
#' @seealso \code{\link{animate}} for options that apply to all animations
#' @keywords hplot
#' @export
#' @examples
#' animate_depth(flea[, 1:6])
display_depth <- function(center = TRUE, half_range = NULL, ...) {
  shades <- hcl(240,
    c = seq(0, 60, length = 100),
    l = seq(80, 20, length = 100)
  )

  init <- function(data) {
    half_range <<- compute_half_range(half_range, data, center)
  }

  render_frame <- function() {
    par(pty = "s", mar = rep(1, 4))
    blank_plot(xlim = c(-1, 1), ylim = c(-1, 1))
    render_transition()
  }
  render_transition <- function() {
    rect(-1, -1, 1, 1, col = "grey80", border = NA)
  }
  render_data <- function(data, proj, geodesic) {
    x <- data %*% proj
    if (center) x <- center(x)
    x <- x / half_range

    depth <- x[, 3]
    # depth ranges between -1 and 1,
    # so depth_std should lie between 0 and 1
    depth_std <- depth / 2 + 0.5
    size <- 0.5 + depth_std * 3
    shade <- shades[round(depth_std * 100)]

    ord <- order(depth_std)
    points(x[ord, 1:2], pch = 20, cex = size[ord], col = shade[ord], ...)
  }

  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}


#' @rdname display_depth
#' @inheritParams animate
#' @export
animate_depth <- function(data, tour_path = grand_tour(3), ...) {
  animate(
    data = data, tour_path = tour_path,
    display = display_depth(...),
    ...
  )
}

#' A guided section tour path.
#'
#' The guided section tour is a variation of the guided tour that is
#' using a section pursuit index for the selection of target planes.
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate_slice}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param index_f the section pursuit index function to optimise. The function
#'   needs to take three arguments, the projected data, the vector of distances
#'   from the current projection plane, and the slice thickness h.
#' @param d target dimensionality
#' @param alpha the initial size of the search window, in radians
#' @param cooling the amount the size of the search window should be adjusted
#'   by after each step
#' @param search_f the search strategy to use
#' @param max.tries the maximum number of unsuccessful attempts to find
#'   a better projection before giving up
#' @param max.i the maximum index value, stop search if a larger value is found
#' @param v_rel relative volume of the slice. If not set, suggested value
#'   is calculated and printed to the screen.
#' @param anchor A vector specifying the reference point to anchor the slice.
#'   If NULL (default) the slice will be anchored at the data center.
#' @param ... arguments sent to the search_f
#' @seealso \code{\link{slice_index}} for an example of an index functions.
#' \code{\link{search_geodesic}}, \code{\link{search_better}},
#'   \code{\link{search_better_random}} for different search strategies
#' @export
#' @examples
#' # Generate samples on a 3d hollow sphere using the geozoo package
#' set.seed(12345)
#' sphere3 <- geozoo::sphere.hollow(3)$points
#' # Columns need to be named before launching the tour
#' colnames(sphere3) <- c("x1", "x2", "x3")
#' # Off-center anchoring
#' anchor3 <- matrix(rep(0.75, 3), ncol=3)
#' # Index setup
#' r_breaks <- linear_breaks(5, 0, 1)
#' a_breaks <- angular_breaks(10)
#' eps <- estimate_eps(nrow(sphere3), ncol(sphere3), 0.1 / 1, 5 * 10, 10, r_breaks)
#' idx <- slice_index(r_breaks, a_breaks, eps, bintype = "polar", power = 1, reweight = TRUE, p = 3)
#' # Running the guided section tour select sections showing a big hole in the center
#' animate_slice(sphere3, guided_section_tour(idx, v_rel = 0.1, anchor = anchor3, max.tries = 5),
#'   v_rel = 0.1, anchor = anchor3
#' )
guided_section_tour <- function(index_f, d = 2, alpha = 0.5, cooling = 0.99,
                                max.tries = 25, max.i = Inf, v_rel = NULL, anchor = NULL,
                                search_f = search_geodesic, ...) {
  h <- NULL

  generator <- function(current, data, tries, ...) {
    if (is.null(current)) {
      return(basis_init(ncol(data), d))
    }

    if (is.null(h)) {
      half_range <- compute_half_range(NULL, data, FALSE)
      v_rel <- compute_v_rel(v_rel, half_range, ncol(data))
      h <<- v_rel^(1 / (ncol(data) - 2))
    }

    index <- function(proj) {
      dist_data <- anchored_orthogonal_distance(proj, data, anchor)
      index_f(as.matrix(data) %*% proj, dist_data, h)
    }

    cur_index <- index(current)

    if (cur_index > max.i) {
      message("Found index ", sprintf("%.3f", cur_index), ",
              larger than selected maximum ", max.i, ". Stopping search.")
      print_final_proj(current)
      return(NULL)
    }

    basis <- search_f(current, alpha, index, tries, max.tries, cur_index = cur_index, ...)
    alpha <<- alpha * cooling

    list(target = basis$target, index = index)
  }

  new_geodesic_path("guided", generator)
}

#' 1d distribution tour path animation.
#'
#' Animate a 1d tour path with a density plot or histogram.
#'
#' @param method display method, histogram or density plot
#' @param center should 1d projection be centered to have mean zero (default: TRUE).
#'   This pins the centre of distribution to the same place, and makes it
#'   easier to focus on the shape of the distribution.
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param col color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".
#' @param rug draw rug plot showing position of actual data points?
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @param density_max allow control of the y range for density plot
#' @param bw binwidth for histogram and density, between 0-1, default 0.2
#' @param scale_density Height of density is scaled at each projection, default FALSE
#' @param ... other arguments passed on to \code{\link{animate}}
#' @seealso \code{\link{animate}} for options that apply to all animations
#' @keywords hplot
#' @export
#' @examples
#' animate_dist(flea[, 1:6])
#'
#' # Change inputs, to color by group, fix y axis, change bin width
#' # and scale bar height or density at each projection
#' animate_dist(flea[, 1:6], col=flea$species, density_max=5)
#' animate_dist(flea[, 1:6], col=flea$species, density_max=5, bw=0.1)
#' animate_dist(flea[, 1:6], col=flea$species, scale_density=TRUE)
#'
#' # When the distribution is not centred, it tends to wander around in a
#' # distracting manner
#' animate_dist(flea[, 1:6], center = FALSE)
#'
#' # Alternatively, you can display the distribution with a histogram
#' animate_dist(flea[, 1:6], method = "hist")
display_dist <- function(method = "density", center = TRUE, half_range = NULL,
                         col = "black", rug = FALSE,
                         palette = "Zissou 1",
                         density_max = 3,
                         bw = 0.2,
                         scale_density = FALSE, ...) {
  if (!requireNamespace("ash", quietly = TRUE)) {
    stop("Please install the ash package", call. = FALSE)
  }

  method <- match.arg(method, c("histogram", "density", "ash"))
  labels <- NULL
  init <- function(data) {
    half_range <<- compute_half_range(half_range, data, center)
    labels <<- abbreviate(colnames(data), 2)
  }

  # If colors are a variable, convert to colors
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    lgps <- levels(gps)
    col <- mapColors(col, palette)
  }
  colrs <- unique(col)
  ngps <- length(colrs)

  render_frame <- function() {
    if (ngps > 5) {
      stop("Please choose a group with 5 or less levels.\n")
    }

    if (ngps == 1)
      par(pty = "m", mar = c(4, 4, 1, 1), mfrow = c(2, 1))
    else {
      #if (method != "density")
        par(pty = "m", mar = c(4, 4, 1, 1), mfrow = c(ngps+1, 1))
      #else
      #  par(pty = "m", mar = c(4, 4, 1, 1))
    }

  }
  render_transition <- function() {
    rect(-1, -1.1, 1.2, density_max, col = "#FFFFFFE6", border = NA)
  }
  render_data <- function(data, proj, geodesic) {

    x <- data %*% proj
    if (center) x <- center(x)
    x <- x / half_range

    # Render projection data
    if (method == "histogram") {
      if (ngps == 1) {
        par(mar = c(0, 4, 0, 1))
        plot(
          x = NA, y = NA, xlim = c(-1, 1.2), ylim = c(0, density_max),
          xlab="", ylab = "Density",
          xaxt = "n", yaxt = "n"
        )
        axis(2, seq(0, density_max, by = 1))
        abline(h = seq(0, density_max, by = 0.5), col = "grey80")

        bins <- hist(x, breaks = seq(-1, 1, bw), plot = FALSE)
        if (scale_density)
          bins$density <- bins$density/max(bins$density) * density_max * 0.8
        with(bins, rect(mids - bw/2, 0, mids + bw/2, density,
          col = col
        ))
        if (rug) {
          segments(x, 0, x, 0.1, ...)
        }

      }
      else {

        for (i in 1:ngps) {
          par(mar = c(0, 4, 0, 1))
          plot(
            x = NA, y = NA, xlim = c(-1, 1.2), ylim = c(0, density_max),
            xlab="", ylab = "Density",
            xaxt = "n", yaxt = "n"
          )
          axis(2, seq(0, density_max, by = 1))
          abline(h = seq(0, density_max, by = 0.5), col = "grey80")

          x.sub <- x[col == colrs[i], ]
          bins <- hist(x.sub, breaks = seq(-1, 1, bw), plot = FALSE)
          if (scale_density)
            bins$density <- bins$density/max(bins$density) * density_max * 0.8
          with(bins, rect(mids - bw/2, 0, mids + bw/2, density,
                          col = colrs[i]
          ))
          text(x=1.0, y=density_max*0.9, labels=lgps[i])
          if (rug) {
            segments(x, 0, x, 0.1, ...)
          }

        }
      }
    } else if (method == "density") {
      if (ngps == 1) {
        par(mar = c(0, 4, 0, 1))
        plot(
          x = NA, y = NA, xlim = c(-1, 1.2), ylim = c(0, density_max),
          xlab="", ylab = "Density",
          xaxt = "n", yaxt = "n"
        )
        axis(2, seq(0, density_max, by = 1))
        abline(h = seq(0, density_max, by = 0.5), col = "grey80")

        if (scale_density) {
          dn <- stats::density(x, bw=bw/2)
          dn$y <- dn$y/max(dn$y) * density_max * 0.8
          polygon(dn, col = col)
        }
        else
          polygon(stats::density(x, bw=bw/2), col = col)
        if (rug) {
          segments(x, 0, x, 0.1, ...)
        }
      }
      else {
        for (i in 1:ngps) {
          par(mar = c(0, 4, 0, 1))
          plot(
            x = NA, y = NA, xlim = c(-1, 1.2), ylim = c(0, density_max),
            xlab="", ylab = "Density",
            xaxt = "n", yaxt = "n"
          )
          axis(2, seq(0, density_max, by = 1))
          abline(h = seq(0, density_max, by = 0.5), col = "grey80")

          x.sub <- x[col == colrs[i], ]
          if (scale_density) {
            dn <- stats::density(x.sub, bw=bw/2)
            dn$y <- dn$y/max(dn$y) * density_max * 0.8
            polygon(dn, col = colrs[i])
          }
          else
            polygon(stats::density(x.sub, bw=bw/2), col = colrs[i])
          text(x=1.0, y=density_max*0.9, labels=lgps[i])
          if (rug) {
            segments(x, 0, x, 0.1, col = colrs[i], ...)
          }
        }
      }
    } else if (method == "ash") {
      if (ngps == 1) {
        par(mar = c(0, 4, 0, 1))
        plot(
          x = NA, y = NA, xlim = c(-1, 1.2), ylim = c(0, density_max),
          xlab="", ylab = "Density",
          xaxt = "n", yaxt = "n"
        )
        axis(2, seq(0, density_max, by = 1))
        abline(h = seq(0, density_max, by = 0.5), col = "grey80")

        if (scale_density) {
          utils::capture.output(ash <- ash::ash1(ash::bin1(x, c(-half_range, half_range))))
          ash$y <- ash$y/max(ash$y) * density_max * 0.8
        }
        else
          utils::capture.output(ash <- ash::ash1(ash::bin1(x, c(-half_range, half_range))))
        #lines(ash, col="black")
        ash$x <- c(min(ash$x), ash$x, max(ash$x))
        ash$y <- c(0, ash$y, 0)
        polygon(ash, col=col)
        if (rug) {
          segments(x, 0, x, 0.1, ...)
        }
      }
      else {
        for (i in 1:ngps) {
          par(mar = c(0, 4, 0, 1))
          plot(
            x = NA, y = NA, xlim = c(-1, 1.2), ylim = c(0, density_max),
            yaxs = "i",
            xlab="", ylab = "Density",
            xaxt = "n", yaxt = "n"
          )
          axis(2, seq(0, 4, by = 1))
          abline(h = seq(0, density_max, by = 0.5), col = "grey80")
          x.sub <- x[col == colrs[i], ]

          if (scale_density) {
            utils::capture.output(ash <- ash::ash1(ash::bin1(x.sub, c(-half_range, half_range))))
            ash$y <- ash$y/max(ash$y) * density_max * 0.8
          }
          else
            utils::capture.output(ash <- ash::ash1(ash::bin1(x.sub, c(-half_range, half_range))))
          # lines(ash, col = colrs[i])
          ash$x <- c(min(ash$x), ash$x, max(ash$x))
          ash$y <- c(0, ash$y, 0)
          polygon(ash, col = colrs[i])
          text(x=1.0, y=density_max*0.9, labels=lgps[i])
          if (rug) {
            segments(x, 0, x, 0.1, col = colrs[i], ...)
          }

        }
      }
    }
    par(mar = c(4, 4, 0, 1))
    plot(
      x = NA, y = NA, xlim = c(-1, 1.2), ylim = c(-1.1, 0),
      xaxs = "i", yaxs = "i",
      xlab = "", ylab = "Projection",
      yaxt = "n"
    )
    lines(c(0, 0), c(-1, 0), col = "grey80")
    lines(c(-1, -1), c(-1.1, 0), col = "grey80")
    lines(c(1, 1), c(-1.1, 0), col = "grey80")
    abline(h = 0)
    box(col = "grey70")

    # Render tour axes
    ax <- seq_along(proj) / length(proj)
    segments(0, -ax, proj, -ax, col = "black", lwd = 3)
    text(1.0, -ax, labels, pos = 4)
  }

  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}


#' @rdname display_dist
#' @inheritParams animate
#' @export
animate_dist <- function(data, tour_path = grand_tour(1), ...) {
  animate(
    data = data, tour_path = tour_path,
    display = display_dist(...), ...
  )
}

#' A guided tour path.
#'
#' Instead of choosing new projections at random like the grand tour, the
#' guided tour always tries to find a projection that is more interesting
#' than the current projection.
#'
#' Currently the index functions only work in 2d.
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param index_f the index function to optimise.
#' @param d target dimensionality
#' @param alpha the initial size of the search window, in radians
#' @param cooling the amount the size of the search window should be adjusted
#'   by after each step
#' @param search_f the search strategy to use: \code{\link{search_geodesic}}, \code{\link{search_better}},
#'   \code{\link{search_better_random}}, \code{\link{search_polish}}. Default is \code{\link{search_geodesic}}.
#' @param max.tries the maximum number of unsuccessful attempts to find
#'   a better projection before giving up
#' @param max.i the maximum index value, stop search if a larger value is found
#' @param n_sample number of samples to generate if \code{search_f} is \code{\link{search_polish}}
#' @param n_jellies only used for \code{search_jellyfish}, the number of jellyfish to use
#' @param ... arguments sent to the search_f
#' @seealso \code{\link{cmass}}, \code{\link{holes}} and \code{\link{lda_pp}}
#'   for examples of index functions.  The function should take a numeric
#'   matrix and return a single number, preferably between 0 and 1.
#' \code{\link{search_geodesic}}, \code{\link{search_better}},
#'   \code{\link{search_better_random}} for different search strategies
#' @export
#' @examples
#' flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
#' animate_xy(flea_std, guided_tour(holes()), sphere = TRUE)
#' \donttest{
#' animate_xy(flea_std, guided_tour(holes(), search_f = search_better_random), sphere = TRUE)
#' animate_dist(flea_std, guided_tour(holes(), 1), sphere = TRUE)
#' animate_xy(flea_std, guided_tour(lda_pp(flea$species)), sphere = TRUE, col = flea$species)
#'
#' # save_history is particularly useful in conjunction with the
#' # guided tour as it allows us to look at the tour path in many different
#' # ways
#' f <- flea_std[, 1:3]
#' tries <- replicate(5, save_history(f, guided_tour(holes())), simplify = FALSE)
#' }
guided_tour <- function(index_f, d = 2, cooling = 0.99, max.tries = 25,
                        max.i = Inf, search_f = search_geodesic,
                        n_jellies = 30, n_sample = 100, alpha = 0.5,...) {
  generator <- function(current, data, tries, ...) {
    index <- function(proj) {
      index_f(as.matrix(data) %*% proj)
    }

    valid_fun <- c(
      "search_geodesic", "search_better", "search_better_random",
      "search_polish", "search_posse", "search_jellyfish"
    )
    method <- valid_fun[vapply(valid_fun, function(x) {
      identical(get(x), search_f)
    }, logical(1))]

    if (is.null(current)) {
      if (method == "search_jellyfish"){
        current <- replicate(n_jellies, basis_random(ncol(data), d), simplify = FALSE)
        cur_index <- sapply(current, index)
        best_id <- which.max(cur_index)
        attr(current, "best_id") <- best_id
        class(current) <- c("multi-bases", class(current))

        rcd_env <- parent.frame(n = 3)
        rcd_env[["record"]] <- dplyr::add_row(
          rcd_env[["record"]], basis = current[-best_id],
          index_val = cur_index[-best_id], info = "initiation",
          method = method, alpha = NA, tries = 1, loop = 1:(length(current)-1)
        )
        rcd_env[["record"]] <- dplyr::add_row(
          rcd_env[["record"]], basis = current[best_id],
          index_val = cur_index[best_id], info = "initiation",
          method = method, alpha = NA, tries = 1, loop = length(current)
        )

        return(current)
      }
      current <- basis_random(ncol(data), d)
      cur_index <- index(current)

      tryCatch({
        rcd_env <- parent.frame(n = 3)
        rcd_env[["record"]] <- dplyr::add_row(
          rcd_env[["record"]], basis = list(current),
          index_val = cur_index, info = "new_basis", method = method,
          alpha = formals(guided_tour)$alpha, tries = 1, loop = 1
        )
      },
      error = function(e){
        assign("record",
               tibble::tibble(
                 basis = list(), index_val = numeric(), info = character(),
                 method = character(), alpha = numeric(), tries = numeric(),
                 loop = numeric()),
               envir = parent.frame())
        rcd_env[["record"]] <- tibble::tibble(
          basis = list(current), index_val = cur_index, info = "new_basis",
          method = method, alpha = formals(guided_tour)$alpha, tries = 1,
          loop = 1
        )
      })

      return(current)
    }

    if (inherits(current, "multi-bases")){
      best_id <- attr(current, "best_id")
      cur_index <- max(sapply(current, index))
    } else{
      cur_index <- index(current)
    }

    if (cur_index > max.i) {
      message("Found index ", sprintf("%.3f", cur_index), ",
              larger than selected maximum ", max.i, ". Stopping search.")
      print_final_proj(current)
      return(NULL)
    }

    # current, alpha = 1, index, max.tries = 5, n = 5, delta = 0.01, cur_index = NA, ..
    basis <- search_f(
      current, alpha = alpha, index = index, tries = tries, max.tries = max.tries,
      cur_index = cur_index, frozen = frozen, n_sample = n_sample, ...)

    if (method == "search_posse") {
      if (!is.null(basis$h)) {
        if (basis$h > 30) {
          alpha <<- alpha * cooling
        }
      }
    } else {
      alpha <<- alpha * cooling
    }

    list(target = basis$target, index = index)
  }

  new_geodesic_path("guided", generator)
}
# globalVariables(c("cur_index", "index"))

#' A local tour path.
#'
#' The local tour alternates between the starting position and a
#' nearby random projection.
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param start initial projection matrix
#' @param angle distance in radians to stay within
#' @export
#' @examples
#' animate_xy(flea[, 1:3], local_tour(basis_init(3, 2)))
#' animate_xy(flea[, 1:3], local_tour(basis_init(3, 2), 0.2))
#' animate_xy(flea[, 1:3], local_tour(basis_random(3, 2), 0.2))
local_tour <- function(start, angle = pi / 4) {
  odd <- TRUE

  generator <- function(current, data, ...) {
    if (odd) {
      new_basis <- start
      odd <<- !odd
      return(new_basis)
    } else {
      new <- basis_random(nrow(start), ncol(start))
      dist <- stats::runif(1, 0, angle)
      new_basis <- step_angle(geodesic_info(start, new), dist)
    }


    target <- new_basis
    list(target = target)
  }

  new_geodesic_path("local", generator)
}

#' Section pursuit index.
#'
#' Calculates a section pursuit index that compares the distribution
#' inside and outside a slice.
#'
#' @param breaks_x binning on the first variable (x or radius).
#' @param breaks_y binning on the second variable (y or angle).
#' @param eps cutoff values to suppress summing up small differences.
#'   Vector with one entry for each bin, can be estimated
#'   using \code{\link{estimate_eps}}.
#' @param bintype select polar (default) or square binning.
#' @param power exponent q used in the index compuatation.
#' @param flip sign of the index computation, select +1 when searching
#'   for low densities and -1 when searching for high densities.
#' @param reweight if TRUE will reweight according to the expected distribution
#'   in a uniform hypersphere (default is FALSE).
#' @param p number of variables in the data (needed for accurate reweighting,
#'   default is 4).
#' @export
slice_index <- function(breaks_x, breaks_y, eps, bintype = "polar", power = 1, flip = 1,
                        reweight = FALSE, p = 4) {
  if (reweight) {
    if (bintype != "polar") {
      message("Reweighting is only defined for polar binning and will be ignored.")
    }
    else {
      message(paste0("Reweighting assuming p=", p))
    }
  }

  resc <- 1
  if (bintype == "polar") {
    resc <- 1 / (1 - (1 / 10)^(1 / power))^power
    message(paste0("Rescaling raw index by a factor ", resc))
  }

  function(mat, dists, h) {

    # call binner, return zero if binning is not recognised
    mat_tab <- slice_binning(mat, dists, h, breaks_x, breaks_y, bintype = bintype)
    if (length(mat_tab) == 1 && mat_tab == 0) {
      return(0)
    }

    # no result if no points inside the slice
    if (length(mat_tab$n[mat_tab$inSlice == TRUE]) == 0) {
      return(0)
    }

    # splitting into in and out of slice
    mat_in <- mat_tab[mat_tab$inSlice == TRUE, ]
    mat_out <- mat_tab[mat_tab$inSlice == FALSE, ]

    if (bintype == "polar" && reweight) {
      # get bin-wise weights
      w_in <- weights_bincount_radial(mat_in, 2)
      w_out <- weights_bincount_radial(mat_out, p)
      # rescale and normalise n
      mat_in$n <- w_in * mat_in$n / sum(mat_in$n)
      mat_out$n <- w_out * mat_out$n / sum(mat_out$n)
    }
    else {
      # normalise n
      mat_in$n <- mat_in$n / sum(mat_in$n)
      mat_out$n <- mat_out$n / sum(mat_out$n)
    }

    # getting binwise density differences
    if (power != 1) {
      x <- flip * (mat_out$n^(1 / power) -
        mat_in$n^(1 / power))
      y <- flip * (mat_out$n -
        mat_in$n)
      x[y < eps] <- 0 # dropping bins based on y
    }
    else {
      x <- flip * (mat_out$n - mat_in$n)
      x[x < eps] <- 0
    }
    if (power != 1) {
      resc * sum(x^power)
    } else {
      resc * sum(x)
    }
  }
}

#' Estimate cutoff eps for section pursuit.
#'
#' @param N total number of points in the input data.
#' @param p number of dimensions of the input data.
#' @param res resolution, (slice radius)/(data radius)
#' @param K total number of bins
#' @param K_theta number of angular bins
#' @param r_breaks boundaries of the radial bins
#' @export
estimate_eps <- function(N, p, res, K, K_theta, r_breaks) {
  r_max <- max(r_breaks)
  ret <- c()
  for (i in 2:length(r_breaks)) {
    r1 <- r_breaks[i - 1]
    r2 <- r_breaks[i]
    delta_i <- (r_max / sqrt(r2^2 - r1^2)) * sqrt(2 * K_theta / N) *
      res^((2 - p) / 2) / sqrt(p - (p - 2) * res^2)
    eps <- delta_i / K
    ret <- c(ret, rep(eps, K_theta))
  }
  ret
}

#' Separately binning observations inside and outside the slice.
#'
#' @param mat projected data points
#' @param dists vector of all point distances from the projection plane.
#' @param h slice thickness.
#' @param breaks_x binning on the first variable (x or radius).
#' @param breaks_y binning on the second variable (y or angle).
#' @param bintype select polar (default) or square binning.
#' @keywords internal
slice_binning <- function(mat, dists, h, breaks_x, breaks_y, bintype = "square") {

  # centering
  mat <- apply(mat, 2, function(x) (x - mean(x)))

  if (bintype == "square") {
    xbin <- cut(mat[, 1], breaks_x)
    ybin <- cut(mat[, 2], breaks_y)
  }
  else if (bintype == "polar") {
    rad <- sqrt(mat[, 1]^2 + mat[, 2]^2)
    ang <- atan2(mat[, 2], mat[, 1])
    xbin <- cut(rad, breaks_x)
    ybin <- cut(ang, breaks_y)
  }
  else {
    message(bintype, " is not a recognised bin type")
    return(0)
  }

  # track which points are inside the current slice
  inSlice <- factor(dists < h)

  # return binned data
  # to match order of radial rescaling vector we pass in ybin first
  ret <- as.data.frame(table(ybin, xbin, inSlice, useNA = "no"))
  colnames(ret) <- c("ybin", "xbin", "inSlice", "n")
  ret
}

#' Returns n equidistant bins between -pi and pi
#'
#' @param n number of bins
#' @export
angular_breaks <- function(n) {
  seq(-pi, pi, length.out = n + 1)
}

#' Returns n equidistant bins between a and b
#'
#' @param n number of bins
#' @param a lower bound
#' @param b upper bound
#' @export
linear_breaks <- function(n, a, b) {
  seq(a, b, length.out = n + 1)
}

#' Cumulative distribution in radius.
#'
#' Fraction of points within radius r given 2D projection of a
#' hypersphere with radius R in p dimensions.
#'
#' @keywords internal
cumulative_radial <- function(r, R, p) {
  1 - (1 - (r / R)^2)^(p / 2)
}


#' Inverse weights for rescaling counts in radial bins.
#'
#' @keywords internal
radial_bin_weight_inv <- function(r1, r2, R, p) {
  cumulative_radial(r2, R, p) - cumulative_radial(r1, R, p)
}

#' Computes weights for the rescaling of radial bin counts.
#'
#' @param histo input histogram (for bin definitions)
#' @param p number of variables
#'
#' @keywords internal
weights_bincount_radial <- function(histo, p) {
  # store weights in vector w
  w <- rep(0, nrow(histo))
  # extract maximum radius from all bins
  R <- max(as.numeric(sub("[^,]*,([^]]*)\\]", "\\1", histo$xbin)))
  n <- length(unique(histo$xbin))
  for (i in 1:nrow(histo)) {
    r1 <- as.numeric(sub("\\((.+),.*", "\\1", histo$xbin[[i]]))
    r2 <- as.numeric(sub("[^,]*,([^]]*)\\]", "\\1", histo$xbin[[i]]))
    w[i] <- radial_bin_weight_inv(r1, r2, R, p)
  }
  1 / (n * w)
}

#' A grand tour path.
#'
#' This method generates target bases by randomly sampling on
#' the space of all d-dimensional planes in p-space.
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param d target dimensionality
#' @param ... arguments sent to the generator
#' @export
#' @examples
#' # All animation methods use the grand tour path by default
#' animate_dist(flea[, 1:6])
#' animate_xy(flea[, 1:6])
#' animate_pcp(flea[, 1:6])
#' animate_pcp(flea[, 1:6], grand_tour(4))
#'
#' # The grand tour is a function:
#' tour2d <- grand_tour(2)
#' is.function(tour2d)
#'
#' # with two parameters, the previous projection and the data set
#' args(tour2d)
#' # if the previous projection is null, it will generate a starting
#' # basis, otherwise the argument is ignored
#' tour2d(NULL, mtcars)
#' # the data argument is just used to determine the correct dimensionality
#' # of the output matrix
#' tour2d(NULL, mtcars[, 1:2])
grand_tour <- function(d = 2, ...) {
  generator <- function(current, data, ...) {
    if (is.null(current)) {
      return(basis_init(ncol(data), d))
    }

    target <- basis_random(ncol(data), d)
    list(target = target)
  }

  new_geodesic_path("grand", generator)
}

#' Generate a random basis
#'
#' @keywords internal
#' @param n dimensionality of data
#' @param d dimensionality of target projection
#' @export
basis_random <- function(n, d = 2) {
  mvn <- matrix(stats::rnorm(n * d), ncol = d)
  orthonormalise(mvn)
}

#' Generate initial basis.
#'
#' First two variables are projected on first two axes.
#'
#' @keywords internal
#' @param n dimensionality of data
#' @param d dimensionality of target projection
#' @export
basis_init <- function(n, d) {
  start <- matrix(0, nrow = n, ncol = d)
  diag(start) <- 1
  start
}

# Tests for tour path generators
#
# Every tour path generator must satisfy three contracts:
#
#   1. CONSTRUCTION  — the object it returns has class "tour_path" and the
#      correct "name" attribute, so new_tour() can dispatch correctly.
#
#   2. GENERATION    — calling the generator with (current_basis, data)
#      returns a geodesic list with at minimum an $ingred component, which
#      itself contains $dist (a positive scalar) and an $interpolate
#      function. Returning NULL signals the tour should stop (used by
#      planned_tour at the end of its sequence).
#
#   3. INTEGRATION   — save_history() + interpolate() round-trip works end-
#      to-end, producing a valid history_array with orthonormal columns.
#      This is the closest thing to an animate() test that can run headlessly.
#
# The orthonormality helper is used throughout to verify basis validity.

# ---------------------------------------------------------------------------
# Shared fixtures
# ---------------------------------------------------------------------------

dat  <- as.matrix(flea[, 1:6])   # 74 x 6
dat4 <- as.matrix(flea[, 1:4])   # 74 x 4  (used for frozen/dependence tests)

# Check that all slices of a history_array have orthonormal columns.
expect_orthonormal_history <- function(h, tol = 1e-6) {
  n <- dim(h)[3]
  for (k in seq_len(n)) {
    b <- matrix(unclass(h)[, , k], nrow = dim(h)[1])
    gram <- crossprod(b)
    expect_equal(gram, diag(ncol(b)), tolerance = tol,
      label = sprintf("basis at frame %d is orthonormal", k))
  }
}


# ===========================================================================
# grand_tour
# ===========================================================================

test_that("grand_tour has correct class and name", {
  tp <- grand_tour(2)
  expect_s3_class(tp, "tour_path")
  expect_equal(attr(tp, "name"), "grand")
})

test_that("grand_tour(2) generates valid geodesic ingredients", {
  tp   <- grand_tour(2)
  from <- basis_random(6, 2)
  g    <- tp(from, dat)
  expect_false(is.null(g$ingred))
  expect_gt(g$ingred$dist, 0)
  expect_true(is.function(g$ingred$interpolate))
  # Interpolated midpoint should be orthonormal
  mid <- g$ingred$interpolate(0.5)
  expect_equal(crossprod(mid), diag(2), tolerance = 1e-6)
})

test_that("grand_tour works for 1D (singleton column)", {
  # Regression: this is the dimension that triggers the Windows/Linux
  # dimension-drop bug in history_array handling.
  tp   <- grand_tour(1)
  from <- basis_random(6, 1)
  g    <- tp(from, dat)
  expect_false(is.null(g$ingred))
  mid  <- g$ingred$interpolate(0.5)
  expect_equal(dim(mid), c(6L, 1L))
  expect_equal(crossprod(mid)[1, 1], 1, tolerance = 1e-6)
})

test_that("grand_tour save_history + interpolate round-trip is orthonormal", {
  set.seed(1)
  h <- save_history(dat, grand_tour(2), max = 5)
  expect_equal(dim(h)[1:2], c(6L, 2L))
  interp <- interpolate(h, 0.05)
  expect_orthonormal_history(interp)
})

test_that("grand_tour(1) save_history + interpolate round-trip is orthonormal", {
  set.seed(1)
  h <- save_history(dat, grand_tour(1), max = 5)
  expect_equal(dim(h)[1:2], c(6L, 1L))
  interp <- interpolate(h, 0.05)
  expect_orthonormal_history(interp)
})


# ===========================================================================
# little_tour
# ===========================================================================

test_that("little_tour has correct class and name", {
  tp <- little_tour()
  expect_s3_class(tp, "tour_path")
  expect_equal(attr(tp, "name"), "little")
})

test_that("little_tour generates axis-parallel bases", {
  # The little tour cycles through axis-parallel projections: each column of
  # the basis has exactly one non-zero entry of ±1 at each extreme.
  h <- save_history(dat4, little_tour(), max = 10)
  expect_equal(dim(h)[1:2], c(4L, 2L))
  # All saved bases should be orthonormal
  expect_orthonormal_history(h)
})

test_that("little_tour save_history + interpolate round-trip works", {
  h      <- save_history(dat4, little_tour(), max = 6)
  interp <- interpolate(h, 0.05)
  expect_orthonormal_history(interp)
})


# ===========================================================================
# planned_tour
# ===========================================================================

test_that("planned_tour has correct class and name", {
  set.seed(1)
  bases <- save_history(dat, grand_tour(2), max = 4)
  tp    <- planned_tour(bases)
  expect_s3_class(tp, "tour_path")
  expect_equal(attr(tp, "name"), "planned")
})

test_that("planned_tour replays exactly the saved bases", {
  set.seed(1)
  bases  <- save_history(dat, grand_tour(2), max = 4)
  # Replay via a second save_history using planned_tour — we should recover
  # a history with at least as many frames as the original.
  replay <- save_history(dat, planned_tour(bases))
  expect_gte(dim(replay)[3], dim(bases)[3])
  expect_orthonormal_history(replay)
})

test_that("planned_tour with cycle = TRUE does not terminate early", {
  set.seed(1)
  bases  <- save_history(dat, grand_tour(2), max = 3)
  # With cycling, requesting more frames than bases should succeed
  replay <- save_history(dat, planned_tour(bases, cycle = TRUE), max = 9)
  expect_gte(dim(replay)[3], 6L)
})

test_that("planned_tour works for 1D saved history", {
  set.seed(1)
  bases  <- save_history(dat, grand_tour(1), max = 4)
  replay <- save_history(dat, planned_tour(bases))
  expect_equal(dim(replay)[2], 1L)
  expect_orthonormal_history(replay)
})


# ===========================================================================
# guided_tour
# ===========================================================================

test_that("guided_tour has correct class and name", {
  tp <- guided_tour(holes())
  expect_s3_class(tp, "tour_path")
  expect_equal(attr(tp, "name"), "guided")
})

test_that("guided_tour with holes index produces valid history", {
  set.seed(42)
  h <- save_history(dat, guided_tour(holes()), max = 5)
  expect_equal(dim(h)[1:2], c(6L, 2L))
  expect_orthonormal_history(h)
})

test_that("guided_tour with cmass index produces valid history", {
  set.seed(42)
  h <- save_history(dat, guided_tour(cmass()), max = 5)
  expect_equal(dim(h)[1:2], c(6L, 2L))
  expect_orthonormal_history(h)
})

test_that("guided_tour index values are non-decreasing along saved path", {
  # The guided tour is an optimisation: each new basis should have a higher
  # (or equal) index value than the previous one.
  #
  # holes() takes the projected data matrix (n x d), not the basis or the
  # original data. We compute proj = dat %*% basis for each saved frame.
  set.seed(42)
  h      <- save_history(dat, guided_tour(holes()), max = 8)
  scores <- sapply(seq_len(dim(h)[3]), function(k) {
    basis <- matrix(unclass(h)[, , k], nrow = 6)
    proj  <- dat %*% basis
    holes()(proj)
  })
  # Allow slight numerical noise with a small tolerance
  diffs <- diff(scores)
  expect_true(all(diffs >= -1e-4),
    label = "holes index is non-decreasing along guided tour path")
})


# ===========================================================================
# local_tour
# ===========================================================================

test_that("local_tour has correct class and name", {
  tp <- local_tour(basis_random(6, 2))
  expect_s3_class(tp, "tour_path")
  expect_equal(attr(tp, "name"), "local")
})

test_that("local_tour stays close to its starting basis", {
  # The angle argument to local_tour() controls the size of the neighbourhood
  # used when searching for the next target basis — it is not a hard ceiling
  # on proj_dist() for every interpolated frame. Intermediate frames can
  # slightly exceed the angle as the tour moves between targets.
  #
  # What we can reliably assert:
  #   1. All frames stay within a reasonable bound (here 2 * angle), ruling
  #      out the tour behaving like an unconstrained grand tour.
  #   2. The local tour produces frames closer to the start on average than
  #      an equivalent grand tour would.
  set.seed(1)
  angle <- pi / 8
  start <- basis_random(6, 2)

  h_local <- save_history(dat, local_tour(start, angle = angle), max = 20)
  h_grand <- save_history(dat, grand_tour(2),                    max = 20)

  dists_local <- sapply(seq_len(dim(h_local)[3]), function(k) {
    proj_dist(start, matrix(unclass(h_local)[, , k], nrow = 6))
  })
  dists_grand <- sapply(seq_len(dim(h_grand)[3]), function(k) {
    proj_dist(start, matrix(unclass(h_grand)[, , k], nrow = 6))
  })

  # No frame should escape a generous bound (2 * angle)
  expect_true(all(dists_local <= 2 * angle),
    label = "all local tour frames stay within 2 * angle of start")

  # Local tour should be closer on average than the grand tour
  expect_lt(mean(dists_local), mean(dists_grand),
    label = "local tour stays closer to start on average than grand tour")
})

test_that("local_tour save_history + interpolate round-trip is orthonormal", {
  set.seed(1)
  start  <- basis_random(6, 2)
  h      <- save_history(dat, local_tour(start), max = 5)
  interp <- interpolate(h, 0.05)
  expect_orthonormal_history(interp)
})

# ===========================================================================
# frozen_tour
# ===========================================================================

test_that("frozen_tour has correct class and name", {
  frozen     <- matrix(NA, nrow = 4, ncol = 2)
  frozen[3,] <- 0.5
  tp         <- frozen_tour(2, frozen)
  expect_s3_class(tp, "tour_path")
  expect_equal(attr(tp, "name"), "frozen")
})

test_that("frozen_tour save_history + interpolate round-trip is orthonormal", {
  frozen     <- matrix(NA, nrow = 4, ncol = 2)
  frozen[1, 1] <- 0.5
  set.seed(1)
  h      <- save_history(dat4, frozen_tour(2, frozen), max = 5)
  interp <- interpolate(h, 0.05)
  expect_orthonormal_history(interp)
})


# ===========================================================================
# radial_tour
# ===========================================================================

test_that("radial_tour has correct class and name", {
  start <- basis_random(6, 2)
  tp    <- radial_tour(start, mvar = 1)
  expect_s3_class(tp, "tour_path")
  expect_equal(attr(tp, "name"), "radial")
})

test_that("radial_tour works with boundary mvar values", {
  # Check that the first and last variable indices both work — these are the
  # boundary cases most likely to break if mvar is used to subset a matrix.
  set.seed(1)
  start <- basis_random(6, 2)

  h_first <- save_history(dat, radial_tour(start, mvar = 1), max = 10)
  expect_equal(dim(h_first)[1:2], c(6L, 2L))
  expect_orthonormal_history(h_first)

  h_last <- save_history(dat, radial_tour(start, mvar = 6), max = 10)
  expect_equal(dim(h_last)[1:2], c(6L, 2L))
  expect_orthonormal_history(h_last)
})

test_that("radial_tour save_history produces orthonormal bases", {
  set.seed(1)
  start <- basis_random(6, 2)
  h     <- save_history(dat, radial_tour(start, mvar = 1), max = 10)
  expect_equal(dim(h)[1:2], c(6L, 2L))
  expect_orthonormal_history(h)
})

test_that("radial_tour drives mvar column through zero during sweep", {
  # The defining behaviour of the radial tour: variable mvar is rotated
  # smoothly out of the projection. At the midpoint of the sweep its
  # contribution (both rows of the mvar column) should be at or near zero.
  # We capture enough frames that the midpoint is well-sampled.
  set.seed(1)
  start <- basis_random(6, 2)
  mvar  <- 3L
  h     <- save_history(dat, radial_tour(start, mvar = mvar), max = 30)

  # Extract the norm of row `mvar` across all frames — it should dip
  # close to zero somewhere in the sequence.
  norms <- sapply(seq_len(dim(h)[3]), function(k) {
    b <- matrix(unclass(h)[, , k], nrow = 6)
    sqrt(sum(b[mvar, ]^2))   # Euclidean norm of the mvar row
  })
  expect_lt(min(norms), 0.1,
    label = "mvar row norm passes through near-zero during radial sweep")
})

test_that("radial_tour works with multiple mvar indices", {
  set.seed(1)
  start <- basis_random(6, 2)
  # Two variables rotated out simultaneously
  h <- save_history(dat, radial_tour(start, mvar = c(2, 5)), max = 10)
  expect_equal(dim(h)[1:2], c(6L, 2L))
  expect_orthonormal_history(h)
})

test_that("radial_tour save_history + interpolate round-trip is orthonormal", {
  set.seed(1)
  start  <- basis_random(6, 2)
  h      <- save_history(dat, radial_tour(start, mvar = 4), max = 10)
  interp <- interpolate(h, 0.05)
  expect_orthonormal_history(interp)
})

#' A pseudo-derivative, line search algorithm along frozen geodesics.
#'
#' @param current starting projection
#' @param index index function
#' @param tries the counter of the outer loop of the opotimiser
#' @param max.tries maximum number of iteration before giving up
#' @param n number of random steps to take to find best direction
#' @param frozen a matrix of frozen variables
#' @param cur_index the index value of the current basis
#' @param ... other arguments being passed into the \code{search_frozen_geodesic()}
#' These three functions perform a corresponding role to
#' \code{\link{search_geodesic}}, \code{\link{find_best_dir}} and
#' \code{\link{find_path_peak}} but for the frozen tour.  They work by
#' zero'ing out the frozen variables and travelling in that restricted
#' subspace.
#'
#' @section To do: eliminate these functions
#' @keywords internal
search_frozen_geodesic <- function(current, index, tries, max.tries = 5, n = 5, frozen, cur_index = NA, ...) {
  cur_index <- index(thaw(current, frozen))

  try <- 1
  while (try < max.tries) {
    # Try 5 random directions and pick the one that has the highest
    # index after a small step in either direction
    direction <- find_best_frozen_dir(current, frozen, index, n)
    best_dir <- direction$basis[[which.max(direction$index_val)]]
    direction_search <- dplyr::mutate(direction, tries = tries, loop = try)
    # Travel halfway round (pi / 4 radians) the sphere in that direction
    # looking for the best projection
    peak <- find_frozen_path_peak(current, best_dir, frozen, index)
    line_search <- dplyr::mutate(peak$best, tries = tries, loop = try)
    new_index <- tail(line_search$index_val, 1)
    new_basis <- tail(line_search$basis, 1)

    rcd_env <- parent.frame(n = 4)
    if (is.null(rcd_env[["record"]])) rcd_env <- parent.frame(n = 1)
    rcd_env[["record"]] <- dplyr::bind_rows(
      rcd_env[["record"]],
      direction_search,
      line_search
    )

    dig3 <- function(x) sprintf("%.3f", x)
    pdiff <- (new_index - cur_index) / cur_index
    if (pdiff > 0.001) {
      message("New index: ", dig3(new_index), " (", dig3(peak$alpha$maximum), " away)", sep = "")
      current <<- new_basis
      cur_index <<- new_index

      return(list(target = new_basis[[1]]))
    }
    message("Best was:  ", dig3(new_index), " (", dig3(peak$alpha$maximum), " away).  Try again...", sep = "")

    try <- try + 1
  }

  NULL
}

#' Find most promising direction in frozen space.
#' @keywords internal
find_best_frozen_dir <- function(old, frozen, index, dist = 0.01, counter = 5) {
  new_basis <- function() freeze(basis_random(nrow(old), ncol(old)), frozen)
  bases <- replicate(counter, new_basis(), simplify = FALSE)
  old <- freeze(old, frozen)

  score <- function(new) {
    interpolator <- geodesic_info(old, new)
    forward <- thaw(step_angle(interpolator, dist), frozen)
    backward <- thaw(step_angle(interpolator, -dist), frozen)

    larger <- max(index(forward), index(backward))

    tibble::tibble(
      basis = c(list(forward), list(backward)),
      index_val = c(index(forward), index(backward)),
      info = "direction_search",
      method = "search_frozen_geodesic"
    )
  }

  ans <- dplyr::bind_rows(lapply(bases, score))
  ans[which.max(ans$index_val), "info"] <- "best_direction_search"
  ans
}

#' Find most highest peak along frozen geodesic.
#' @keywords internal
find_frozen_path_peak <- function(old, new, frozen, index, max_dist = pi / 4) {
  interpolator <- geodesic_info(freeze(old, frozen), freeze(new, frozen))

  index_pos <- function(alpha) {
    index(thaw(step_angle(interpolator, alpha), frozen))
  }

  alpha <- stats::optimize(index_pos, c(-max_dist, max_dist), maximum = TRUE, tol = 0.01)

  best <- dplyr::tibble(
    basis = list(thaw(step_angle(interpolator, alpha$maximum), frozen)),
    index_val = alpha$objective,
    info = "best_line_search",
    method = "search_frozen_geodesic"
  )

  # xgrid <- seq(-max_dist, max_dist, length = 100)
  # index <- sapply(xgrid, index_pos)
  # plot(xgrid, index, type = "l")

  list(best = best, alpha = alpha)
}

#' Compute distance matrix from bases.
#'
#' @param history history of the plots
#' @export
#' @examples
#' # This code is to be used as an example but you should increase
#' # the max from 2 to 50, say, to check tour coverage.
#' flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
#' grand <- interpolate(save_history(flea_std, max = 2), 0.2)
#' # The grand tour  -----------------------------
#' # Look at the tour path in a tour, how well does it cover a sphere
#' # Using MDS to summarise the high-d space of projections
#' # Last basis is a duplicate, needs removing
#' d <- path_dist(grand[,,-dim(grand)[[3]]])
#' ord <- as.data.frame(MASS::isoMDS(d)$points)
#' require(ggplot2)
#' ggplot(data = ord, aes(x=V1, y=V2)) +
#'   geom_path() +
#'   coord_equal() +
#'   labs(x = NULL, y = NULL)
#'
#' # Compare five guided tours  -----------------------------
#' holes1d <- guided_tour(holes(), 1)
#' tour_reps <- replicate(5, save_history(flea_std, holes1d, max = 2),
#'   simplify = FALSE
#' )
#' tour_reps2 <- lapply(tour_reps, interpolate, 0.2)
#'
#' bases <- unlist(lapply(tour_reps2, as.list), recursive = FALSE)
#' class(bases) <- "history_list"
#' index_values <- paths_index(tour_reps2, holes())
#' index_values$step <- index_values$step.1
#' d <- path_dist(bases)
#' ord <- as.data.frame(cmdscale(d, 2))
#'
#' info <- cbind(ord, index_values)
#' ggplot(data = info, aes(x = step, y = value, group = try)) +
#'   geom_line()
#' ##ggplot(data = info, aes(x = V1, y = V2, group = try)) +
#' ##  geom_path() +
#' ##  geom_point(aes(size = value)) +
#' ##  coord_equal()
#' ##last_plot() + facet_wrap(~try)
path_dist <- function(history) {
  history <- as.array(history)
  n <- dim(history)[3]
  d <- matrix(NA, nrow = n, ncol = n)

  for (i in seq_len(n)) {
    for (j in seq_len(i - 1)) {
      d[i, j] <- proj_dist(history[[i]], history[[j]])
    }
  }
  stats::as.dist(d)
}

#' Star glyph tour path animation.
#'
#' Animate a nD tour path with star glyphs.
#'
#' Currently, scaling doesn't seem to be computed absolutely correctly, as
#' centres move around as well as outside points.
#'
#' @param ... other arguments passed on to \code{\link{stars}}
#' @seealso \code{\link{animate}} for options that apply to all animations
#' @keywords hplot
#' @export
#' @examples
#' animate_stars(flea[1:10, 1:6])
#' animate_stars(flea[1:10, 1:6], grand_tour(5))
#' animate_stars(flea[, 1:6], grand_tour(5))
#' animate_stars(flea[1:10, 1:6], grand_tour(5),
#'   col.stars = rep("grey50", 10), radius = FALSE
#' )
display_stars <- function(...) {
  render_data <- function(data, proj, geodesic) {
    x <- data %*% proj
    x <- (x + 2) / 4
    stars(x, scale = FALSE, ...)
  }

  list(
    init = nul,
    render_frame = nul,
    render_transition = nul,
    render_data = render_data,
    render_target = nul
  )
}


#' @rdname display_stars
#' @inheritParams animate
#' @export
animate_stars <- function(data, tour_path = grand_tour(3), ...) {
  animate(
    data = data, tour_path = tour_path,
    display = display_stars(...), ...
  )
}

#' Generate geodesic path.
#'
#' Wrap basis generation method with a function that computes the
#' geodesic interpolation from the previous frame to the next frame, and
#' provides convenient access to all the information about the path.
#'
#' Frozen variables allow us to keep certain values of the projection
#' fixed and generate a geodesic across the subspace generated by those
#'
#' @param current starting projection
#' @param target ending projection
#' @param frozen matrix giving frozen variables, as described in
#'   \code{\link{freeze}}
#' @keywords internal
#' @export
#' @return
#'   \item{interpolate}{A function with single parameter in [0, 1] that
#'     returns an interpolated frame between the current and future frames.
#'     0 gives the current plane, 1 gives the new target frame in plane of
#'     current frame.}
#'   \item{dist}{The distance, in radians, between the current and target
#'     frames.}
#'  \item{Fa}{The current frame.}
#'  \item{Fz}{The new target frame.}
#'  \item{tau}{The principle angles between the current and target frames.}
#'  \item{Ga}{The current plane.}
#'  \item{Gz}{The target plane.}
#' @examples
#' a <- basis_random(4, 2)
#' b <- basis_random(4, 2)
#' path <- geodesic_path(a, b)
#'
#' path$dist
#' all.equal(a, path$interpolate(0))
#' # Not true generally - a rotated into plane of b
#' all.equal(b, path$interpolate(1))
geodesic_path <- function(current, target, frozen = NULL, ...) {
  if (is.null(frozen)) {
    # Regular geodesic
    geodesic <- geodesic_info(current, target)

    interpolate <- function(pos) {
      step_fraction(geodesic, pos)
    }
  } else {
    # Frozen geodesic
    current_froz <- freeze(current, frozen)
    target_froz <- freeze(target, frozen)

    geodesic <- geodesic_info(current_froz, target_froz)

    interpolate <- function(pos) {
      thaw(step_fraction(geodesic, pos), frozen)
    }
  }

  list(
    interpolate = interpolate,
    Fa = current,
    Fz = target,
    Ga = geodesic$Ga,
    Gz = geodesic$Gz,
    tau = geodesic$tau,
    dist = proj_dist(current, target)
  )
}

#' Calculate information required to interpolate along a geodesic path between
#' two frames.
#'
#' The methdology is outlined in
#' \url{http://www-stat.wharton.upenn.edu/~buja/PAPERS/paper-dyn-proj-algs.pdf}
#' and
#' \url{http://www-stat.wharton.upenn.edu/~buja/PAPERS/paper-dyn-proj-math.pdf},
#' and the code follows the notation outlined in those papers:
#'
#' \itemize{
#'   \item p = dimension of data
#'   \item d = target dimension
#'   \item F = frame, an orthonormal p x d matrix
#'   \item Fa = starting frame, Fz = target frame
#'   \item Fa'Fz = Va lamda  Vz' (svd)
#'   \item Ga = Fa Va, Gz = Fz Vz
#'   \item tau = principle angles
#' }
#' @keywords internal
#' @param Fa starting frame, will be orthonormalised if necessary
#' @param Fz target frame, will be orthonormalised if necessary
#' @param epsilon epsilon used to determine if an angle is effectively equal
#'   to 0
geodesic_info <- function(Fa, Fz, epsilon = 1e-6) {
  if (!is_orthonormal(Fa)) {
    # message("Orthonormalising Fa")
    Fa <- orthonormalise(Fa)
  }
  if (!is_orthonormal(Fz)) {
    # message("Orthonormalising Fz")
    Fz <- orthonormalise(Fz)
  }

  # if (Fa.equivalent(Fz)) return();
  # message("dim Fa",nrow(Fa),ncol(Fa),"dim Fz",nrow(Fz),ncol(Fz))

  # Compute the SVD: Fa'Fz = Va lambda Vz' --------------------------------
  sv <- svd(t(Fa) %*% Fz)

  # R returns the svd from smallest to largest -------------------------------
  nc <- ncol(Fa)
  lambda <- sv$d[nc:1]
  Va <- sv$u[, nc:1]
  Vz <- sv$v[, nc:1]

  # Compute frames of principal directions (planes) ------------------------
  Ga <- Fa %*% Va
  Gz <- Fz %*% Vz

  # Form an orthogonal coordinate transformation --------------------------
  Ga <- orthonormalise(Ga)
  Gz <- orthonormalise(Gz)
  Gz <- orthonormalise_by(Gz, Ga)

  # Compute and check principal angles -----------------------
  tau <- suppressWarnings(acos(lambda))
  badtau <- is.nan(tau) | tau < epsilon
  Gz[, badtau] <- Ga[, badtau]
  tau[badtau] <- 0

  list(Va = Va, Ga = Ga, Gz = Gz, tau = tau)
}

#' Step along an interpolated path by fraction of path length.
#'
#' @keywords internal
#' @param interp interpolated path
#' @param fraction fraction of distance between start and end planes
step_fraction <- function(interp, fraction) {
  # Interpolate between starting and end planes
  #  - must multiply column wise (hence all the transposes)
  G <- t(
    t(interp$Ga) * cos(fraction * interp$tau) +
      t(interp$Gz) * sin(fraction * interp$tau)
  )

  # rotate plane to match frame Fa
  orthonormalise(G %*% t(interp$Va))
}

#' Step along an interpolated path by angle in radians.
#'
#' @keywords internal
#' @param interp interpolated path
#' @param angle angle, in radians
step_angle <- function(interp, angle) {
  step_fraction(interp, angle / sqrt(sum(interp$tau^2)))
}

#' Parallel coordinates tour path animation.
#'
#' Animate a nD tour path with a parallel coordinates plot.
#'
#' The lines show the observations, and the points, the values of the
#' projection matrix.
#'
#' @param ... other arguments passed on to \code{\link{animate}}
#' @seealso \code{\link{animate}} for options that apply to all animations
#' @keywords hplot
#' @export
#' @examples
#' animate_pcp(flea[, 1:6], grand_tour(3))
#' animate_pcp(flea[, 1:6], grand_tour(5))
display_pcp <- function(...) {
  labels <- NULL
  limit <- NULL
  init <- function(data) {
    labels <<- abbreviate(colnames(data), 2)
    limit <<- max(sqrt(rowSums(data^2)))
  }

  render_frame <- function() {
    blank_plot(xlim = c(0, 1), ylim = c(-limit, limit))
  }
  render_transition <- function() {
    rect(0, -limit, 1, limit, col = "#FFFFFFE6", border = NA)
  }
  render_data <- function(data, proj, geodesic) {
    d <- ncol(proj)
    xpos <- 0.05 + seq(0, 0.9, length = d)
    ys <- as.vector(t(cbind(data %*% proj, NA)))
    xs <- rep(c(xpos, NA), length = length(ys))

    # Grid lines
    segments(xpos, limit, xpos, -limit, col = "grey90")
    segments(0, 0, 1, 0, col = "grey90")

    # Projection values
    label_df <- data.frame(
      x = xpos,
      y = as.vector(t(proj)),
      label = rep(labels, each = d)
    )
    with(
      subset(label_df, abs(y) > 0.05),
      text(x, y, label = label, col = "grey50", pos = 4)
    )
    with(label_df, points(x, y, col = "grey50", pch = 20))

    # Data values
    lines(xs, ys)
  }

  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}
# globalVariables("y")

#' @rdname display_pcp
#' @inheritParams animate
#' @export
animate_pcp <- function(data, tour_path = grand_tour(3), ...) {
  animate(
    data = data, tour_path = tour_path,
    display = display_pcp(...), ...
  )
}

#' Interpolate geodesically between bases.
#'
#' This function takes a set of bases and produces a tour by geodesically
#' interpolating between each basis
#'
#' @param basis_set input basis set
#' @param angle target distance (in radians) between bases
#' @param cycle For \code{\link{planned_tour}} cycle through continuously (TRUE) or stop after first pass (FALSE)
#' @keywords hplot
#' @export
#' @examples
#' t1 <- save_history(flea[, 1:6], grand_tour(1), max = 3)
#' dim(t1)
#' dim(interpolate(t1, 0.01))
#' dim(interpolate(t1, 0.05))
#' dim(interpolate(t1, 0.1))
#' t2 <- save_history(flea[, 1:6], grand_tour(2), max = 2)
#' dim(interpolate(t2, 0.05))
interpolate <- function(basis_set, angle = 0.05, cycle = FALSE) {
  basis_set <- as.array(basis_set)
  n <- dim(basis_set)[3]
  if (n < 2) {
    return(basis_set)
  }

  record <-
    tibble::tibble(
      basis = list(),
      index_val = numeric(),
      info = character(),
      method = character(),
      alpha = numeric(),
      tries = numeric(),
      loop = numeric()
    )
  # Estimate number of bases in output
#  dists <- sapply(2:n, function(i) {
#    proj_dist(apply(basis_set[,,i - 1], 2, c), apply(basis_set[,,i], 2, c))
#  })
  dists <- sapply(2:n, function(i) {
    proj_dist(
      matrix(unclass(basis_set)[, , i - 1], nrow = dim(basis_set)[1]),
      matrix(unclass(basis_set)[, , i],     nrow = dim(basis_set)[1])
    )
  })
  steps <- sum(ceiling(dists / angle)) * 2

  new_basis <- rep(NA, steps)
  new_basis[1] <- TRUE

  # Initialise result storage
  projs <- array(NA_real_, c(dim(basis_set)[1:2], steps))

  i <- 1
#  tour <- new_tour(basis_set[, , 1], planned_tour(basis_set, cycle))
  tour <- new_tour(
    matrix(unclass(basis_set)[, , 1], nrow = dim(basis_set)[1]),
    planned_tour(basis_set, cycle)
  )
  step <- tour(0)
  stop_next <- FALSE # use bool to stop after adding final projection

  while (TRUE) { # need to add final projection after generator runs out, so using break instead of while condition
    new_basis[i] <- step$step <= 0
    projs[, , i] <- step$proj
    if (step$step == -1) break # this signals we have reached the last target
    i <- i + 1
    step <- tour(angle)
  }

  # Trim off extra bases
  projs <- projs[, , seq_len(i), drop = FALSE]
  new_basis <- new_basis[seq_len(i), drop = FALSE]

  attr(projs, "new_basis") <- new_basis
  attr(projs, "data") <- attr(basis_set, "data")
  class(projs) <- c("history_array", class(projs))
  projs
}

#' Draw anaglyphs with base graphics.
#'
#' @param d3 3d numeric matrix giving position of points
#' @param blue blue colour (for right eye)
#' @param red red colour (for left eye)
#' @param cex size of the point to be plotted.  Defaults to 1.
#' @keywords internal
anaglyph <- function(d3, blue, red, cex = 1) {
  d2 <- project3d(d3)

  with(d2, points(right, y, col = blue, pch = 20, cex = cex))
  with(d2, points(left, y, col = red, pch = 20, cex = cex))
}

#' Stereographic projection
#'
#'  Math from http://dogfeathers.com/java/3dproj.html
#'
#' @keywords internal
#' @param d3 3d numeric matrix giving position of points
#' @param length width of plot, mm
#' @param z0 distance from eye to screen, mm
#' @param d half interpupilary distance, mm
project3d <- function(d3, length = par("din")[1] * 25.4, z0 = 300, d = 30) {
  length <- length * 0.3
  x <- d3[, 1] * length
  y <- d3[, 2] * length
  # Squash z dimension a bit more
  z <- (1.5 + d3[, 3]) * length / 2

  d2 <- data.frame(
    left =  (z0 * x - z * d) / (z0 - z),
    right = (z0 * x + z * d) / (z0 - z),
    y =     (z0 * y) / (z0 - z)
  ) / length * 0.5
}


#' Anaglpyh tour path animation.
#'
#' Uses red-blue anaglyphs to display a 3d tour path.  You'll need some red-
#' blue glasses to get much out of this displays!
#'
#' @param blue blue colour (for right eye)
#' @param red red colour (for left eye)
#' @param cex size of the point to be plotted.  Defaults to 1.
#' @param ... other arguments passed on to \code{\link{animate}}
#' @keywords hplot
#' @export
#' @examples
#' animate_stereo(flea[, 1:6])
display_stereo <- function(blue, red, cex = 1, ...) {
  labels <- NULL
  init <- function(data, ...) {
    labels <<- abbreviate(colnames(data), 2)
  }
  render_frame <- function() {
    par(pty = "s", bg = "grey85")
    blank_plot(xlim = c(-1, 1), ylim = c(-1, 1))
  }
  render_transition <- function() {
    # rect(-1, -1, 1, 1, col="#D9D9D9E6", border=NA)
  }
  render_data <- function(data, proj, geodesic) {
    render_frame()
    anaglyph(data %*% proj, blue, red, cex = cex)

    axes <- project3d(proj)
    with(axes, {
      segments(0, 0, right, y, col = blue)
      segments(0, 0, left, y, col = red)
      text(right, y, col = blue, label = labels)
      text(left, y, col = red, label = labels)
    })
  }



  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}


#' @rdname display_stereo
#' @inheritParams animate
#' @export
animate_stereo <- function(data, tour_path = grand_tour(3), blue = rgb(0, 0.91, 0.89), red = rgb(0.98, 0.052, 0), ...) {
  animate(
    data = data, tour_path = tour_path,
    display = display_stereo(blue, red, ...),
    ...
  )
}

# Compute maximum distance from data to origin.
#
# @keywords internal
max_dist <- function(data, center = FALSE) {
  max(sqrt(rowSums(data^2)))
}

compute_half_range <- function(half_range, data, center) {
  if (!is.null(half_range)) {
    return(half_range)
  }

  if (center) {
    data <- center(data)
  }
  half_range <- max_dist(data, center)
  message("Using half_range ", format(half_range, digits = 2))
  half_range
}

#' Display 2D tour projections displayed separately by groups
#'
#' This function is designed to allow comparisons across multiple groups,
#'  especially for examining things like two (or more) different models
#'  on the same data. The primary display is a scatterplot, with lines or
#'  contours overlaid.
#'
#' @param axes position of the axes: center, bottomleft or off
#' @param centr if TRUE, centers projected data to (0,0).  This pins the
#'  center of data cloud and make it easier to focus on the changing shape
#'  rather than position.
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param edges A two column integer matrix giving indices of ends of lines.
#' @param col color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".
#' @param pch shape of the point to be plotted.  Defaults to 20.
#' @param cex size of the point to be plotted.  Defaults to 1.
#' @param group_by variable to group by. Must have less than 25 unique values.
#' @param plot_xgp if TRUE, plots points from other groups in light grey
#' @param edges.col colour of edges to be plotted, Defaults to "black"
#' @param edges.width line width for edges, default 1
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @param shapeset numbers corresponding to shapes in base R points, to use for mapping
#'        categorical variable to shapes, default=c(15:17, 23:25)
#' @param axislablong text labels only for the long axes in a projection, default FALSE
#' @param ...  other arguments passed on to \code{\link{animate}} and
#'   \code{\link{display_groupxy}}
#' @export
#' @examples
#' animate_groupxy(flea[, 1:6], col = flea$species,
#'   pch = flea$species, group_by = flea$species)
#' animate_groupxy(flea[, 1:6], col = flea$species,
#'   pch = flea$species, group_by = flea$species,
#'   plot_xgp = FALSE)
#' # Edges example
#' x <- data.frame(x1=runif(10, -1, 1), x2=runif(10, -1, 1), x3=runif(10, -1, 1))
#' x$cl <- factor(c(rep("A", 3), rep("B", 3), rep("C", 4)))
#' x.edges <- cbind(from=c(1,2, 4,5, 7,8,9), to=c(2,3, 5,6, 8,9,10))
#' x.edges.col <- factor(c(rep("A", 2), rep("B", 2), rep("C", 3)))
#' animate_groupxy(x[,1:3], col=x$cl, group_by=x$cl, edges=x.edges, edges.col=x.edges.col)
display_groupxy <- function(centr = TRUE, axes = "center", half_range = NULL,
                            col = "black", pch = 20, cex = 1,
                            edges = NULL, edges.col = "black", edges.width=1,
                            group_by = NULL, plot_xgp = TRUE,
                            palette = "Zissou 1", shapeset=c(15:17, 23:25),
                            axislablong = FALSE, ...) {
  labels <- NULL

  # If colors are a variable, convert to colors
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    col <- mapColors(col, palette)
  }
  if (is.factor(edges.col) | !areColors(edges.col)) {
    edges.gps <- edges.col
    edges.col <- mapColors(edges.col, palette)
  }
  # If shapes are a variable, convert shapes
  if (is.factor(pch)) {
    shapes <- mapShapes(pch, shapeset)
  } else {
    shapes <- pch
  }

  init <- function(data) {
    half_range <<- compute_half_range(half_range, data, centr)
    labels <<- abbreviate(colnames(data), 3)
  }

  if (!is.null(edges)) {
    if (!is.matrix(edges) && ncol(edges) >= 2) {
      stop("Edges matrix needs two columns, from and to, only.")
    }
  }

  render_frame <- function() {
    par(pty = "s", mar = rep(0.1, 4))
  }
  render_transition <- function() {
    rect(-1, -1, 1, 1, col = "#FFFFFFE6", border = NA)
  }
  render_data <- function(data, proj, geodesic) {
    gps <- unique(group_by)
    ngps <- length(gps)

    edge.gps <- unique(edges.col)

    if (ngps > 24) {
      stop("Please choose a group with 24 or less levels.\n")
    }
    grid <- ceiling(sqrt(ngps + 1))
    par(mfrow = c(grid, grid))

    # Render projected points
    x <- data %*% proj
    if (centr) x <- center(x)
    x <- x / half_range

    blank_plot(xlim = c(-1, 1), ylim = c(-1, 1))
    draw_tour_axes(proj, labels, limits = 1, axes, longlabels=axislablong)
    # add a legend, only if a variable was used
    if (is.factor(gps)) {
      numcol <- unique(col)
      if (length(numcol) > 1) {
        legend("topright", legend=unique(gps),
               col=numcol, pch=15)
      }
    }
    if (is.factor(pch)) {
      numpch <- unique(shapes)
      if (length(numpch) > 1) {
        legend("bottomright", legend=unique(pch),
               col="black", pch=unique(shapes))
      }
    }

    if (ngps < 2) {
      points(x, col = col, pch = shapes, cex = cex, new = FALSE)
      if (!is.null(edges)) {
        segments(
          x[edges[, 1], 1], x[edges[, 1], 2],
          x[edges[, 2], 1], x[edges[, 2], 2],
          col = edges.col,
          lwd = edges.width
        )
      }
    }
    else {
      for (i in 1:ngps) {
        x.sub <- x[group_by == gps[i], ]
        col.sub <- if (length(col) == nrow(x)) col[group_by == gps[i]] else col
        pch.sub <- if (length(shapes) == nrow(x)) shapes[group_by == gps[i]] else shapes
        cex.sub <- if (length(cex) == nrow(x)) cex[group_by == gps[i]] else cex

        blank_plot(xlim = c(-1, 1), ylim = c(-1, 1))
        if (plot_xgp) {
          points(x[group_by != gps[i], ],
            col = "#DEDEDEDE", new = FALSE,
            shapes = if (length(shapes) > 1) shapes[group_by != gps[i]] else shapes,
            cex = if (length(cex) > 1) cex[group_by != gps[i]] else cex
          )
        }
        points(x.sub, col = col.sub, pch = pch.sub, cex = cex.sub, new = FALSE)

        # This is potentially fragile!
        # It assumes the edge colour groups are the same as the point colour groups
        if (!is.null(edges)) {
          segments(
            x[edges[edges.gps == gps[i], 1], 1], x[edges[edges.gps == gps[i], 1], 2],
            x[edges[edges.gps == gps[i], 2], 1], x[edges[edges.gps == gps[i], 2], 2],
            col = edges.col[edges.gps == gps[i]],
            lwd = edges.width
          )
        }
      }
    }
  }

  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}

#' @rdname display_groupxy
#' @inheritParams animate
#' @export
animate_groupxy <- function(data, tour_path = grand_tour(), ...) {
  animate(data, tour_path, display_groupxy(...))
  par(mfrow = c(1, 1))
}

#' Flea beatle measurements
#'
#' This data is from a paper by A. A. Lubischew, "On the Use of Discriminant
#' Functions in Taxonomy", Biometrics, Dec 1962, pp.455-477. Data is
#' standardized, and original units are in flea_raw.
#'
#' \itemize{
#'   \item tars1, width of the first joint of the first tarsus in microns
#'   (the sum of measurements for both tarsi)
#'   \item tars2, the same for the second joint
#'   \item head, the maximal width of the head between the external edges of the
#'   eyes in 0.01 mm
#'   \item ade1, the maximal width of the aedeagus in the fore-part in microns
#'   \item ade2, the front angle of the aedeagus ( 1 unit = 7.5 degrees)
#'   \item ade3, the aedeagus width from the side in microns
#'   \item species, which species is being examined - concinna, heptapotamica, heikertingeri
#' }
#'
#' @name Flea measurements
#' @aliases flea, flea_raw
#' @docType data
#' @format A 74 x 7 numeric array
#' @keywords datasets
#' @examples
#'
#' head(flea)
#' animate_xy(flea[, -7])
#' animate_xy(flea[, -7], col = flea[, 7])
"flea"

#' Turnable laser measurements from Bellcore
#'
#'
#' This data came from an investigation of an experimental laser at Bellcore.
#' It was a tunable laser, in the sense that both its wavelength and power
#' output were controllable.
#'
#' Rotation helped the experimental physicists to characterize the laser, which
#' turned out not to be a very good one, due to its unstable operating region.
#'
#' This data initially came to the statistics research group when Janette
#' Cooper asked Paul Tukey to help her analyze the data she had collected to
#' describe the laser.
#'
#' \itemize{
#'  \item ifront, current applied to the front of the laser
#'  \item iback, current applied to the back of the laser
#'  \item power, output power
#'  \item lambda, output wavelength
#' }
#'
#'
#' @name Laser measurements
#' @aliases laser
#' @docType data
#' @format A 64 x 4 numeric array
#' @keywords datasets
#' @examples
#'
#' head(laser)
#' animate_xy(laser[, -4])
NULL

#' Olive oil samples from Italy
#'
#' This data is from a paper by Forina, Armanino, Lanteri, Tiscornia (1983)
#' Classification of Olive Oils from their Fatty Acid Composition, in Martens
#' and Russwurm (ed) Food Research and Data Anlysis. We thank Prof. Michele
#' Forina, University of Genova, Italy for making this dataset available.
#'
#' \itemize{
#'   \item region Three super-classes of Italy: North, South and the
#'     island of Sardinia
#'   \item area Nine collection areas: three from North, four from South
#'     and 2 from Sardinia
#'   \item palmitic, palmitoleic, stearic, oleic, linoleic, linolenic,
#'     arachidic, eicosenoic fatty acids percent x 100
#' }
#'
#' @name Olive oil measurements
#' @aliases olive
#' @docType data
#' @format A 572 x 10 numeric array
#' @keywords datasets
#' @examples
#'
#' head(olive)
#' animate_xy(olive[, c(7, 9, 10)])
#' animate_xy(olive[, c(7, 9, 10)], col = olive[, 1])
NULL

#' Monthly ozone measurements over Central America
#'
#'
#' This data set is a subset of the data from the 2006 ASA Data expo challenge.
#' The data are monthly ozone averages on a very coarse 24 by 24 grid covering
#' Central America, from Jan 1995 to Dec 2000. The data is stored in a 3d area
#' with the first two dimensions representing latitude and longitude, and the
#' third representing
#' time.
#'
#' @name Ozone measurements
#' @aliases ozone
#' @docType data
#' @format A 24 x 24 x 72 numeric array
#' @keywords datasets
#' @examples
#'
#' example(display_image)
NULL

#' Ratings of different locations across North America
#'
#'
#' The "places data" were distributed to interested ASA members a few years ago
#' so that they could apply contemporary data analytic methods to describe
#' these data and then present results in a poster session at the ASA annual
#' conference.  Latitude and longitude have been added by Paul Tukey.
#'
#' ____________________________________________________________________
#'
#' The first dataset is taken from the Places Rated Almanac, by Richard Boyer
#' and David Savageau, copyrighted and published by Rand McNally. This book
#' order (SBN) number is 0-528-88008-X, and it retails for $14.95 .  The data
#' are reproduced on disk by kind permission of the publisher, and with the
#' request that the copyright notice of Rand McNally, and the names of the
#' authors appear in any paper or presentation using these data.
#'
#' The nine rating criteria used by Places Rated Almanac are: Climate and
#' Terrain Housing Health Care and Environment Crime Transportation Education
#' The Arts Recreation Economics
#'
#' For all but two of the above criteria, the higher the score, the better.
#' For Housing and Crime, the lower the score the better.
#'
#' The scores are computed using the following component statistics for each
#' criterion (see the Places Rated Almanac for details):
#'
#' Climate and Terrain: very hot and very cold months, seasonal temperature
#' variation, heating- and cooling-degree days, freezing days, zero-degree
#' days, ninety-degree days.
#'
#' Housing: utility bills, property taxes, mortgage payments.
#'
#' Health Care and Environment: per capita physicians, teaching hospitals,
#' medical schools, cardiac rehabilitation centers, comprehensive cancer
#' treatment centers, hospices, insurance/hospitalization costs index,
#' flouridation of drinking water, air pollution.
#'
#' Crime: violent crime rate, property crime rate.
#'
#' Transportation: daily commute, public transportation, Interstate highways,
#' air service, passenger rail service.
#'
#' Education: pupil/teacher ratio in the public K-12 system, effort index in
#' K-12, accademic options in higher education.
#'
#' The Arts: museums, fine arts and public radio stations, public television
#' stations, universities offering a degree or degrees in the arts, symphony
#' orchestras, theatres, opera companies, dance companies, public libraries.
#'
#' Recreation: good restaurants, public golf courses, certified lanes for
#' tenpin bowling, movie theatres, zoos, aquariums, family theme parks,
#' sanctioned automobile race tracks, pari-mutuel betting attractions, major-
#' and minor- league professional sports teams, NCAA Division I football and
#' basketball teams, miles of ocean or Great Lakes coastline, inland water,
#' national forests, national parks, or national wildlife refuges, Consolidated
#' Metropolitan Statistical Area access.
#'
#' Economics: average household income adjusted for taxes and living costs,
#' income growth, job growth.
#'
#'
#' @name Places Ratings
#' @aliases places
#' @docType data
#' @format A 329 x 14 numeric array
#' @keywords datasets
#' @examples
#'
#' head(places)
#' animate_xy(places[, 1:9])
NULL

#' Rat CNS Gene Expression
#'
#'
#' Columns:
#'
#' e11  e13  e15  e18  e21   p0   p7  p14    a class1 class2
#' \itemize{
#'   \item e11, an ebryonic timepoint from the original data with the number
#'      corresponding to the day
#'   \item e13, an ebryonic timepoint from the original data with the number
#'      corresponding to the day
#'   \item e15, an ebryonic timepoint from the original data with the number
#'      corresponding to the day
#'   \item e18, an ebryonic timepoint from the original data with the number
#'      corresponding to the day
#'   \item e21, an ebryonic timepoint from the original data with the number
#'     corresponding to the day
#'   \item p0, a postnatal timpoint from the original data with the number
#'     corresponding to the day
#'   \item p7, a postnatal timpoint from the original data with the number
#'     corresponding to the day
#'   \item p14, a postnatal timpoint from the original data with the number
#'     corresponding to the day
#'   \item a, a postnatal timpoint from the original data. It is equivalent to
#'     p90.
#'   \item class1, is the high-level class: its range is 1:4
#'   \item class2, breaks down the high-level classes, so its range is 1:14
#' }
#'
#' Rows: Each case is a gene (or gene family?) And each cell is the gene
#' expression level for that gene at time t, averaging a few measured values
#' and normalizing using the maximum expression value for that gene.
#'
#' Reference (available on the web at pnas.org): Large-scale temporal gene
#' expression mapping of central nervous system development by X. Wen, S.
#' Fuhrman, G. S. Michaels, D. B. Carr, S. Smith, J. L. Barker, R. Somogyi in
#' the Proceedings of the National Academy of Science, Vol 95, pp. 334-339,
#' January 1998
#'
#'
#' @name Rat CNS
#' @aliases ratcns
#' @docType data
#' @format A 112 x 11 numeric array
#' @references https://www.pnas.org
#' @keywords datasets
#' @examples
#'
#' head(ratcns)
#' animate_xy(ratcns[, 1:8], col = ratcns[, 10])
NULL

#' Tropical Atmosphere Ocean data
#'
#'
#' This is a subset of data taken from the NOAA web site
#' https://www.pmel.noaa.gov/tao/. The data is generated from recording
#' instruments on a grid of buoys laid out over the Pacific Ocean.  The grid
#' was setup to monitor El Nino and La Nina events. This subset contains
#' measurements from 5 locations (0deg/110W, 2S/110W, 0deg/95W,2S/95W,5S/95W)
#' and two time points Nov-Jan 1993 (normal), 1997 (El Nino). There are missing
#' values in this data set, which need to be removed, or imputed before running
#' a tour.
#'
#'
#' @name Tropical Atmosphere Ocean
#' @aliases tao
#' @docType data
#' @format A 736 x 8 numeric array
#' @references https://www.pmel.noaa.gov/tao/
#' @keywords datasets
NULL

#' Saved history of guided tour with holes
#'
#' This data was generated from the following code:
#' set.seed(2020)
#' t1 <- save_history(flea[, 1:6], guided_tour(holes()), max = 100)
#' attr(t1, "class") <- NULL
#' And used as an example for search_polish() to start
#' optimising from the best projection from search_geodesic.
#' t1 is a 3D array or 2D projections.
#'
#' @name t1
#' @aliases t1
#' @docType data
#' @keywords datasets
NULL

#' Scagnostic indexes.
#'
#' Compute the scagnostic measures from the cassowaryr package
#' @export
stringy <- function(){
  function(mat){
    cassowaryr::sc_stringy(mat[,1], mat[,2])
  }
}


#' Maximum and total information coefficient index.
#'
#' Compute the maximum and total information coefficient indexes,
#' see \code{minerva::mine}.
#'
#' @rdname indexes
#' @export
MIC <- function(){
  function(mat){
    minerva::mine(mat[,1], mat[,2], alpha = 0.3,  est = "mic_e")$MIC
  }
}

#' @rdname indexes
#' @export
TIC <- function(){
  function(mat){
    minerva::mine(mat[,1], mat[,2], est = "mic_e", alpha = 0.3)$TIC
  }
}

#' Distance correlation index.
#'
#' Computes the distance correlation based index on 2D projections of the data.
#' \code{dcor2d_2} uses the faster implementation of the distance correlation
#' for bivariate data, see \code{energy::dcor2d}.
#'
#' @keywords hplot
#' @importFrom stats na.omit
#' @export
#' @rdname dcor
dcor2d <- function() {
  function(mat) {
    xy <- na.omit(data.frame(x = mat[, 1], y = mat[, 2]))
    measure <- with(xy, energy::dcor(x, y))
    return(measure)
  }
}

#' @rdname dcor
#' @export
dcor2d_2 <- function() {
  function(mat) {
    xy <- na.omit(data.frame(x = mat[, 1], y = mat[, 2]))
    measure <- with(xy, energy::dcor2d(x, y, type = "U"))
    return(measure)
  }
}

#' Spline/loess based index.
#'
#' Compares the variance in residuals of a fitted
#' spline/loess model to the overall variance to find
#' functional dependence in 2D projections
#' of the data.
#'
#' @keywords hplot
#' @importFrom stats residuals var
#' @rdname spline-loess
#' @export
splines2d <- function() {
  function(mat) {
    mat <- as.data.frame(mat)
    colnames(mat) <- c("x", "y")
    kx <- ifelse(length(unique(mat$x[!is.na(mat$x)])) < 20, 3, 10)
    ky <- ifelse(length(unique(mat$y[!is.na(mat$y)])) < 20, 3, 10)
    mgam1 <- mgcv::gam(y ~ s(x, bs = "cr", k = kx), data=mat)
    mgam2 <- mgcv::gam(x ~ s(y, bs = "cr", k = ky), data=mat)
    measure <- max(1 - var(residuals(mgam1), na.rm = T) / var(mat$y, na.rm = T),
                   1 - var(residuals(mgam2), na.rm = T) / var(mat$x, na.rm = T))
    return(measure)
  }
}

#' @rdname spline-loess
#' @importFrom stats loess
#' @export
loess2d <- function() {
  function(mat) {
    mat <- as.data.frame(mat)
    colnames(mat) <- c("x", "y")
    loess_fit <- loess(y ~ x, data = mat, span = 0.05)
    loess_fit2 <- loess(x ~ y, data = mat, span = 0.05)
    measure <- max(1 - var(residuals(loess_fit), na.rm = T) / var(mat$y, na.rm = T),
                   1 - var(residuals(loess_fit2), na.rm = T) / var(mat$y, na.rm = T)
    )
    return(measure)
  }
}


#' Normality index.
#'
#' Compares the similarity between the projected distribution and a normal distribution.
#' \itemize{
#' \item norm_bin: compares the count in 100 histogram bins
#' \item norm_kol: compares the cdf based on the Kolmogorov–Smirnov test (KS test)
#' }
#' @param nr The number of rows in the target matrix
#'
#' @keywords hplot
#' @importFrom stats rnorm
#' @export
norm_bin <- function(nr) {
  norm <- rnorm(nr)

  function(mat) {
    stopifnot(length(norm) == nrow(mat))
    norm <- scale(norm)
    mat_s <- scale(mat)
    norm_bin_count <- ash::bin1(norm, c(min(mat), max(mat)), 100)$nc
    mat_bin_count <- ash::bin1(mat_s, c(min(mat), max(mat)), 100)$nc
    diff <- sum((mat_bin_count - norm_bin_count)^2) / nrow(mat)

    diff
  }
}


#' @export
#' @rdname norm_bin
#' @examples
#' # manually compute the norm_kol index
#' # create the index function
#' set.seed(123)
#' index <- norm_kol(nrow(flea[, 1:3]))
#' # create the projection
#' proj <- matrix(c(1, 0, 0), nrow = 3)
#' # pre-process the example data
#' flea_s <- sphere_data(flea[, 1:3])
#' # produce the index value
#' index(flea_s %*% proj)
norm_kol <- function(nr) {
  norm <- rnorm(nr)

  function(mat) {
    stopifnot(length(norm) == nrow(mat))
    as.numeric(stats::ks.test(mat, norm)$statistic)
  }
}

#' Holes index.
#'
#' Calculates the holes index. See Cook and Swayne (2007)
#' Interactive and Dynamic Graphics for Data Analysis for equations.
#'
#' @keywords hplot
#' @export
holes <- function() {
  function(mat) {
    n <- nrow(mat)
    d <- ncol(mat)

    num <- 1 - 1 / n * sum(exp(-0.5 * rowSums(mat^2)))
    den <- 1 - exp(-d / 2)

    num / den
  }
}

#' Central mass index.
#'
#' Calculates the central mass index.  See Cook and Swayne (2007)
#' Interactive and Dynamic Graphics for Data Analysis for equations.
#'
#' @keywords hplot
#' @export
cmass <- function() {
  function(mat) {
    n <- nrow(mat)
    d <- ncol(mat)

    num <- 1 - 1 / n * sum(exp(-0.5 * rowSums(mat^2)))
    den <- 1 - exp(-d / 2)

    1 - num / den
  }
}

#' Skewness index.
#'
#' Calculates the skewness index. See Cook, Buja and Cabrera (1993)
#' Projection pursuit indexes based on orthonormal function expansions
#'  for equations.
#'
#' @keywords hplot
#' @export
skewness <- function() {
  function(mat) {
    n <- nrow(mat)
    d <- ncol(mat)

    idx <- mean(rowSums(mat * exp(-0.5 * mat^2)))

    idx
  }
}

#' LDA projection pursuit index.
#'
#' Calculate the LDA projection pursuit index.  See Cook and Swayne (2007)
#' Interactive and Dynamic Graphics for Data Analysis for equations.
#'
#' @param cl class to be used.  Such as "color"
#' @keywords hplot
#' @export
lda_pp <- function(cl) {
  if (length(unique(cl)) == 0) {
    stop("You need to select the class variable!")
  }
  if (length(unique(cl)) == 1) {
    stop("LDA index needs at least two classes!")
  }

  function(mat) {
    if (ncol(mat) > 1) {
      fit <- stats::manova(mat ~ cl)

      1 - summary(fit, test = "Wilks")$stats[[3]]
    } else {
      summary(stats::aov(mat ~ cl))[[1]][1,4]
    }
  }
}

#' PDA projection pursuit index.
#'
#' Calculate the PDA projection pursuit index.  See Lee and Cook (2009)
#' A Projection Pursuit Index for Large p, Small n Data
#'
#' @param cl class to be used.  Such as "color"
#' @param lambda shrinkage parameter (0 = no shrinkage, 1 = full shrinkage)
#' @keywords hplot
#' @export
pda_pp <- function(cl, lambda = 0.2) {
  if (length(unique(cl)) == 0) {
    stop("You need to select the class variable!")
  }
  if (length(unique(cl)) < 2) {
    stop("PDA index needs at least two classes!")
  }

  # Convert class to sequential integers, and sort
  cl.i <- as.integer(as.factor(cl))
  cl.sort <- order(cl.i)
  cl <- cl.i[cl.sort]

  function(mat) {
    mat <- as.matrix(mat)

    # Reorder so classes are adjacent
    if (ncol(mat) == 1) {
      mat <- as.matrix(mat[cl.sort, ])
    } else {
      mat <- mat[cl.sort, ]
    }

    ngroup <- table(cl.i) # the obs. no. in each class, stored in table
    groups <- length(ngroup) # no. of classes
    gname <- names(ngroup) # names of classes, now is integer 1,2,3...

    .CalIndex(
      nrow(mat), ncol(mat), groups, mat, cl, gname,
      as.integer(ngroup), lambda
    )
  }
}

# @param n no. of obs,
# @param p no. of variables,
# @param groups no. of classes,
# @param fvals sorted matrix
# @param groupraw sorted class label of whole matrix
# @param gname names of classes, now is integer 1,2,3...;
# @param ngroup the obs. no. in each class
# @param lambda shrinkage parameter (0 = no shrinkage, 1 = full shrinkage)
.CalIndex <- function(n, p, groups, fvals, groupraw, gname, ngroup, lambda) {
  # int i, j, k, right, left
  g <- groups
  mean <- matrix(rep(0, g * p), g, p)
  ovmean <- matrix(rep(0, p), p)
  group <- matrix(rep(0, n), n) # the class label of each obs

  right <- n - 1
  left <- 0

  group <- groupraw

  val <- 0

  # Calculate mean for within class and the overall mean
  for (i in 1:n) {
    for (j in 1:p) {
      mean[group[i], j] <- mean[group[i], j] + fvals[i, j] / ngroup[group[i]]
      ovmean[j] <- ovmean[j] + fvals[i, j] / n
    }
  }

  cov <- matrix(rep(0, p * p), p, p)
  tempcov <- matrix(rep(0, p * p), p, p)

  for (i in 1:n) {
    for (j in 1:p) {
      for (k in 1:j) {
        cov[k, j] <- cov[k, j] + (1 - lambda) * (fvals[i, j] -
          mean[group[i], j]) * (fvals[i, k] - mean[group[i], k])
        cov[j, k] <- cov[k, j]
      }
    }
  }

  for (j in 1:p) {
    cov[j, j] <- cov[j, j] + lambda * n
  }

  tempcov <- cov

  # =========================================================
  if (ncol(tempcov) > 1) {
    pivot <- matrix(rep(0, p), p)
    det <- 0

    val <- .ludcomp(tempcov, p, pivot)
  }
  else {
    val <- as.numeric(tempcov)
  }
  # ===========================================================

  for (j in 1:p) {
    for (k in 1:p) {
      for (i in 1:g) {
        cov[j, k] <- cov[j, k] + (1 - lambda) * ngroup[i] *
          (mean[i, j] - ovmean[j]) * (mean[i, k] - ovmean[k])
      }
    }
  }

  tempcov <- cov
  if (ncol(tempcov) > 1) {
    tempval <- .ludcomp(tempcov, p, pivot)
  } else {
    tempval <- tempcov
  }

  if (tempval < 0.00000001) {
    val <- 0
    print("ZERO VARIANCE!")
  }
  else {
    val <- 1 - val / tempval
  }

  return(val)
}

# ==============================================================================
.ludcomp <- function(a, n, pivot) {
  det <- 1
  temp <- 0
  s <- matrix(rep(0, n), n)
  for (i in 1:n) {
    s[i] <- a[i, 1 + 1]
    for (j in 2:n) {
      if (s[i] < a[i, j]) {
        s[i] <- a[i, j]
      }
    }
  }

  for (k in 1:(n - 1)) {
    for (i in k:n)
    {
      temp <- abs(a[i, k] / s[i])
      if (i == k) {
        c <- temp
        pivot[k] <- i
      }
      else if (c < temp) {
        c <- temp
        pivot[k] <- i
      }
    }

    if (pivot[k] != k) {
      det <- det * (-1)
      for (j in k:n) {
        temp <- a[k, j]
        a[k, j] <- a[pivot[k], j]
        a[pivot[k], j] <- temp
      }
      temp <- s[k]
      s[k] <- s[pivot[k]]
      s[pivot[k]] <- temp
    }

    for (i in (k + 1):n) {
      temp <- a[i, k] / a[k, k]
      a[i, k] <- temp
      for (j in (k + 1):n) {
        a[i, j] <- a[i, j] - temp * a[k, j]
      }
    }
    det <- det * a[k, k]
  }

  det <- det * a[n, n]

  return(det)
}

#' A jellyfish optimiser for projection pursuit guided tour
#'
#' @param current starting projection, a list of basis of class "multi-bases"
#' @param index index function
#' @param tries the counter of the outer loop of the opotimiser
#' @param max.tries the maximum number of iteration before giving up
#' @param ... other arguments being passed into the \code{search_jellyfish()}
#' @param verbose whether to print out the progress messages
#' @keywords optimize
#' @importFrom stats runif
#' @rdname jellyfish
#' @export
#' @examples
#' library(dplyr)
#' res <- animate_xy(flea[, 1:6], guided_tour(lda_pp(cl = flea$species),
#' search_f = search_jellyfish))
#' bases <- res |> filter(loop == 1) |> pull(basis) |> check_dup(0.1)
#' animate_xy(data = flea[,1:6], tour_path = planned_tour(bases), col = flea$species)
search_jellyfish <- function(current, index, tries, max.tries = 50, verbose = FALSE, ...) {
  rcd_env <- parent.frame(n = 4)
  if (is.null(rcd_env[["record"]])) rcd_env <- parent.frame(n = 1)
  best_jelly <- current[[attr(current, "best_id")]]
  last_tries <- tries - 1
  if (tries == 2){
    current_idx <- sapply(current, index)
  } else{
    current_idx <- rcd_env[["record"]] |>
      dplyr::filter(tries == last_tries) |>
      dplyr::pull(index_val)
  }

  c_t = abs((1 - tries / max.tries) * (2 * runif(1) - 1))

  if (c_t >= 0.5) {
    trend = best_jelly - 3 * runif(1) * Reduce("+", current) / length(current) # eq 9
    target = lapply(current, function(x) {
      orthonormalise(x + runif(1) * trend)
    })
  } else if (runif(1) > (1 - c_t)) {
    # type A passive
    target = lapply(current, function(x) {
      orthonormalise(x + 0.1 * runif(1))
    }) # eq 12
  } else {
    # type B active
    # generate random jelly fish j and its index value
    update_typeB <- function(current, idx){
      idx_jelly_i <- current_idx[idx]; jelly_i <- current
      jelly_j_no <- sample(1:length(current_idx), 1)
      idx_jelly_j <- current_idx[jelly_j_no]
      current_basis <- rcd_env[["record"]] |>
        dplyr::filter(tries == last_tries) |> dplyr::pull(basis)
      jelly_j <- current_basis[[jelly_j_no]]
      if (idx_jelly_i > idx_jelly_j) {
        new_i = orthonormalise(jelly_i + runif(1) * (jelly_i - jelly_j))
      } else {
        new_i = orthonormalise(jelly_i + runif(1) * (jelly_j - jelly_i)) # eq 16
      }
      return(new_i)
    }
    target <- mapply(update_typeB, current, 1:length(current), SIMPLIFY = FALSE)
    #target = lapply(current, function(i) {update_typeB(i, current)})
  }

  target <- mapply(correct_orientation, current, target, SIMPLIFY = FALSE)
  target_idx <- sapply(target, index)

  # check for NA for index computation (skinny gives NA if the alphahull can't be computed)
  na_idx <- which(is.na(target_idx))
  if (length(na_idx) > 0) {
    target[na_idx] <- current[na_idx]
    target_idx[na_idx] <- current_idx[na_idx]
  }

  # if the target is worse than current, use current
  worse_id <- current_idx > target_idx
  target[worse_id] <- current[worse_id]
  target_idx[worse_id] <- current_idx[worse_id]

  best_id <- which.max(target_idx)
  if (verbose) message("Target: ", sprintf("%.3f", max(target_idx)))
  attr(target, "best_id") <- best_id
  class(target) <- c("multi-bases", class(target))

  rcd_env[["record"]] <- dplyr::add_row(
    rcd_env[["record"]], basis = unclass(target),
    index_val = target_idx, info = "jellyfish_update",
    tries = tries, method = "search_jellyfish", alpha = NA,
    loop = 1:length(target)
  )
  rcd_env[["record"]] <- dplyr::mutate(
    rcd_env[["record"]],
    info = ifelse(tries == max(tries) & loop == best_id, "current_best", info)
  )

  if (tries >= max.tries) {
    print_final_proj(target[[attr(target, "best_id")]])
    rcd_env[["record"]] <- dplyr::mutate(
      rcd_env[["record"]],
      id = dplyr::row_number()
    )
    NULL
  } else{
    tries <- tries + 1
    return(list(target = target))
  }
}

#' @param bases a list of bases extracted from the data collection object, see examples
#' @param min_dist the minimum distance between two bases
#' @rdname jellyfish
#' @export
check_dup <- function(bases, min_dist) {
  res <- list()
  res[[1]] <- bases[[1]]
  i <- 1; j <- 2
  while(j <= length(bases)){
    if (proj_dist(bases[[i]], bases[[j]]) >= min_dist) {
      res <- c(res, list(bases[[j]]))
      i <- i + 1; j <- j + 1
    } else{
      j <- j + 1
    }
  }
  return(res)
}

globalVariables(c("loop", "basis"))

#' Compute index values for a tour history.
#'
#' @param history list of bases produced by \code{\link{save_history}}
#'   (or otherwise)
#' @param index_f index function to apply to each basis
#' @param data dataset to be projected on to bases
#' @keywords hplot
#' @seealso \code{\link{save_history}} for options to save history
#' @export
#' @examples
#' fl_holes <- save_history(flea[, 1:6], guided_tour(holes()), sphere = TRUE)
#' path_index(fl_holes, holes())
#' ## path_index(fl_holes, cmass())
#'
#' plot(path_index(fl_holes, holes()), type = "l")
#' ## plot(path_index(fl_holes, cmass()), type = "l")
#'
#' \donttest{
#' # Use interpolate to show all intermediate bases as well
#' hi <- path_index(interpolate(fl_holes), holes())
#' hi
#' plot(hi)
#' }
path_index <- function(history, index_f, data = attr(history, "data")) {
  index <- function(proj) {
    index_f(as.matrix(data) %*% proj)
  }

  structure(
    apply(history, 3, index),
    class = "path_index"
  )
}

#' Plot history index with ggplot2.
#'
#' @keywords internal hplot
#' @export
plot.path_index <- function(x, ...) {
    pth <- data.frame(x=seq_along(x), y=unclass(x))
    ggplot2::ggplot(pth, ggplot2::aes(x, y)) +
      ggplot2::geom_line() +
      ggplot2::labs(x = "step", y = "index")
}

#' Compute index value for many histories.
#'
#' This is a convenience method that returns a data frame summarising the
#' index values for multiple tour paths.
#'
#' @keywords internal
#' @param bases_list list of histories produced by \code{\link{save_history}}
#' @param index_f index function to apply to each projection
#' @export
#' @examples
#' # The max.tries is low for satisfying CRAN checks
#' # Increase it for using in practice
#' holes1d <- guided_tour(holes(), 1, max.tries=2)
#' # Perform guided tour 5 times, saving results
#' tries <- replicate(5, save_history(flea[, 1:6], holes1d), simplify = FALSE)
#' # Interpolate between target bases
#' itries <- lapply(tries, interpolate)
#'
#' paths <- paths_index(itries, holes())
#' head(paths)
#'
#' if (require(ggplot2)) {
#'   ggplot(data = paths, aes(x=step, y=value, group = try)) + geom_line()
#'   ## ggplot(data = paths, aes(x=step, y=improvement, group = try)) + geom_line()
#' }
paths_index <- function(bases_list, index_f) {
  indices <- lapply(bases_list, path_index, index_f)
  data.frame(
    try = rep(seq_along(indices), sapply(indices, length)),
    step = unlist(sapply(indices, seq_along)),
    value = unlist(indices),
    improvement = unlist(lapply(indices, function(x) c(0, diff(x))))
  )
}

#' Display tour path with a sage scatterplot
#'
#' Animate a 2D tour path with a sage scatterplot that
#' uses a radial transformation on the projected points to re-allocate
#' the volume projected across the 2D plane.
#'
#' @param axes position of the axes: center, bottomleft or off
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param col color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".
#' @param pch marker for points. Defaults to 20.
#' @param gam scaling of the effective dimensionality for rescaling. Defaults to 1.
#' @param R scale for the radial transformation.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @param axislablong text labels only for the long axes in a projection, default FALSE
#' @param ...  other arguments passed on to \code{\link{animate}} and
#'   \code{\link{display_sage}}
#' @export
#' @examples
#' # Generate uniform samples in a 10d sphere using the geozoo package
#' sphere10 <- geozoo::sphere.solid.random(10)$points
#' # Columns need to be named before launching the tour
#' colnames(sphere10) <- paste0("x", 1:10)
#' # Standard grand tour display, points cluster near center
#' animate_xy(sphere10)
#' # Sage display, points are uniformly distributed across the disk
#' animate_sage(sphere10)
display_sage <- function(axes = "center", half_range = NULL,
                         col = "black", pch = 20, gam = 1, R = NULL,
                         palette = "Zissou 1",
                         axislablong = FALSE, ...) {
  labels <- NULL
  peff <- NULL

  # If colors are a variable, convert to colors
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    col <- mapColors(col, palette)
  }

  init <- function(data) {
    half_range <<- compute_half_range(half_range, data, TRUE)
    R <<- compute_half_range(R, data, TRUE)
    labels <<- abbreviate(colnames(data), 3)
    peff <<- gam * ncol(data)
  }


  render_frame <- function() {
    par(pty = "s", mar = rep(0.1, 4))
    blank_plot(xlim = c(-1, 1), ylim = c(-1, 1))
  }
  render_transition <- function() {
    rect(-1, -1, 1, 1, col = "#FFFFFFE6", border = NA)
  }

  render_data <- function(data, proj, geodesic) {
    draw_tour_axes(proj, labels, 1, axes, longlabels=axislablong)

    # Projecte data and center
    x <- data %*% proj
    x <- center(x)
    # Calculate polar coordinates
    rad <- sqrt(x[, 1]^2 + x[, 2]^2)
    ang <- atan2(x[, 2], x[, 1])
    rad <- pmin(rad, R)
    # transform with cumulative to get uniform distribution in radius
    rad <- cumulative_radial(rad, R, peff)
    # square-root is the inverse of the cumulative of a uniform disk (rescaling to maximum radius = 1)
    rad <- sqrt(rad)
    # transform back to x, y coordinates
    x[, 1] <- rad * cos(ang)
    x[, 2] <- rad * sin(ang)
    # scale down maximum radius from 1 to fit drawing range
    # also allow to change drawing range with half_range
    x <- 0.9 * (half_range / R) * x
    points(x, col = col, pch = pch)
  }


  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}


#' Calculate radial cumulative distribtuion
#'
#' Calculate fraction of volume of a pD sphere with radius R projected
#' within the projected radius r.
#'
#' @param r projected radius
#' @param R radius of pD sphere
#' @param p input dimension
#' @return fraction of volume
#' @export
cumulative_radial <- function(r, R, p) {
  1 - (1 - (r / R)^2)^(p / 2)
}

#' @rdname display_sage
#' @inheritParams animate
#' @export
animate_sage <- function(data, tour_path = grand_tour(), ...) {
  animate(data, tour_path, display_sage(...))
}

#' Compute Andrews' curves
#'
#' This function takes a numeric vector of input, and returns a function which
#' allows you to compute the value of the Andrew's curve at every point along
#' its path from -pi to pi.
#'
#' @param x input a new parameter
#' @return a function with single argument, theta
#' @export
#' @examples
#' a <- andrews(1:2)
#' a(0)
#' a(-pi)
#' grid <- seq(-pi, pi, length = 50)
#' a(grid)
#'
#' plot(grid, andrews(1:2)(grid), type = "l")
#' plot(grid, andrews(runif(5))(grid), type = "l")
andrews <- function(x) {
  n <- length(x)
  y <- rep(x[1] / sqrt(2), length(t))

  function(t) {
    for (i in seq(2, n, by = 1)) {
      val <- i %/% 2 * t
      y <- y + x[i] * (if (i %% 2 == 0) sin(val) else cos(val))
    }
    y / n
  }
}


#' Andrews' curves tour path animation.
#'
#' Animate a nD tour path with Andrews' curves.  For more details about
#' Andrew's curves, see \code{\link{andrews}}
#'
#' @param col color to be plotted.  Defaults to "black"
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @param ... other arguments passed on to \code{\link{animate}}
#' @seealso \code{\link{animate}} for options that apply to all animations
#' @keywords hplot
#' @export
#' @examples
#' animate_andrews(flea[, 1:6])
#' animate_andrews(flea[, 1:6], grand_tour(d = 3))
#' animate_andrews(flea[, 1:6], grand_tour(d = 6))
#'
#' # It's easy to experiment with different tour paths:
#' animate_andrews(flea[, 1:6], guided_tour(cmass()))
display_andrews <- function(col = "black", palette = "Zissou 1", ...) {
  grid <- NULL
  init <- function(data) {
    grid <<- seq(-pi, pi, length = 50)
  }

  # If colors are a variable, convert to colors
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    col <- mapColors(col, palette)
  }

  render_frame <- function() {
    blank_plot(xlim = c(-pi, pi), ylim = c(-1, 1))
  }
  render_transition <- function() {
    # rect(-pi, -1, pi, 1, col="#FFFFFF", border=NA)
  }
  render_data <- function(data, proj, geodesic) {
    xd <- data %*% proj
    xd <- rescale(xd)

    values <- lapply(seq_len(nrow(xd)), function(i) {
      rbind(
        cbind(grid, andrews(xd[i, ])(grid)),
        NA, NA
      )
    })

    render_frame()
    segments(-pi, 0, pi, 0)
    if (length(col) == 1) {
      col <- rep(col, nrow(data))
    }
    clrs <- rep(col, rep(52, nrow(data)))
    nclrs <- unique(clrs)
    ldat <- do.call("rbind", values)
    if (length(nclrs) > 1) {
      for (i in 1:length(nclrs)) {
        lines(ldat[clrs == nclrs[i], ], col = nclrs[i])
      }
    }
    else {
      lines(ldat)
    }
  }

  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}


#' @rdname display_andrews
#' @inheritParams animate
#' @export
animate_andrews <- function(data, tour_path = grand_tour(3),
                            col = "black", ...) {
  animate(
    data = data, tour_path = tour_path,
    display = display_andrews(col, ...), ...
  )
}

#' Search very locally to find slightly better projections to polish a broader search.

#' @param current the current projection basis
#' @param alpha the angle used to search the target basis from the current basis
#' @param index index function
#' @param tries the counter of the outer loop of the opotimiser
#' @param polish_max_tries maximum number of iteration before giving up
#' @param cur_index the index value of the current basis
#' @param n_sample number of samples to generate
#' @param polish_cooling percentage of reduction in polish_alpha when no better basis is found
#' @param ... other arguments being passed into the \code{search_polish()}
#' @keywords optimize
#' @export
#' @examples
#' data(t1)
#' best_proj <- t1[, , dim(t1)[3]]
#' attr(best_proj, "data") <- NULL
#' best_proj <- unclass(drop(best_proj))
#' animate_xy(
#'   flea[, 1:6],
#'   guided_tour(holes()),
#'     search_f = search_polish(
#'          polish_max_tries = 5),
#'   start = best_proj
#' )
search_polish <- function(current, alpha = 0.5, index, tries, polish_max_tries = 30,
                          cur_index = NA, n_sample = 100, polish_cooling = 1, ...) {
  if (is.na(cur_index)) cur_index <- index(current)
  try <- 1

  while (try <= polish_max_tries) {
    # could use replicate here
    basis <- lapply(1:n_sample, function(x) {
      tibble::tibble(basis = list(basis_nearby(current,
        alpha = alpha
      )))
    })
    polish <- dplyr::mutate(dplyr::bind_rows(basis),
      index_val = vapply(basis, function(x) index(x), double(1)),
      alpha = round(alpha, 4), tries = tries, info = "polish",
      loop = try, method = "search_polish"
    )

    # could use which.max
    best_row <- dplyr::filter(polish, index_val == max(index_val))
    best_row <- dplyr::mutate(best_row, info = "loop_best", method = "search_polish")

    rcd_env <- parent.frame(n = 4)
    if (is.null(rcd_env[["record"]])) rcd_env <- parent.frame(n = 1)
    rcd_env[["record"]] <- dplyr::bind_rows(rcd_env[["record"]], polish, best_row)
    rcd_env[["record"]] <- dplyr::mutate(
      rcd_env[["record"]],
      id = dplyr::row_number()
    )

    if (best_row$index_val > cur_index) {
      polish_dist <- proj_dist(current, best_row$basis[[1]])
      polish_pdiff <- (best_row$index_val - cur_index) / cur_index

      # check condition 1: the two bases can't be too close

      if (polish_dist < 1e-3) {
        message("The new basis is too close to the current one!")
        message("current basis: ", current, "cur_index: ", sprintf("%.3f", cur_index))
        return(list(target = current, alpha = alpha))
      }

      # check condition 2: there needs to be certain improvement

      if (polish_pdiff < 1e-5) {
        message("The improvement is too small!")
        message("current basis: ", current, "cur_index: ", sprintf("%.3f", cur_index))
        return(list(target = current, alpha = alpha))
      }

      message("better basis found, index_val = ", sprintf("%.3f", best_row$index_val))
      cur_index <- best_row$index_val
      current <- best_row$basis[[1]]
      best_row <- dplyr::mutate(best_row,
        info = "loop_best",
        method = "search_polish"
      )

      rcd_env[["record"]] <- dplyr::bind_rows(rcd_env[["record"]], best_row)

    } else {
      polish_cooling <- polish_cooling * 0.95
      alpha <- alpha * polish_cooling
      message("alpha gets updated to", alpha, "\n")

      # check condition 3: alpha can't be too small

      if (alpha < 0.01) {
        message("alpha is", alpha, "and it is too small!")
        message("current basis: ", current, "cur_index: ", sprintf("%.3f", cur_index))
        return(list(target = current, alpha = alpha))
      }
    }

    try <- try + 1
  }

  message("No better bases found after ", polish_max_tries, " tries.  Giving up.")
  print_final_proj(current)
  return(list(target = current, alpha = alpha))
}
# globalVariables("index_val")

#' A pseudo-derivative, line search algorithm.
#'
#' This is a novel method for finding more interesting projections for the
#' guided tour.  It works by first taking a small step in \code{n} random
#' directions, and then picking the direction that looks most promising
#' (based on the height of the index function), which is effectively a gradient search.
#' Then it performs a linear search along the geodesic in that direction,
#' traveling up to half way around the sphere.
#'
#' You should not to have call this function directly, but should supply it
#' to the \code{\link{guided_tour}} as a search strategy.
#'
#' @param current starting projection
#' @param alpha maximum distance to travel (currently ignored)
#' @param index interestingness index function
#' @param tries the counter of the outer loop of the opotimiser
#' @param max.tries maximum number of failed attempts before giving up
#' @param n number of random steps to take to find best direction
#' @param delta step size for evaluation of best direction
#' @param cur_index index value for starting projection, set NA if it needs to
#'   be calculated
#' @param ... other arguments being passed into the \code{search_geodesic()}
#' @keywords optimize
#' @export
#' @examples
#' animate_xy(flea[, 1:6], guided_tour(holes(), search_f = search_geodesic))
search_geodesic <- function(current, alpha = 1, index, tries, max.tries = 5, ...,  n = 5,
                            delta = 0.01, cur_index = NA) {
  if (is.na(cur_index)) cur_index <- index(current)

  try <- 1
  while (try < max.tries) {
    # Try 5 random directions and pick the one that has the highest
    # index after a small step in either direction
    direction <- find_best_dir(current, index, counter = n, dist = delta)
    best_dir <- direction$basis[[which.max(direction$index_val)]]
    direction_search <- dplyr::mutate(direction, tries = tries, loop = try)

    # Travel halfway round (pi / 4 radians) the sphere in that direction
    # looking for the best projection
    peak <- dplyr::mutate(find_path_peak(current, best_dir, index), tries = tries, loop = try)
    new_index <- tail(peak$index_val, 1)
    new_basis <- tail(peak$basis, 1)

    rcd_env <- parent.frame(n = 4)
    if (is.null(rcd_env[["record"]])) rcd_env <- parent.frame(n = 1)
    rcd_env[["record"]] <- dplyr::bind_rows(
      rcd_env[["record"]],
      direction_search,
      peak
    )
    rcd_env[["record"]] <- dplyr::mutate(
      rcd_env[["record"]],
      id = dplyr::row_number()
    )

    if (cur_index == 0 | new_index == 0) {
      warning("either the cur_index or the new_index is zero!")
    } else {
      pdiff <- (new_index - cur_index) / cur_index
      if (pdiff > 0.001 & proj_dist(current, new_basis[[1]]) > 1e-2) { # FIXME: pdiff should pbly be a changeable parameter
        message("Target: ", sprintf("%.3f", new_index), ", ",
                sprintf("%.1f", pdiff * 100), "% better ")
        current <- new_basis
        cur_index <- new_index

        return(list(target = new_basis[[1]]))
      }
    }

    try <- try + 1
  }

  message("No better bases found after ", max.tries, " tries.  Giving up.")
  print_final_proj(current)

  NULL
}

#' Find the most promising direction to travel in.
#'
#' Starting from the current projection, pick \code{counter} random location
#' and take a small step towards and away from each location.  The most
#' promising direction has the highest value of the \code{index} function.
#'
#' @keywords optimize internal
#' @param old current projection
#' @param index interestingness index function
#' @param dist step size in radians, should be small
#' @param counter of random steps to take
find_best_dir <- function(old, index, dist = 0.01, counter = 5, ...) {

  # change the original parameter tries to counter since it conflicts with the tries in geodesic-path.r

  bases <- replicate(counter, basis_random(nrow(old), ncol(old)),
    simplify = FALSE
  )

  score <- function(new) {
    interpolator <- geodesic_info(old, new)
    forward <- step_angle(interpolator, dist)
    backward <- step_angle(interpolator, -dist)

    larger <- max(index(forward), index(backward))

    tibble::tibble(
      basis = c(list(forward), list(backward)),
      index_val = c(index(forward), index(backward)),
      info = "direction_search",
      method = "search_geodesic"
    )
  }

  ans <- dplyr::bind_rows(lapply(bases, score))
  ans[which.max(ans$index_val), "info"] <- "best_direction_search"
  ans
}

#' Find the most interesting projection along a geodesic.
#'
#' Use \code{\link{optimize}} to find the most interesting projection amongst
#' all projections on a geodesic.  This method assumes that the function is
#' continuous with a single maximum, but seems to do ok even if there are
#' multiple maxima.
#'
#' @param old currention project
#' @param new projection that gives direction to travel in
#' @param index interestingness index function
#' @param max_dist maximum distance to travel along in radians
#' @keywords optimize internal
find_path_peak <- function(old, new, index, max_dist = pi / 4, ...) {
  interpolator <- geodesic_info(old, new)

  index_pos <- function(alpha) index(step_angle(interpolator, alpha))

  alpha <- stats::optimize(index_pos, c(-max_dist, max_dist), maximum = TRUE, tol = 0.01)

  best <- tibble::tibble(
    basis = list(step_angle(interpolator, alpha$maximum)),
    index_val = alpha$objective,
    info = "best_line_search",
    method = "search_geodesic"
  )

  best
}
# globalVariables("tries")

#' Display tour path with principal component scores with original axes
#'
#' Animate a 2D tour path on data that has been transformed into
#' principal components, and also show the original variable axes.
#'
#' @param axes position of the axes: center, bottomleft or off
#' @param center if TRUE, centers projected data to (0,0).  This pins the
#'  center of data cloud and make it easier to focus on the changing shape
#'  rather than position.
#' @param half_range half range to use when calculating limits of projected.
#'   If not set, defaults to maximum distance from origin to each row of data.
#' @param col color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".
#' @param pch shape of the point to be plotted.  Defaults to 20.
#' @param cex size of the point to be plotted.  Defaults to 1.
#' @param pc_coefs coefficients relating the original variables to
#'   principal components. This is required.
#' @param edges A two column integer matrix giving indices of ends of lines.
#' @param edges.col colour of edges to be plotted, Defaults to "black.
#' @param rescale Default FALSE. If TRUE, rescale all variables to range [0,1].
#' @param palette name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"
#' @param axislablong text labels only for the long axes in a projection, default FALSE
#' @param ...  other arguments passed on to \code{\link{animate}} and
#'   \code{\link{display_slice}}
#' @export
#' @examples
#' flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
#' flea_pca <- prcomp(flea_std, center = FALSE, )
#' flea_coefs <- flea_pca$rotation[, 1:3]
#' flea_scores <- flea_pca$x[, 1:3]
#' animate_pca(flea_scores, pc_coefs = flea_coefs)
display_pca <- function(center = TRUE, axes = "center", half_range = NULL,
                        col = "black", pch = 20, cex = 1,
                        pc_coefs = NULL,
                        edges = NULL, edges.col = "black",
                        palette = "Zissou 1",
                        axislablong = FALSE, ...) {
  labels <- NULL

  # If colors are a variable, convert to colors
  if (is.factor(col) | !areColors(col)) {
    gps <- col
    col <- mapColors(col, palette)
  }

  init <- function(data) {
    half_range <<- compute_half_range(half_range, data, center)
    labels <<- abbreviate(rownames(pc_coefs), 3)
    stopifnot(ncol(pc_coefs) == ncol(data))
  }

  if (!is.null(edges)) {
    if (!is.matrix(edges) && ncol(edges) == 2) {
      stop("Edges matrix needs two columns, from and to, only.")
    }
  }

  render_frame <- function() {
    par(pty = "s", mar = rep(0.1, 4))
    blank_plot(xlim = c(-1, 1), ylim = c(-1, 1))
  }
  render_transition <- function() {
    rect(-1, -1, 1, 1, col = "#FFFFFFE6", border = NA)
  }

  render_data <- function(data, proj, geodesic) {
    # Render axes
    pc_axes <- pc_coefs %*% proj
    draw_tour_axes(pc_axes, labels, limits = 1, axes, longlabels=axislablong)

    # Render projected points
    x <- data %*% proj
    if (center) x <- center(x)
    x <- x / half_range
    points(x, col = col, pch = pch, cex = cex)

    if (!is.null(edges)) {
      segments(x[edges[, 1], 1], x[edges[, 1], 2],
        x[edges[, 2], 1], x[edges[, 2], 2],
        col = edges.col
      )
    }
  }


  list(
    init = init,
    render_frame = render_frame,
    render_transition = render_transition,
    render_data = render_data,
    render_target = nul
  )
}

#' @rdname display_pca
#' @inheritParams animate
#' @export
animate_pca <- function(data, tour_path = grand_tour(), rescale = FALSE, ...) {
  animate(data, tour_path, display_pca(...), rescale = rescale)
}

#' Normalise a numeric matrix.
#'
#' Ensure that columns of a numeric matrix have norm 1
#'
#' @keywords internal algebra
#' @param x numeric matrix or vector
#' @export
normalise <- function(x) {
  if (is.matrix(x)) {
    lengths <- sqrt(colSums(x^2, na.rm = TRUE))
    sweep(x, 2, lengths, "/")
  } else {
    x / sqrt(sum(x^2))
  }
}

#' Orthonormalise using modified Gram-Schmidt process.
#'
#' @keywords internal algebra
#' @param x numeric matrix
#' @export
orthonormalise <- function(x) {
  x <- normalise(x) # to be conservative

  if (ncol(x) > 1) {
    for (j in seq_len(ncol(x))) {
      for (i in seq_len(j - 1)) {
        x[, j] <- x[, j] - as.vector(crossprod(x[, j], x[, i])) * x[, i]
      }
    }
  }

  normalise(x)
}


#' Test if a numeric matrix is orthonormal.
#'
#' @keywords internal algebra
#' @param x numeric matrix
#' @param tol tolerance used to test floating point differences
#' @export
is_orthonormal <- function(x, tol = 0.001) {
  if(is.numeric(x) == FALSE) stop("'x', expected to be numeric and coercable to matrix.")
  if (!is.matrix(x)) x <- as.matrix(x)
  #stopifnot(is.matrix(x))
  nc <- ncol(x)
  iter <- seq_len(nc)
  for (j in iter) {
    if (abs(1-sqrt(sum(x[, j]^2))) > tol) {
      return(FALSE)
    }
  }
  if (nc > 1) {
    # dot product between columns is close to zero
    for (j in iter[2:nc]) {
      rem <- setdiff(iter, j)
      for (i in rem) {
        if (abs(sum(x[, j] * x[, i])) > tol) {
          return(FALSE)
        }
      }
    }
  }
  TRUE
}

#' Orthonormalise one matrix by another.
#'
#' This ensures that each column in x is orthogonal to the corresponding
#' column in by.
#'
#' @keywords internal algebra
#' @param x numeric matrix
#' @param by numeric matrix, same size as x
#' @returns orthonormal numeric matrix
#' @export
orthonormalise_by <- function(x, by) {
  stopifnot(ncol(x) == ncol(by))
  stopifnot(nrow(x) == nrow(by))

  x <- normalise(x)
  by <- normalise(by)

  for (j in seq_len(ncol(x))) {
    for (k in seq_len(ncol(by))) {
      x[, j] <- x[, j] - as.vector(crossprod(x[, j], by[, k])) * by[, k]
      x[, j] <- normalise(x[, j])
    }
  }

  # Last step, columns new matrix to orthonormal
  if (ncol(x) > 1) {
    for (j in 2:ncol(x)) {
      x[, j] <- x[, j] - as.vector(crossprod(x[, j], x[, j-1])) * x[, j-1]
      normalise(x[, j])
    }
  }

  return(x)
}

#' Calculate the distance between two bases.
#'
#' Computes the Frobenius norm between two bases, in radians.  This is
#' equals to the Euclidean norm of the vector of principal angles between
#' the two subspaces.
#
#' @param x projection matrix a
#' @param y projection matrix b
#' @keywords algebra
#' @export
proj_dist <- function(x, y) sqrt(sum((x %*% t(x) - y %*% t(y))^2))

#' Calculate the Mahalanobis distance between points and center.
#'
#' Computes the Mahalanobis distance using a provided variance-covariance
#' matrix of observations from 0.
#
#' @param x matrix of data
#' @param vc pre-determined variance-covariance matrix
#' @keywords algebra
#' @export
mahal_dist <- function(x, vc) {
  n <- dim(x)[1]
  p <- dim(x)[2]
  mn <- rep(0, p)
  ev <- eigen(vc)
  vcinv <- ev$vectors %*% diag(1/ev$values) %*% t(ev$vectors)
  x <- x - matrix(rep(mn, n), ncol = p, byrow = T)
  dx <- NULL
  for (i in 1:n)
    dx <- c(dx, x[i, ] %*% vcinv %*% as.matrix(x[i, ]))
  return(dx)
}

#' A little tour path.
#'
#' The little tour is a planned tour that travels between all axis parallel
#' projections. (John McDonald named this type of tour.)
#'
#' Usually, you will not call this function directly, but will pass it to
#' a method that works with tour paths like \code{\link{animate}},
#' \code{\link{save_history}} or \code{\link{render}}.
#'
#' @param d target dimensionality
#' @param reorder logical, default = TRUE, to randomise the order in which
#' @export
#' @examples
#' animate_xy(flea[, 1:6], little_tour())
#' animate_pcp(flea[, 1:6], little_tour(3))
#' animate_scatmat(flea[, 1:6], little_tour(3))
#' animate_pcp(flea[, 1:6], little_tour(4))
little_tour <- function(d = 2, reorder = TRUE) {
  little <- NULL
  step <- 0

  generator <- function(current, data, ...) {
    if (is.null(little)) {
      # Initialise bases
      little <<- bases_little(ncol(data), d, reorder)
    }
    step <<- (step %% length(little)) + 1
    if (is.null(current)) {
      # need an intercept to start the little tour
      return(little[[step]])
    }
    target <- little[[step]]
    list(target = target)
  }

  new_geodesic_path("little", generator)
}

#' Generate bases for the little tour
#'
#' @keywords internal
#' @param p dimensionality of data
#' @param d dimensionality of target projection
#' bases are visited
bases_little <- function(p, d = 2, reorder = TRUE) {
  b <- diag(rep(1, p))
  vars <- utils::combn(p, d)
  # Let's make the list twice as long, so it visits pairs multiple
  # times, and in different order
  if (reorder)
    vars <- cbind(vars[, sample(1:ncol(vars))], vars[, sample(1:ncol(vars))])
  lapply(seq_len(ncol(vars)), function(i) b[, vars[, i]])
}

#' Generate a geodesic path between bases supplied by generator
#'
#' A tour path is a function that when called with the current projection
#' and data set, generates sequence of \code{\link{geodesic_path}}s.  The
#' path can either span the whole space of orthonormal matrices, the default
#' or be restricted to a subspace with the frozen argument.  More details
#' are given in the documentation for \code{\link{freeze}}.
#'
#' Subsequent frames are guaranteed to be at least 0.001 radians away from
#' the current frame.  If after 10 tries the generator does not give a new
#' basis at least this far away then we give up.
#'
#' If a suitable new basis can not be found, the path function returns NULL
#' indicating that the tour should stop.
#'
#' @param name name to give tour path
#' @param generator basis generator function
#' @param frozen matrix giving frozen variables, as described in
#'   \code{\link{freeze}}
#' @export
#' @keywords internal
new_geodesic_path <- function(name, generator, frozen = NULL, ...) {
  tries <- 1 # Needed for guided

  tour_path <- function(current, data, ...) {
    if (is.null(current)) {
      return(generator(NULL, data, tries, ...))
    }

    # Keep trying until we get a frame that's not too close to the
    # current frame
    dist <- 0
    while (dist < 1e-3) {
      if (name %in% c("guided", "frozen-guided")) tries <<- tries + 1

      gen <- generator(current, data, tries, ...)
      target <- gen$target
      if (inherits(target, "multi-bases")) return(list(target = target))

      # generator has run out, so give up
      if (is.null(target)) {
        return(NULL)
      }

      dist <- proj_dist(current, target)
      if (dist < 1e-2) {
        return(NULL)
      }

    }
    list(ingred = geodesic_path(current, target, frozen, ...), index = gen$index, tries = tries)
  }

  structure(
    tour_path,
    name = name,
    class = c("tour_path", "function")
  )
}


#' @export
"print.tour_path" <- function(x, ...) {
  message("Tour path: ", attr(x, "name"))

  # params <- as.list(environment(get("generator", environment(g))))
  # str(params)
}

---
title: "Introduction to tourr"
author: "Ursula Laa"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to tourr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  cache = FALSE,
  eval = FALSE,
  rmarkdown.html_vignette.check_title = FALSE
)
```

A tour animates a display of interpolated low-dimensional projections from high-dimensional data, to explore the shape of the multivariate distribution. The shape might be elliptical if the distribution is roughly normal, or there may be clusters corresponding to grouping the data according to known or unknown variables. There might be outliers, the can be identified because they are separated from the other observations, or move on different paths that other points. It may be that there are curvilinear patterns indicating nonlinear association between some variables.

This vignette shows how to use the `tourr` package to generate animations using different tour and display types.

```{r setup}
library(tourr)
```


## Quick start

To get started we can simply generate a tour animation for any numeric data matrix with default settings. Here we look at the flea data available in the package, drop the last column (the grouping variable), and call

```{r anim1}
animate(flea[,1:6])
```

![Frame from first tour animation made with above code (and also code below).](figures/anim1.png){width=400 fig-alt="Scatterplot overlaid by a circle and line segments representing the projection coefficients. The points form a shallow arch, and there are 6 line segments pointing in various directions around a clock labelled tr2, ad1, hed, ad1, tr1, ad2."}

Here we only see a single frame that is generated on the tour path, but if you call `animate` in your console you will launch a tour animation that you can view in your graphics window.

We could have also launched the same tour by specifying the default selection explicitly

```{r anim2}
animate(flea[,1:6],
        tour_path = grand_tour(),
        display = display_xy()
        )
```

The most common application is to use a grand tour, which means that the projections (or bases) are randomly selected and give a global overview of the distribution. The default display is a two-dimensional scatterplot display.

## Tour types

As we have seen, the default tour type is a grand tour. By selecting the second argument, `tour_path`, in the animate function we can change this, choosing from the tour types that have been implemented in the package. The most relevant options are:

- *grand tour*: randomly selected bases
- *guided tour*: basis selection is optimising an index function
- *planned tour*: interpolates between a set of input bases
- *little tour*: a planned tour between all axis parallel projections
- *local tour*: alternating between the starting projection and randomly selected nearby projections
- *slice tour*: points outside a given orthogonal distance of the projection plane appear smaller.
- *sage tour*: reverse the curse of dimensionalityt piling, in the 2D projection

For example, the guided tour can be used to move towards more interesting views of the distribution as the animation progresses. To measure the interestingness of each projection we need to define an index function which will get maximised along the tour path. You can define your own index function, or use one of those available in the `tourr` package. For our example flea data we can use the holes index which is looking for projections with low densities near the center, and can often find views that reveal clustering.

```{r anim3}
animate(flea[,1:6],
        tour_path = guided_tour(holes()),
        display = display_xy())
```

```
Converting input data to the required matrix format.
Press Esc to stop tour running
Target: 0.733, 17.5% better 
Using half_range 4.4
Target: 0.872, 19.0% better 
Target: 1.106, 26.8% better 
Target: 1.243, 12.3% better 
Target: 1.245, 0.2% better 
Target: 1.258, 1.1% better 
Target: 1.261, 0.2% better 
Target: 1.268, 0.6% better 
Target: 1.271, 0.2% better 
Target: 1.272, 0.1% better 
Target: 1.273, 0.1% better 
Target: 1.276, 0.2% better 
Target: 1.278, 0.1% better 
No better bases found after 25 tries.  Giving up.
Final projection: 
-0.151  -0.738  
 0.488  -0.378  
 0.379  -0.153  
 0.600   0.024  
-0.092  -0.516  
 0.476   0.145  
```

![Final projection from a guided tour shows a hint of three clusters.](figures/anim2.png){width=400 fig-alt="Scatterplot overlaid by a circle and line segments representing the projection coefficients. The points fall into three elliptical clusters at top right, bottom left, and middle right. The line segments ad2, tr1 point down, and the other four roughly to the right."}

When running the full guided tour for this example, feedback on the optimisation is provided in the console, and the final view shows three clusters (not very separated) that correspond to the three species in the dataset. A random start is used be default, which will produce different results each run. To get the same result repeatedly, you need to set the seed. By mapping the species to color we can see how the groups get teased apart in the guided tour. 

```{r anim4}
animate(flea[,1:6],
        tour_path = guided_tour(holes()),
        display = display_xy(col = flea$species))
```

![The three clusters correspond to the species variable.](figures/anim3.png){width=400 fig-alt="Approximately the same as the previous plot, oriented in the opposite direction so the middle cluster is now towards the top. The legend has blue as Concinna, yellow as Heikert, red as Heptapot. Yellow cluster is at lower left, red cluster is at top right, and blue is in between."}

Alternatively, a colour vector can be created manually and passed into the guided tour.

```{r anim5}
# defining the color palette
clrs <- c("#486030", "#c03018", "#f0a800")
# mapping the species vector onto a color vector
flea_col <- clrs[as.numeric(flea$species)]
# the color vector specifies the color for each point
# and gets passed into the display function
animate(flea[,1:6],
        tour_path = guided_tour(holes()),
        display = display_xy(col = flea_col))
```


## Display types

The default display is showing a scatterplot of the data projected to 2D. Depending on the number of dimensions we project onto (this is called $d$), we can choose different display types. For example, we could look at 1D projections in a density display, a 3D projection using depth cues, or higher dimensional projections in a parallel coordinate or scatterplot matrix display.

When changing the display type via the `display` argument, we need to make sure that the basis generation is matching the corresponding number of dimensions $d$. We can pass in this information when generating the tour path. To work with 1D projections and a density display, we can launch the animation as

```{r anim6}
animate(flea[,1:6],
        tour_path = grand_tour(d = 1),
        display = display_dist()
        )
```

but there is also a shortcut available

```{r anim7}
animate_dist(flea[,1:6])
```

![A tour of 1D projections displayed as a denisty plot.](figures/anim4.png){width=400 fig-alt="Density plot at the top showing two modes. On the bottom is a representation of the axes, as horizontal line segments: t1, t2 point to the left, and hd, a1, a2, a3 point to the right." }

## Saving a tour

After exploring the different options, we may have identified a particularly interesting tour that we may want to save or share.

The first option is to save the animation (or its individual frames). This is possible through the `render` functions, that save the frame views to png or pdf format. Another option, perhaps more convenient, is to directly save the full animation to a gif file. To use the function, you will need to install the `gifski` package.

An advanced alternative is to save the tour path, which can later be replayed as a planned tour, and allows us to look at the same tour in different displays. For example we can save a default grand tour path for the flea data and then replay it in a scatterplot display

```{r anim8}
t1 <- save_history(flea[,1:6], max = 3)
animate(flea[,1:6], planned_tour(t1))
```

---
title: "Tour types"
author: "Claude and Di Cook"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tour types}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = TRUE
)
library(tourr)
```

A tour animates a sequence of low-dimensional linear projections of
high-dimensional data. Think of it like shining a light on an object and
watching the shadow: any single shadow (projection) shows you one view of the
shape, but watching many shadows in sequence — as the object slowly rotates —
lets you build up a picture of the whole structure. Clusters, outliers,
non-linear patterns, and variable associations that would be invisible in a
single 2D plot can emerge as the tour progresses.

The two main ingredients in any tour are the **projection dimension** `d`
(how many dimensions to project down to, usually 1 or 2) and the **tour
type**, which controls how the sequence of projections is chosen. This
vignette walks through each tour type, explains when to use it, and shows
the code needed to run it.

Throughout we use the `flea` dataset (74 flea beetle measurements on 6
numeric variables), which has been standardised to have mean 0 and unit variance. The views can be misleading if your data is not standardised, or variables have not been scaled to be on comparable units. 

---

## Grand tour

The grand tour generates target projections by sampling uniformly from the
space of all `d`-dimensional planes and interpolates smoothly between them.
Over time it visits every possible projection, giving a global overview of the
data. It is the right starting point when you do not know what structure to
look for.

```{r grand, eval = FALSE}
# 2D projections — the default
f <- flea[,1:6] # For convenience
animate_xy(f)

# 1D projections shown as a density
animate_dist(f)
```

Points that stand apart from the main cloud in many projections are likely
outliers. Groups of points that stay together through many projections, then
occasionally merge with others, suggest clusters. Because the grand tour
samples randomly it will eventually show every view, but it may take a while
to stumble on the most revealing one — that is where the guided tour helps.

---

## Guided tour

The guided tour adds a direction to the search: it uses **projection
pursuit** to move towards projections that score highly on a chosen index
function. Each new target is not random but chosen to increase the index,
so the tour makes progress towards the structure you care about rather than
wandering at random.

The index function is the key choice:

- `holes()` looks for projections with low density near the centre and high
  density near the edges — a signature of separated clusters.
- `cmass()` looks for projections with high density near the centre —
  useful for finding outliers that sit away from a dense core.
- `lda_pp()` and `pda_pp()` use known class labels to find projections where
  groups are well separated.

```{r guided, eval = FALSE}
# Find cluster structure without class labels
animate_xy(f, tour_path = guided_tour(holes()))

# Same search, colour by known species once structure is found
animate_xy(f,
  tour_path = guided_tour(holes()),
  col = flea$species
)

# Use class labels directly to separate groups
animate_xy(f,
  tour_path = guided_tour(lda_pp(flea$species)),
  col = flea$species
)
```

The guided tour terminates when no better projection can be found, so it will
stop on its own. The final view it settles on is usually the most revealing
one for that index. If it stops too quickly — without finding visible structure
— try `sphere = TRUE` in `animate()` to remove linear associations first, or
start from a different random seed.

---

## Little tour

The little tour is a planned tour between all axis-parallel projections: it
cycles through every pair of original variables in turn. This makes it a
systematic way to survey all the marginal 2D views of the data, equivalent to
watching a slow animated scatterplot matrix.

```{r little, eval = FALSE}
animate_xy(f, tour_path = little_tour())
```

Because the little tour only visits axis-parallel projections it can miss
structure that lives in oblique directions — a cluster that separates along a
diagonal will not appear. It is best used as a sanity check or a complement
to the grand tour rather than a primary exploration tool.

---

## Planned tour

The planned tour replays a sequence of bases you have already saved. The
typical workflow is to run a grand tour with `save_history()`, keep the saved
path, and then replay it — perhaps in a different display, or with colour
added after the fact.

```{r planned, eval = FALSE}
# Save a grand tour path
set.seed(42)
t1 <- save_history(f, max = 10)

# Replay it in a scatterplot
animate_xy(f, tour_path = planned_tour(t1))

# Replay the same path with species colour added
animate_xy(f,
  tour_path = planned_tour(t1),
  col = flea$species
)

# Cycle continuously through the saved bases
animate_xy(f, tour_path = planned_tour(t1, cycle = TRUE))
```

The planned tour is also how you share a specific tour with someone else:
save the history with `save_history()`, save the object to disk with
`saveRDS()`, and send it alongside your code.

---

## Local tour

The local tour makes small movements around a chosen starting projection. At
each step it picks a new target that is within a specified angular distance
(`angle`, in radians) of the current position, so the view never strays far
from where you started. This is useful for examining whether a pattern you
have spotted is robust — if the structure survives small perturbations of the
projection it is a real feature rather than a coincidence of a particular
viewing angle.

```{r local, eval = FALSE}
# Start from a specific projection (e.g. one saved from a guided tour)
start <- basis_random(6, 2)

# Explore a small neighbourhood of angle pi/4
animate_xy(f,
  tour_path = local_tour(start, angle = pi / 4),
  col = flea$species
)
```

A tighter `angle` keeps the view closer to the starting point. A wider angle
effectively becomes a grand tour. Values around `pi/4` to `pi/8` are a
reasonable starting range.

---

## Radial tour

The radial tour answers a specific and practical question: *how important is
a particular variable to the pattern I can currently see?* Starting from a
chosen projection, it rotates one variable at a time smoothly out of the
projection plane and then back in again, like a dial being turned to zero and
back. Watching what happens to the structure in the plot as the variable is
removed tells you exactly how much that variable contributes.

The two required arguments are `start`, the projection matrix to begin from,
and `mvar`, the index (or indices) of the variable(s) to rotate in and out.
The `start` projection is typically one you have already found interesting —
for example the final view of a guided tour.

```{r radial, eval = FALSE}
# Use a saved projection as the starting point — here we take the
# end-point of a guided tour run with holes()
set.seed(42)
guided_path  <- save_history(f, guided_tour(holes()), sphere = TRUE)
start        <- matrix(guided_path[, , dim(guided_path)[3]], nrow = 6)

# Rotate variable 4 (elytra width) out and back in
animate_xy(f,
  tour_path = radial_tour(start, mvar = 4),
  col = flea$species,
  rescale = TRUE
)

# Rotate two variables simultaneously to see their joint contribution
animate_xy(f,
  tour_path = radial_tour(start, mvar = c(3, 4)),
  col = flea$species,
  rescale = TRUE
)
```

The radial tour also works with other display types. If the starting
projection is 1D, use `animate_dist()`:

```{r radial-1d, eval = FALSE}
start1d <- basis_random(6, 1)
animate_dist(f, radial_tour(start1d, mvar = 2), rescale = TRUE)
```

**Reading the result.** When the variable is rotated out, watch whether the
structure (clusters, gaps, outlier positions) collapses or survives:

- Structure *disappears* when the variable is removed → the variable is
  essential to producing that pattern.
- Structure *survives* → the pattern exists without that variable; other
  variables carry it.
- Structure *changes shape but does not vanish* → the variable contributes
  partially; removing it reveals a different facet of the same underlying
  structure.

This makes the radial tour one of the most useful tools for variable
selection and for building intuition about which measurements drive the
patterns in your data. It is often run after a guided tour: find an
interesting projection with the guided tour, then probe each variable in turn
with the radial tour to understand what is making the pattern.

---

## Frozen tour

The frozen tour holds the projection coefficients for some variables fixed at
specified values and lets the remaining coefficients vary freely under a grand
tour. This is useful when you already know that a particular variable or
direction is important and want to explore the remaining variation while keeping
that variable anchored in view.

Frozen values are specified with a matrix of `NA` and numeric entries: `NA`
means the coefficient varies freely; a number fixes it.

```{r frozen, eval = FALSE}
# Fix variable 3 to contribute equally to both axes
frozen      <- matrix(NA, nrow = 6, ncol = 2)
frozen[3, ] <- 0.5

animate_xy(f, tour_path = frozen_tour(2, frozen))
```

The frozen tour is the most specialised of the tour types and is most useful
when you have domain knowledge that a particular variable should always be
visible — for example, when verifying a hypothesis about the role of a specific
measurement.

---

## Choosing the right tour

| Goal | Recommended tour |
|---|---|
| General exploration, no prior knowledge | `grand_tour()` |
| Find clusters or outliers efficiently | `guided_tour(holes())` |
| Find class separation with labels | `guided_tour(lda_pp(class))` |
| Survey all pairwise marginal views | `little_tour()` |
| Replay or share a specific path | `planned_tour(saved_history)` |
| Stress-test a pattern found elsewhere | `local_tour(start)` |
| Assess which variables drive a pattern | `radial_tour(start, mvar)` |
| Anchor one variable while exploring others | `frozen_tour(d, frozen)` |

A common workflow is to start with the grand tour to get an overview, switch
to the guided tour to pursue interesting features, use the radial tour to
understand which variables are responsible for those features, and then use
the local tour to confirm that the pattern is stable under small perturbations.
The planned tour is then used to communicate findings by replaying the key
paths with informative colours or labels.

---

## Saving a tour to a file

All tour types can be rendered to a GIF with `render_gif()`. Replace
`animate_xy()` with `render_gif()` and supply a `gif_file` path and
`frames` count.

```{r gif, eval = FALSE}
render_gif(
  f,
  tour_path = guided_tour(holes()),
  display   = display_xy(col = flea$species),
  gif_file  = "guided_flea.gif",
  frames    = 60
)
```

## Further reading

Cook and Laa (2024), *Interactively Exploring High-Dimensional Data and Models
in R*, Chapman & Hall/CRC, provides a comprehensive treatment of tour methods
and their application to clustering, dimension reduction, and classification.
The online version is available at <https://dicook.github.io/mulgar_book/>.

---
title: "Labelling observations in animate_xy"
author: "Claude and Di Cook"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Labelling observations in animate_xy}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = TRUE
)
library(tourr)
```

Labels can be added to individual observations in `animate_xy` via the
`obs_labels` argument of `display_xy`. This is useful when you want to track
specific points — outliers, known cases of interest, or a named subset — as
they move through the tour.

## Basic usage

Pass a character vector of the same length as the number of rows in your data.
Each element is the label that will be drawn next to the corresponding point.
Turning axes off keeps the plot uncluttered when labels are present.

```{r basic, eval = FALSE}
animate_xy(flea[,1:6],
  obs_labels = as.character(1:nrow(flea)),
  axes = "off"
)
```

Row numbers make a reasonable default label because they let you trace a point
back to a specific row of your data frame.

![Projection from a grand tour showing all the points labelled by row number. ](figures/obs1.png){width=400 fig-alt="Scatterplot with 74 points, and each one is numbered. Mostly numbers are unreaddable, except for a few on the outside of the cloud like 41 at the bottom."}

## Labelling a subset of observations

Labelling every point can become hard to read in larger datasets. To label
only a subset, supply an empty string `""` for the observations you want
to leave unlabelled. Only the non-empty strings are drawn.

Here we label just the five most extreme outliers, identified in advance
by their Mahalanobis distance:

```{r subset, eval = FALSE}
# Find the five points furthest from the multivariate centre
d2    <- mahalanobis(flea[,1:6], colMeans(flea[,1:6]), cov(flea[,1:6]))
top5  <- order(d2, decreasing = TRUE)[1:5]

# Build a label vector: empty for most points, row number for the top 5
lbls <- rep("", nrow(flea))
lbls[top5] <- as.character(top5)

animate_xy(flea[,1:6], obs_labels = lbls, axes="bottomleft")
```

![Projection from a grand tour showing some of the points labelled by row number, furthest from the mean according to Mahalanobis distance. ](figures/obs2.png){width=400 fig-alt="Scatterplot with 74 points. Five outlying points are labelled: 5, 22, 41, 67, 68."}

## Combining labels with colour

Labels and the `col` argument work independently, so you can colour by group
while still labelling individual points. A common workflow is to colour all
observations by a grouping variable and then label only the few points you
want to call out by name.

```{r colour, eval = FALSE}
# Colour by species; label only row 1 and row 22 (arbitrary examples)
lbls <- rep("", nrow(flea))
lbls[c(1, 22)] <- rownames(flea)[c(1, 22)]

animate_xy(flea[,1:6],
  col       = flea$species,
  obs_labels = lbls,
  axes      = "bottomleft"
)
```

![Projection from a grand tour showing two points labelled by row number. ](figures/obs3.png){width=400 fig-alt="Scatterplot with 74 points coloured as blue, red, yellow matching the three species Concinna, Heptapot. and Heikert. Two points are labelled: 1 and 22."}

## Saving a labelled tour to a GIF

Interactive use is fine for exploration, but if you want to share a labelled
tour you can render it to a GIF with `render_gif`. The `frames` argument
controls how many frames are captured before the file is written; `width` and
`height` are in pixels.

```{r gif, eval = FALSE}
lbls <- rep("", nrow(f))
lbls[c(1, 22)] <- rownames(flea)[c(1, 22)]

render_gif(
  flea[,1:6],
  tour_path = grand_tour(),
  display   = display_xy(
    col        = flea$species,
    obs_labels = labs,
    axes       = "bottomleft"
  ),
  gif_file = "labelled_tour.gif",
  frames   = 60,
  width    = 400,
  height   = 400
)
```

## A note on label placement

Labels are drawn by the underlying call to `text()` in base graphics,
positioned directly at each point's projected coordinates. In dense regions
of a projection labels will overlap, which is expected — as the tour rotates,
points separate and the labels become readable in turn. If overlap is a
persistent problem, consider labelling a smaller subset or reducing `cex` via
the `...` passthrough:

```{r cex, eval = FALSE}
animate_xy(flea[,1:6],
  obs_labels = as.character(1:nrow(f)),
  axes = "off",
  cex  = 0.6
)
```

---
title: "Edges in scatterplot displays"
author: "Ursula Laa"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Edges in scatterplot displays}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  eval = FALSE,
  comment = "#>",
  rmarkdown.html_vignette.check_title = FALSE
)
```

```{r setup}
library(tourr)
```

Showing connections between points can be useful in different applications, including geometric shapes, indicating patterns in sequential series or showing connections suggested by a model (for example in clustering). These edges can be added in the 2D scatterplot display (`display_xy`) through the `edges` argument, and should be specified as a two column integer matrix of the indices of ends of lines.

For example, we can generate cube vertices and edges using the `geozoo` package. The information stored as `points` constitutes the input data, and `edges` can directly be passed into the display.

```{r cube3}
# generate 3D cube vertices
cube <- geozoo::cube.iterate(3)
# data is stored points, edges contains the needed two column matrix for connecting points
cube$points
cube$edges
# call grand tour with the scatterplot (xy) display, turning off axes
animate_xy(cube$points, edges = cube$edges, axes = "off")
```

![Edges added to the vertices of a 3D cube.](figures/cube3.png){width=400 fig-alt="Recognisable 3D clube wireframe."}

We can use the same functions for higher dimensional cubes as well (but would not print the data).

```{r cube5}
cube5 <- geozoo::cube.iterate(5)
animate_xy(cube5$points, edges = cube5$edges, axes = "off")
```

![Edges added to the vertices of a 5D cube.](figures/cube5.png){width=400 fig-alt="Wireframe of a 5D cube, where the 32 vertices are connected appropriately by edges. All vertices and edges are visible."}

It can often be useful to connect points with edges, for example sequential points in a time series, or points that are connected by a model. As a simple example we can connect centroids found when clustering the flea dataset:

```{r}
# get centroids of 3 clusters of the flea data
n <- nrow(flea)
set.seed(1019)
flea_centroids <- stats::kmeans(flea[,1:6], 3)$centers
flea_aug <- rbind(flea[,1:6], flea_centroids)
col <- c(rep("black", n), rep("orange", 3))
flea_edges <- matrix(c(n+1, n+2, n+1, n+3, n+2, n+3), ncol=2, byrow = TRUE)
animate_xy(flea_aug, edges = flea_edges, 
           col = col, edges.col = "orange", 
           edges.width = 3)
```

![Connecting the k-centroids with edges.](figures/kmeans.png){width=400 fig-alt="Scatterplot overlaid by a circle and line segments representing the projection coefficients. The key information is an orange triangle visibly connecting the centres of three clusters."}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{t1}
\alias{t1}
\title{Saved history of guided tour with holes}
\description{
This data was generated from the following code:
set.seed(2020)
t1 <- save_history(flea[, 1:6], guided_tour(holes()), max = 100)
attr(t1, "class") <- NULL
And used as an example for search_polish() to start
optimising from the best projection from search_geodesic.
t1 is a 3D array or 2D projections.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{Flea measurements}
\alias{Flea measurements}
\alias{flea}
\alias{flea,}
\alias{flea_raw}
\title{Flea beatle measurements}
\format{
A 74 x 7 numeric array
}
\usage{
flea
}
\description{
This data is from a paper by A. A. Lubischew, "On the Use of Discriminant
Functions in Taxonomy", Biometrics, Dec 1962, pp.455-477. Data is
standardized, and original units are in flea_raw.
}
\details{
\itemize{
  \item tars1, width of the first joint of the first tarsus in microns
  (the sum of measurements for both tarsi)
  \item tars2, the same for the second joint
  \item head, the maximal width of the head between the external edges of the
  eyes in 0.01 mm
  \item ade1, the maximal width of the aedeagus in the fore-part in microns
  \item ade2, the front angle of the aedeagus ( 1 unit = 7.5 degrees)
  \item ade3, the aedeagus width from the side in microns
  \item species, which species is being examined - concinna, heptapotamica, heikertingeri
}
}
\examples{

head(flea)
animate_xy(flea[, -7])
animate_xy(flea[, -7], col = flea[, 7])
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tourr-package.R
\docType{package}
\name{tourr-package}
\alias{tourr}
\alias{tourr-package}
\title{tourr: Tour Methods for Multivariate Data Visualisation}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Implements geodesic interpolation and basis generation functions that allow you to create new tour methods from R.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://ggobi.github.io/tourr/}
  \item Report bugs at \url{https://github.com/ggobi/tourr/issues}
}

}
\author{
\strong{Maintainer}: Dianne Cook \email{dicook@monash.edu} (\href{https://orcid.org/0000-0002-3813-7155}{ORCID})

Authors:
\itemize{
  \item Dianne Cook \email{dicook@monash.edu} (\href{https://orcid.org/0000-0002-3813-7155}{ORCID})
  \item Hadley Wickham \email{h.wickham@gmail.com} (\href{https://orcid.org/0000-0003-4757-117X}{ORCID}) [contributor]
}

Other contributors:
\itemize{
  \item Nick Spyrison \email{nicholas.spyrison@monash.edu} (\href{https://orcid.org/0000-0002-8417-0212}{ORCID}) [contributor]
  \item Ursula Laa \email{ursula.laa@boku.ac.at} (\href{https://orcid.org/0000-0002-0249-6439}{ORCID}) [contributor]
  \item H. Sherry Zhang \email{huizezhangsh@gmail.com} (\href{https://orcid.org/0000-0002-7122-1463}{ORCID}) [contributor]
  \item Stuart Lee \email{stuart.lee1@monash.edu} (\href{https://orcid.org/0000-0003-1179-8436}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slicing-util.r
\name{anchored_orthogonal_distance}
\alias{anchored_orthogonal_distance}
\title{Calculate orthogonal distances}
\usage{
anchored_orthogonal_distance(plane, data, anchor = NULL)
}
\arguments{
\item{plane}{matrix specifying the projection plane}

\item{data}{data frame or matrix}

\item{anchor}{A vector specifying the reference point to anchor the plane
If NULL (default) the slice will be anchored at the origin.}
}
\value{
distance vector
}
\description{
For each datapoint this function calculates the orthogonal distance from the
anchored projection plane.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-guided.r
\name{guided_tour}
\alias{guided_tour}
\title{A guided tour path.}
\usage{
guided_tour(
  index_f,
  d = 2,
  cooling = 0.99,
  max.tries = 25,
  max.i = Inf,
  search_f = search_geodesic,
  n_jellies = 30,
  n_sample = 100,
  alpha = 0.5,
  ...
)
}
\arguments{
\item{index_f}{the index function to optimise.}

\item{d}{target dimensionality}

\item{cooling}{the amount the size of the search window should be adjusted
by after each step}

\item{max.tries}{the maximum number of unsuccessful attempts to find
a better projection before giving up}

\item{max.i}{the maximum index value, stop search if a larger value is found}

\item{search_f}{the search strategy to use: \code{\link{search_geodesic}}, \code{\link{search_better}},
\code{\link{search_better_random}}, \code{\link{search_polish}}. Default is \code{\link{search_geodesic}}.}

\item{n_jellies}{only used for \code{search_jellyfish}, the number of jellyfish to use}

\item{n_sample}{number of samples to generate if \code{search_f} is \code{\link{search_polish}}}

\item{alpha}{the initial size of the search window, in radians}

\item{...}{arguments sent to the search_f}
}
\description{
Instead of choosing new projections at random like the grand tour, the
guided tour always tries to find a projection that is more interesting
than the current projection.
}
\details{
Currently the index functions only work in 2d.

Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
animate_xy(flea_std, guided_tour(holes()), sphere = TRUE)
\donttest{
animate_xy(flea_std, guided_tour(holes(), search_f = search_better_random), sphere = TRUE)
animate_dist(flea_std, guided_tour(holes(), 1), sphere = TRUE)
animate_xy(flea_std, guided_tour(lda_pp(flea$species)), sphere = TRUE, col = flea$species)

# save_history is particularly useful in conjunction with the
# guided tour as it allows us to look at the tour path in many different
# ways
f <- flea_std[, 1:3]
tries <- replicate(5, save_history(f, guided_tour(holes())), simplify = FALSE)
}
}
\seealso{
\code{\link{cmass}}, \code{\link{holes}} and \code{\link{lda_pp}}
  for examples of index functions.  The function should take a numeric
  matrix and return a single number, preferably between 0 and 1.
\code{\link{search_geodesic}}, \code{\link{search_better}},
  \code{\link{search_better_random}} for different search strategies
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{print_final_proj}
\alias{print_final_proj}
\title{Print out the final projection basis}
\usage{
print_final_proj(basis)
}
\arguments{
\item{basis}{the projection basis to print, used in search_* functions}
}
\description{
Print out the final projection basis
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-frozen.r
\name{freeze}
\alias{freeze}
\alias{thaw}
\title{Freeze and thaw matrices}
\usage{
freeze(input, frozen)

thaw(input, frozen)
}
\description{
Some terminology:
  * frozen variables: the variables that have fixed values
  * warm variables: the remaining variables that vary freely
}
\details{
A frozen matrix specifies which variables to fix in a projection matrix.
Warm variables should be missing (\code{NA}) while frozen variables should
be set to their fixed values.
}
\examples{
frozen <- matrix(NA, nrow = 4, ncol = 2)
frozen[3, ] <- .5

input <- basis_random(4, 2)
freeze(input, frozen)
thaw(input, frozen)
freeze(basis_random(4, 2), frozen)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-grand.r
\name{basis_init}
\alias{basis_init}
\title{Generate initial basis.}
\usage{
basis_init(n, d)
}
\arguments{
\item{n}{dimensionality of data}

\item{d}{dimensionality of target projection}
}
\description{
First two variables are projected on first two axes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interesting-indices.r
\name{lda_pp}
\alias{lda_pp}
\title{LDA projection pursuit index.}
\usage{
lda_pp(cl)
}
\arguments{
\item{cl}{class to be used.  Such as "color"}
}
\description{
Calculate the LDA projection pursuit index.  See Cook and Swayne (2007)
Interactive and Dynamic Graphics for Data Analysis for equations.
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-frozen.r
\name{frozen_guided_tour}
\alias{frozen_guided_tour}
\title{The frozen guided tour}
\usage{
frozen_guided_tour(frozen, index_f, d = 2, max.tries = 25)
}
\arguments{
\item{frozen}{matrix of frozen variables, as described in
\code{\link{freeze}}}

\item{index_f}{the index function to optimise.}

\item{d}{target dimensionality}

\item{max.tries}{the maximum number of unsuccessful attempts to find
a better projection before giving up}
}
\description{
The frozen guided tour
}
\examples{
frozen <- matrix(NA, nrow = 4, ncol = 2)
frozen[3, ] <- .5
animate_xy(flea[, 1:4], frozen_guided_tour(frozen, holes()))
}
\seealso{
\code{\link{cmass}}, \code{\link{holes}} and \code{\link{lda_pp}}
  for examples of index functions.  The function should take a numeric
  matrix and return a single number, preferrably between 0 and 1.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-frozen.r
\name{frozen_tour}
\alias{frozen_tour}
\title{A frozen tour path.}
\usage{
frozen_tour(d = 2, frozen)
}
\arguments{
\item{d}{target dimensionality}

\item{frozen}{matrix of frozen variables, as described in
\code{\link{freeze}}}
}
\description{
A frozen tour fixes some of the values of the orthonormal projection
matrix and allows the others to vary freely according to any of the
other tour methods.  This frozen tour is a frozen grand tour.  See
\code{\link{frozen_guided_tour}} for a frozen guided tour.
}
\details{
Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
frozen <- matrix(NA, nrow = 4, ncol = 2)
frozen[3, ] <- .5
animate_xy(flea[, 1:4], frozen_tour(2, frozen))

frozen <- matrix(NA, nrow = 4, ncol = 2)
frozen[1, 1] <- 0.5
animate_xy(flea[, 1:4], frozen_tour(2, frozen))

# Doesn't work - a bug?
frozen <- matrix(NA, nrow = 4, ncol = 2)
frozen[1:2, 1] <- 1 / 4
animate_xy(flea[, 1:4], frozen_tour(2, frozen))

\dontrun{
# This freezes one entire direction which causes a problem,
# and is caught by error handling.
# If you want to do this it would be best with a dependence
# tour, with one variable set one axis, eg 3rd variable to
# x axis would be indicated from the code below
frozen <- matrix(NA, nrow = 4, ncol = 2)
frozen[3, ] <- c(0, 1)
animate_xy(flea[, 1:4], frozen_tour(2, frozen))
}

# Two frozen variables in five.
frozen <- matrix(NA, nrow = 5, ncol = 2)
frozen[3, ] <- .5
frozen[4, ] <- c(-.2, .2)
animate_xy(flea[, 1:5], frozen_tour(2, frozen))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geodesic-path.r
\name{new_geodesic_path}
\alias{new_geodesic_path}
\title{Generate a geodesic path between bases supplied by generator}
\usage{
new_geodesic_path(name, generator, frozen = NULL, ...)
}
\arguments{
\item{name}{name to give tour path}

\item{generator}{basis generator function}

\item{frozen}{matrix giving frozen variables, as described in
\code{\link{freeze}}}
}
\description{
A tour path is a function that when called with the current projection
and data set, generates sequence of \code{\link{geodesic_path}}s.  The
path can either span the whole space of orthonormal matrices, the default
or be restricted to a subspace with the frozen argument.  More details
are given in the documentation for \code{\link{freeze}}.
}
\details{
Subsequent frames are guaranteed to be at least 0.001 radians away from
the current frame.  If after 10 tries the generator does not give a new
basis at least this far away then we give up.

If a suitable new basis can not be found, the path function returns NULL
indicating that the tour should stop.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-pca.r
\name{display_pca}
\alias{display_pca}
\alias{animate_pca}
\title{Display tour path with principal component scores with original axes}
\usage{
display_pca(
  center = TRUE,
  axes = "center",
  half_range = NULL,
  col = "black",
  pch = 20,
  cex = 1,
  pc_coefs = NULL,
  edges = NULL,
  edges.col = "black",
  palette = "Zissou 1",
  axislablong = FALSE,
  ...
)

animate_pca(data, tour_path = grand_tour(), rescale = FALSE, ...)
}
\arguments{
\item{center}{if TRUE, centers projected data to (0,0).  This pins the
center of data cloud and make it easier to focus on the changing shape
rather than position.}

\item{axes}{position of the axes: center, bottomleft or off}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{col}{color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".}

\item{pch}{shape of the point to be plotted.  Defaults to 20.}

\item{cex}{size of the point to be plotted.  Defaults to 1.}

\item{pc_coefs}{coefficients relating the original variables to
principal components. This is required.}

\item{edges}{A two column integer matrix giving indices of ends of lines.}

\item{edges.col}{colour of edges to be plotted, Defaults to "black.}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}

\item{axislablong}{text labels only for the long axes in a projection, default FALSE}

\item{...}{other arguments passed on to \code{\link{animate}} and
\code{\link{display_slice}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}

\item{rescale}{Default FALSE. If TRUE, rescale all variables to range [0,1].}
}
\description{
Animate a 2D tour path on data that has been transformed into
principal components, and also show the original variable axes.
}
\examples{
flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
flea_pca <- prcomp(flea_std, center = FALSE, )
flea_coefs <- flea_pca$rotation[, 1:3]
flea_scores <- flea_pca$x[, 1:3]
animate_pca(flea_scores, pc_coefs = flea_coefs)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-little.r
\name{little_tour}
\alias{little_tour}
\title{A little tour path.}
\usage{
little_tour(d = 2, reorder = TRUE)
}
\arguments{
\item{d}{target dimensionality}

\item{reorder}{logical, default = TRUE, to randomise the order in which}
}
\description{
The little tour is a planned tour that travels between all axis parallel
projections. (John McDonald named this type of tour.)
}
\details{
Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
animate_xy(flea[, 1:6], little_tour())
animate_pcp(flea[, 1:6], little_tour(3))
animate_scatmat(flea[, 1:6], little_tour(3))
animate_pcp(flea[, 1:6], little_tour(4))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-faces.r
\name{display_faces}
\alias{display_faces}
\alias{animate_faces}
\title{Chernoff faces tour path animation.}
\usage{
display_faces(...)

animate_faces(data, tour_path = grand_tour(3), ...)
}
\arguments{
\item{...}{other arguments passed on to \code{\link{animate}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a nD tour path with Chernoff's faces.  Can display up to 18
dimensions.
}
\details{
This function requires the \code{aplpack} package to draw the
Chernoff faces.  See \code{\link[aplpack]{faces}} for more details.
}
\examples{
# The drawing code is fairly slow, so this animation works best with a
# limited number of cases
flea_s <- rescale(flea[,1:6])
animate_faces(flea_s[19:24, 1:6])

animate_faces(flea_s[19:24, 1:6], grand_tour(5))
}
\seealso{
\code{\link{animate}} for options that apply to all animations
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-andrews.r
\name{display_andrews}
\alias{display_andrews}
\alias{animate_andrews}
\title{Andrews' curves tour path animation.}
\usage{
display_andrews(col = "black", palette = "Zissou 1", ...)

animate_andrews(data, tour_path = grand_tour(3), col = "black", ...)
}
\arguments{
\item{col}{color to be plotted.  Defaults to "black"}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}

\item{...}{other arguments passed on to \code{\link{animate}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a nD tour path with Andrews' curves.  For more details about
Andrew's curves, see \code{\link{andrews}}
}
\examples{
animate_andrews(flea[, 1:6])
animate_andrews(flea[, 1:6], grand_tour(d = 3))
animate_andrews(flea[, 1:6], grand_tour(d = 6))

# It's easy to experiment with different tour paths:
animate_andrews(flea[, 1:6], guided_tour(cmass()))
}
\seealso{
\code{\link{animate}} for options that apply to all animations
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path-index.r
\name{path_index}
\alias{path_index}
\title{Compute index values for a tour history.}
\usage{
path_index(history, index_f, data = attr(history, "data"))
}
\arguments{
\item{history}{list of bases produced by \code{\link{save_history}}
(or otherwise)}

\item{index_f}{index function to apply to each basis}

\item{data}{dataset to be projected on to bases}
}
\description{
Compute index values for a tour history.
}
\examples{
fl_holes <- save_history(flea[, 1:6], guided_tour(holes()), sphere = TRUE)
path_index(fl_holes, holes())
## path_index(fl_holes, cmass())

plot(path_index(fl_holes, holes()), type = "l")
## plot(path_index(fl_holes, cmass()), type = "l")

\donttest{
# Use interpolate to show all intermediate bases as well
hi <- path_index(interpolate(fl_holes), holes())
hi
plot(hi)
}
}
\seealso{
\code{\link{save_history}} for options to save history
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{rescale}
\alias{rescale}
\title{Rescale a matrix or data frame}
\usage{
rescale(df)
}
\arguments{
\item{df}{data frame or matrix}
}
\description{
Standardise each column to have range [0, 1]
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-sage.R, R/section-pursuit.r
\name{cumulative_radial}
\alias{cumulative_radial}
\title{Calculate radial cumulative distribtuion}
\usage{
cumulative_radial(r, R, p)

cumulative_radial(r, R, p)
}
\arguments{
\item{r}{projected radius}

\item{R}{radius of pD sphere}

\item{p}{input dimension}
}
\value{
fraction of volume
}
\description{
Calculate fraction of volume of a pD sphere with radius R projected
within the projected radius r.

Fraction of points within radius r given 2D projection of a
hypersphere with radius R in p dimensions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-guided-anomaly.r
\name{guided_anomaly_tour}
\alias{guided_anomaly_tour}
\title{A guided anomaly tour path.}
\usage{
guided_anomaly_tour(
  index_f,
  d = 2,
  alpha = 0.5,
  cooling = 0.99,
  max.tries = 25,
  max.i = Inf,
  ellipse,
  ellc = NULL,
  ellmu = NULL,
  search_f = search_geodesic,
  ...
)
}
\arguments{
\item{index_f}{the section pursuit index function to optimise. The function
needs to take two arguments, the projected data, indexes of anomalies.}

\item{d}{target dimensionality}

\item{alpha}{the initial size of the search window, in radians}

\item{cooling}{the amount the size of the search window should be adjusted
by after each step}

\item{max.tries}{the maximum number of unsuccessful attempts to find
a better projection before giving up}

\item{max.i}{the maximum index value, stop search if a larger value is found}

\item{ellipse}{pxp variance-covariance matrix defining ellipse, default NULL.
Useful for comparing data with some hypothesized null.}

\item{ellc}{This can be considered the equivalent of a critical value, used to
scale the ellipse larger or smaller to capture more or fewer anomalies. Default 3.}

\item{ellmu}{This is the centre of the ellipse corresponding to the mean of the
normal population. Default vector of 0's}

\item{search_f}{the search strategy to use}

\item{...}{arguments sent to the search_f}
}
\description{
The guided anomaly tour is a variation of the guided tour that is
using an ellipse to determine anomalies on which to select target planes.
}
\details{
Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate_xy}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
animate_xy(flea[, 1:6], guided_anomaly_tour(anomaly_index(),
  ellipse=cov(flea[,1:6])), ellipse=cov(flea[,1:6]), axes="off")
}
\seealso{
\code{\link{slice_index}} for an example of an index functions.
\code{\link{search_geodesic}}, \code{\link{search_better}},
  \code{\link{search_better_random}} for different search strategies
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{Tropical Atmosphere Ocean}
\alias{Tropical Atmosphere Ocean}
\alias{tao}
\title{Tropical Atmosphere Ocean data}
\format{
A 736 x 8 numeric array
}
\description{
This is a subset of data taken from the NOAA web site
https://www.pmel.noaa.gov/tao/. The data is generated from recording
instruments on a grid of buoys laid out over the Pacific Ocean.  The grid
was setup to monitor El Nino and La Nina events. This subset contains
measurements from 5 locations (0deg/110W, 2S/110W, 0deg/95W,2S/95W,5S/95W)
and two time points Nov-Jan 1993 (normal), 1997 (El Nino). There are missing
values in this data set, which need to be removed, or imputed before running
a tour.
}
\references{
https://www.pmel.noaa.gov/tao/
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-xy.r
\name{draw_tour_axes}
\alias{draw_tour_axes}
\title{Draw tour axes on the projected data with base graphics}
\usage{
draw_tour_axes(
  proj,
  labels,
  limits = 1,
  position = "center",
  axis.col = "grey50",
  axis.lwd = 1,
  axis.text.col = "grey50",
  longlabels = FALSE,
  ...
)
}
\arguments{
\item{proj}{matrix of projection coefficients}

\item{labels}{variable names for the axes, of length the same
as the number of rows of proj}

\item{limits}{value setting the lower and upper limits of
projected data, default 1}

\item{position}{position of the axes: center (default),
bottomleft or off}

\item{axis.col}{colour of axes, default "grey50"}

\item{axis.lwd}{linewidth of axes, default 1}

\item{axis.text.col}{colour of axes text, default "grey50"}

\item{longlabels}{text labels only for the long axes in a projection, default FALSE}

\item{...}{other arguments passed}
}
\description{
Draw tour axes on the projected data with base graphics
}
\examples{
data(flea)
flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
prj <- basis_random(ncol(flea[,1:6]), 2)
flea_prj <- as.data.frame(as.matrix(flea_std) \%*\% prj)
par(pty = "s", mar = rep(0.1, 4))
plot(flea_prj$V1, flea_prj$V2,
     xlim = c(-3, 3), ylim = c(-3, 3),
     xlab="P1", ylab="P2")
draw_tour_axes(prj, colnames(flea)[1:6], limits=3)

plot(flea_prj$V1, flea_prj$V2,
     xlim = c(-3, 3), ylim = c(-3, 3),
     xlab="P1", ylab="P2")
draw_tour_axes(prj, colnames(flea)[1:6], limits=3, position="bottomleft")
draw_tour_axes(prj, colnames(flea)[1:6], longlabels=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-image.r
\name{display_image}
\alias{display_image}
\alias{animate_image}
\title{Image tour path animation.}
\usage{
display_image(xs, ys, ...)

animate_image(data, tour_path = grand_tour(1), ...)
}
\arguments{
\item{xs}{x limit that is used in making the size of the plot}

\item{ys}{y limit that is used in making the size of the plot}

\item{...}{other arguments passed on to \code{\link{animate}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a 1d tour path with an image plot.  This animation requires a
different input data structure, a 3d array.  The first two dimensions are
locations on a grid, and the 3rd dimension gives the observations to be
mixed with the tour.
}
\examples{
str(ozone)
animate_image(ozone)
}
\seealso{
\code{\link{animate}} for options that apply to all animations
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{mapShapes}
\alias{mapShapes}
\title{Map vector of factors to pch}
\usage{
mapShapes(x, shapeset)
}
\arguments{
\item{x}{vector}

\item{shapeset}{vector of integers indicating point shapes}
}
\description{
Map vector of factors to pch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear-algebra.r
\name{orthonormalise_by}
\alias{orthonormalise_by}
\title{Orthonormalise one matrix by another.}
\usage{
orthonormalise_by(x, by)
}
\arguments{
\item{x}{numeric matrix}

\item{by}{numeric matrix, same size as x}
}
\value{
orthonormal numeric matrix
}
\description{
This ensures that each column in x is orthogonal to the corresponding
column in by.
}
\keyword{algebra}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-depth.r
\name{display_depth}
\alias{display_depth}
\alias{animate_depth}
\title{Display 3d projection with depth cues}
\usage{
display_depth(center = TRUE, half_range = NULL, ...)

animate_depth(data, tour_path = grand_tour(3), ...)
}
\arguments{
\item{center}{should projected data be centered to have mean zero (default:
TRUE). This pins the centre of the data to the same place, and makes it
easier to focus on the shape.}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{...}{other arguments passed on to \code{\link{animate}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Suggestion to use gray background and colour saturation (instead of
gray shading) by Graham Wills.
}
\examples{
animate_depth(flea[, 1:6])
}
\seealso{
\code{\link{animate}} for options that apply to all animations
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interesting-indices.r
\name{stringy}
\alias{stringy}
\title{Scagnostic indexes.}
\usage{
stringy()
}
\description{
Compute the scagnostic measures from the cassowaryr package
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/section-pursuit.r
\name{slice_index}
\alias{slice_index}
\title{Section pursuit index.}
\usage{
slice_index(
  breaks_x,
  breaks_y,
  eps,
  bintype = "polar",
  power = 1,
  flip = 1,
  reweight = FALSE,
  p = 4
)
}
\arguments{
\item{breaks_x}{binning on the first variable (x or radius).}

\item{breaks_y}{binning on the second variable (y or angle).}

\item{eps}{cutoff values to suppress summing up small differences.
Vector with one entry for each bin, can be estimated
using \code{\link{estimate_eps}}.}

\item{bintype}{select polar (default) or square binning.}

\item{power}{exponent q used in the index compuatation.}

\item{flip}{sign of the index computation, select +1 when searching
for low densities and -1 when searching for high densities.}

\item{reweight}{if TRUE will reweight according to the expected distribution
in a uniform hypersphere (default is FALSE).}

\item{p}{number of variables in the data (needed for accurate reweighting,
default is 4).}
}
\description{
Calculates a section pursuit index that compares the distribution
inside and outside a slice.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear-algebra.r
\name{mahal_dist}
\alias{mahal_dist}
\title{Calculate the Mahalanobis distance between points and center.}
\usage{
mahal_dist(x, vc)
}
\arguments{
\item{x}{matrix of data}

\item{vc}{pre-determined variance-covariance matrix}
}
\description{
Computes the Mahalanobis distance using a provided variance-covariance
matrix of observations from 0.
}
\keyword{algebra}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/section-pursuit.r
\name{estimate_eps}
\alias{estimate_eps}
\title{Estimate cutoff eps for section pursuit.}
\usage{
estimate_eps(N, p, res, K, K_theta, r_breaks)
}
\arguments{
\item{N}{total number of points in the input data.}

\item{p}{number of dimensions of the input data.}

\item{res}{resolution, (slice radius)/(data radius)}

\item{K}{total number of bins}

\item{K_theta}{number of angular bins}

\item{r_breaks}{boundaries of the radial bins}
}
\description{
Estimate cutoff eps for section pursuit.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{Olive oil measurements}
\alias{Olive oil measurements}
\alias{olive}
\title{Olive oil samples from Italy}
\format{
A 572 x 10 numeric array
}
\description{
This data is from a paper by Forina, Armanino, Lanteri, Tiscornia (1983)
Classification of Olive Oils from their Fatty Acid Composition, in Martens
and Russwurm (ed) Food Research and Data Anlysis. We thank Prof. Michele
Forina, University of Genova, Italy for making this dataset available.
}
\details{
\itemize{
  \item region Three super-classes of Italy: North, South and the
    island of Sardinia
  \item area Nine collection areas: three from North, four from South
    and 2 from Sardinia
  \item palmitic, palmitoleic, stearic, oleic, linoleic, linolenic,
    arachidic, eicosenoic fatty acids percent x 100
}
}
\examples{

head(olive)
animate_xy(olive[, c(7, 9, 10)])
animate_xy(olive[, c(7, 9, 10)], col = olive[, 1])
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{sphere_data}
\alias{sphere_data}
\title{Sphere a matrix (or data frame) by transforming variables to
principal components.}
\usage{
sphere_data(df)
}
\arguments{
\item{df}{data frame or matrix}
}
\description{
Sphering is often useful in conjunction with the guided tour, as it
removes simpler patterns that may conceal more interesting findings.
}
\keyword{manip}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{Places Ratings}
\alias{Places Ratings}
\alias{places}
\title{Ratings of different locations across North America}
\format{
A 329 x 14 numeric array
}
\description{
The "places data" were distributed to interested ASA members a few years ago
so that they could apply contemporary data analytic methods to describe
these data and then present results in a poster session at the ASA annual
conference.  Latitude and longitude have been added by Paul Tukey.
}
\details{
____________________________________________________________________

The first dataset is taken from the Places Rated Almanac, by Richard Boyer
and David Savageau, copyrighted and published by Rand McNally. This book
order (SBN) number is 0-528-88008-X, and it retails for $14.95 .  The data
are reproduced on disk by kind permission of the publisher, and with the
request that the copyright notice of Rand McNally, and the names of the
authors appear in any paper or presentation using these data.

The nine rating criteria used by Places Rated Almanac are: Climate and
Terrain Housing Health Care and Environment Crime Transportation Education
The Arts Recreation Economics

For all but two of the above criteria, the higher the score, the better.
For Housing and Crime, the lower the score the better.

The scores are computed using the following component statistics for each
criterion (see the Places Rated Almanac for details):

Climate and Terrain: very hot and very cold months, seasonal temperature
variation, heating- and cooling-degree days, freezing days, zero-degree
days, ninety-degree days.

Housing: utility bills, property taxes, mortgage payments.

Health Care and Environment: per capita physicians, teaching hospitals,
medical schools, cardiac rehabilitation centers, comprehensive cancer
treatment centers, hospices, insurance/hospitalization costs index,
flouridation of drinking water, air pollution.

Crime: violent crime rate, property crime rate.

Transportation: daily commute, public transportation, Interstate highways,
air service, passenger rail service.

Education: pupil/teacher ratio in the public K-12 system, effort index in
K-12, accademic options in higher education.

The Arts: museums, fine arts and public radio stations, public television
stations, universities offering a degree or degrees in the arts, symphony
orchestras, theatres, opera companies, dance companies, public libraries.

Recreation: good restaurants, public golf courses, certified lanes for
tenpin bowling, movie theatres, zoos, aquariums, family theme parks,
sanctioned automobile race tracks, pari-mutuel betting attractions, major-
and minor- league professional sports teams, NCAA Division I football and
basketball teams, miles of ocean or Great Lakes coastline, inland water,
national forests, national parks, or national wildlife refuges, Consolidated
Metropolitan Statistical Area access.

Economics: average household income adjusted for taxes and living costs,
income growth, job growth.
}
\examples{

head(places)
animate_xy(places[, 1:9])
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-radial.r
\name{radial_tour}
\alias{radial_tour}
\title{A radial tour path.}
\usage{
radial_tour(start, mvar = 1, ...)
}
\arguments{
\item{start}{initial projection matrix}

\item{mvar}{variable(s) chosen to rotate out}

\item{...}{additional arguments for drawing}
}
\description{
The radial tour rotates a chosen variable axis out of the current
projection.
}
\details{
Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
animate_xy(flea[, 1:6], radial_tour(basis_random(6, 2), mvar = 4), rescale=TRUE)
animate_xy(flea[, 1:6], radial_tour(basis_random(6, 2), mvar = c(3,4)), rescale=TRUE)
animate_dist(flea[, 1:6], radial_tour(basis_random(6, 1), mvar = 4), rescale=TRUE)
animate_scatmat(flea[, 1:6], radial_tour(basis_random(6, 3), mvar = 4), rescale=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/animate.r
\name{animate}
\alias{animate}
\title{Animate a tour path.}
\usage{
animate(
  data,
  tour_path = grand_tour(),
  display = display_xy(),
  start = NULL,
  aps = 1,
  fps = 10,
  max_frames = Inf,
  rescale = FALSE,
  sphere = FALSE,
  ...
)
}
\arguments{
\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}

\item{display}{takes the display that is suppose to be used, defaults to
the xy display}

\item{start}{projection to start at, if not specified, uses default
associated with tour path}

\item{aps}{target angular velocity (in radians per second)}

\item{fps}{target frames per second (defaults to 15, to accommodate RStudio graphics device)}

\item{max_frames}{the maximum number of bases to generate.  Defaults to
Inf for interactive use (must use Ctrl + C to terminate), and 1 for
non-interactive use.}

\item{rescale}{Default FALSE. If TRUE, rescale all variables to range [0,1]?}

\item{sphere}{if true, sphere all variables}

\item{...}{ignored}
}
\value{
an (invisible) list of bases visited during this tour
}
\description{
This is the function that powers all of the tour animations.  If you want
to write your own tour animation method, the best place to
start is by looking at the code for animation methods that have already
implemented in the package.
}
\details{
See \code{\link{render}} to render animations to disk.
}
\examples{
f <- flea[, 1:6]
animate(f, grand_tour(), display_xy())
# or in short
animate(f)
animate(f, max_frames = 30)
\donttest{
animate(f, max_frames = 10, fps = 1, aps = 0.1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear-algebra.r
\name{is_orthonormal}
\alias{is_orthonormal}
\title{Test if a numeric matrix is orthonormal.}
\usage{
is_orthonormal(x, tol = 0.001)
}
\arguments{
\item{x}{numeric matrix}

\item{tol}{tolerance used to test floating point differences}
}
\description{
Test if a numeric matrix is orthonormal.
}
\keyword{algebra}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{center}
\alias{center}
\title{Center a numeric vector by subtracting off its mean.}
\usage{
center(x)
}
\arguments{
\item{x}{numeric vector}
}
\description{
Center a numeric vector by subtracting off its mean.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path-dist.r
\name{path_dist}
\alias{path_dist}
\title{Compute distance matrix from bases.}
\usage{
path_dist(history)
}
\arguments{
\item{history}{history of the plots}
}
\description{
Compute distance matrix from bases.
}
\examples{
# This code is to be used as an example but you should increase
# the max from 2 to 50, say, to check tour coverage.
flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
grand <- interpolate(save_history(flea_std, max = 2), 0.2)
# The grand tour  -----------------------------
# Look at the tour path in a tour, how well does it cover a sphere
# Using MDS to summarise the high-d space of projections
# Last basis is a duplicate, needs removing
d <- path_dist(grand[,,-dim(grand)[[3]]])
ord <- as.data.frame(MASS::isoMDS(d)$points)
require(ggplot2)
ggplot(data = ord, aes(x=V1, y=V2)) +
  geom_path() +
  coord_equal() +
  labs(x = NULL, y = NULL)

# Compare five guided tours  -----------------------------
holes1d <- guided_tour(holes(), 1)
tour_reps <- replicate(5, save_history(flea_std, holes1d, max = 2),
  simplify = FALSE
)
tour_reps2 <- lapply(tour_reps, interpolate, 0.2)

bases <- unlist(lapply(tour_reps2, as.list), recursive = FALSE)
class(bases) <- "history_list"
index_values <- paths_index(tour_reps2, holes())
index_values$step <- index_values$step.1
d <- path_dist(bases)
ord <- as.data.frame(cmdscale(d, 2))

info <- cbind(ord, index_values)
ggplot(data = info, aes(x = step, y = value, group = try)) +
  geom_line()
##ggplot(data = info, aes(x = V1, y = V2, group = try)) +
##  geom_path() +
##  geom_point(aes(size = value)) +
##  coord_equal()
##last_plot() + facet_wrap(~try)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geodesic.r
\name{geodesic_info}
\alias{geodesic_info}
\title{Calculate information required to interpolate along a geodesic path between
two frames.}
\usage{
geodesic_info(Fa, Fz, epsilon = 1e-06)
}
\arguments{
\item{Fa}{starting frame, will be orthonormalised if necessary}

\item{Fz}{target frame, will be orthonormalised if necessary}

\item{epsilon}{epsilon used to determine if an angle is effectively equal
to 0}
}
\description{
The methdology is outlined in
\url{http://www-stat.wharton.upenn.edu/~buja/PAPERS/paper-dyn-proj-algs.pdf}
and
\url{http://www-stat.wharton.upenn.edu/~buja/PAPERS/paper-dyn-proj-math.pdf},
and the code follows the notation outlined in those papers:
}
\details{
\itemize{
  \item p = dimension of data
  \item d = target dimension
  \item F = frame, an orthonormal p x d matrix
  \item Fa = starting frame, Fz = target frame
  \item Fa'Fz = Va lamda  Vz' (svd)
  \item Ga = Fa Va, Gz = Fz Vz
  \item tau = principle angles
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interesting-indices.r
\name{pda_pp}
\alias{pda_pp}
\title{PDA projection pursuit index.}
\usage{
pda_pp(cl, lambda = 0.2)
}
\arguments{
\item{cl}{class to be used.  Such as "color"}

\item{lambda}{shrinkage parameter (0 = no shrinkage, 1 = full shrinkage)}
}
\description{
Calculate the PDA projection pursuit index.  See Lee and Cook (2009)
A Projection Pursuit Index for Large p, Small n Data
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interesting-indices.r
\name{MIC}
\alias{MIC}
\alias{TIC}
\title{Maximum and total information coefficient index.}
\usage{
MIC()

TIC()
}
\description{
Compute the maximum and total information coefficient indexes,
see \code{minerva::mine}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-xy.r
\name{display_xy}
\alias{display_xy}
\alias{animate_xy}
\title{Display tour path with a scatterplot}
\usage{
display_xy(
  center = TRUE,
  axes = "center",
  half_range = NULL,
  col = "black",
  pch = 20,
  cex = 1,
  edges = NULL,
  edges.col = "black",
  edges.width = 1,
  obs_labels = NULL,
  ellipse = NULL,
  ellc = NULL,
  ellmu = NULL,
  ellmarks = TRUE,
  palette = "Zissou 1",
  shapeset = c(15:17, 23:25),
  axislablong = FALSE,
  ...
)

animate_xy(data, tour_path = grand_tour(), ...)
}
\arguments{
\item{center}{if TRUE, centers projected data to (0,0).  This pins the
center of data cloud and make it easier to focus on the changing shape
rather than position.}

\item{axes}{position of the axes: center, bottomleft or off}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{col}{color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".}

\item{pch}{shape of the point to be plotted, can be a factor or integer.  Defaults to 20.}

\item{cex}{size of the point to be plotted.  Defaults to 1.}

\item{edges}{A two column integer matrix giving indices of ends of lines.}

\item{edges.col}{colour of edges to be plotted, Defaults to "black"}

\item{edges.width}{line width for edges, default 1}

\item{obs_labels}{vector of text labels to display}

\item{ellipse}{pxp variance-covariance matrix defining ellipse, default NULL. Useful for
comparing data with some null hypothesis}

\item{ellc}{This can be considered the equivalent of a critical value, used to
scale the ellipse larger or smaller to capture more or fewer anomalies. Default 3.}

\item{ellmu}{This is the centre of the ellipse corresponding to the mean of the
normal population. Default vector of 0's}

\item{ellmarks}{mark the extreme points with red crosses, default TRUE}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}

\item{shapeset}{numbers corresponding to shapes in base R points, to use for mapping
categorical variable to shapes, default=c(15:17, 23:25)}

\item{axislablong}{text labels only for the long axes in a projection, default FALSE}

\item{...}{other arguments passed on to \code{\link{animate}} and
\code{\link{display_xy}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a 2D tour path with a scatterplot.
}
\examples{
animate_xy(flea[, 1:6])
animate(flea[, 1:6], tour_path = grand_tour(), display = display_xy())
# This won't do anything because the flea data is standardised
# but use rescale option to force scaling before displaying
animate(flea[, 1:6],
  tour_path = grand_tour(),
  display = display_xy(),
  rescale = TRUE
)
animate(flea[, 1:6],
  tour_path = grand_tour(),
  display = display_xy(half_range = 0.5)
)
animate_xy(flea[, 1:6], tour_path = little_tour())
animate_xy(flea[, 1:3], tour_path = guided_tour(holes()), sphere = TRUE)
animate_xy(flea[, 1:6], center = FALSE)

# The default axes are centered, like a biplot, but there are other options
animate_xy(flea[, 1:6], axes = "bottomleft")
animate_xy(flea[, 1:6], axes = "off")
animate_xy(flea[, 1:6], dependence_tour(c(1, 2, 1, 2, 1, 2)),
  axes = "bottomleft"
)

animate_xy(flea[, -7], col = flea$species)
animate_xy(flea[, -7], col = flea$species,
             pch = flea$species)

animate_xy(flea[, -7], col = flea$species,
  obs_labels=as.character(1:nrow(flea)), axes="off")

# You can also draw lines
edges <- matrix(c(1:5, 2:6), ncol = 2)
animate(
  flea[, 1:6], grand_tour(),
  display_xy(axes = "bottomleft", edges = edges)
)
# An ellipse can be drawn on the data using a specified var-cov
animate_xy(flea[, 1:6], axes = "off", ellipse=cov(flea[,1:6]))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-grand.r
\name{basis_random}
\alias{basis_random}
\title{Generate a random basis}
\usage{
basis_random(n, d = 2)
}
\arguments{
\item{n}{dimensionality of data}

\item{d}{dimensionality of target projection}
}
\description{
Generate a random basis
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour.r
\name{new_tour}
\alias{new_tour}
\title{Create a new tour.}
\usage{
new_tour(data, tour_path, start = NULL, ...)
}
\arguments{
\item{data}{the data matrix to be projected}

\item{tour_path}{basis generator, a function that generates a new basis,
called with the previous projection and the data set.  For more
complicated tour paths, this will need to be a closure with local
variables.  Should return NULL if the tour should terminate}

\item{start}{starting projection, if omitted will use default projection
from generator}
}
\value{
a function with single argument, step_size.  This function returns
 a list containing the new projection, the current target and the number
 of steps taken towards the target.
}
\description{
The tour function provides the common machinery behind all tour methods:
interpolating from basis to basis, and generating new bases when necessary.
You should not have to call this function.
}
\details{
If you are intended to call \code{new_tour()} from the global environment, try \code{save_history()} and then animate with a \code{planned_tour()}. See \code{\link[tourr]{save_history}} for examples on this.
}
\seealso{
\code{\link{save_history}}, \code{\link{render}} and
  \code{\link{animate}} for examples of functions that use this function
  to run dynamic tours.
}
\keyword{dynamic}
\keyword{hplot}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-sage.R
\name{display_sage}
\alias{display_sage}
\alias{animate_sage}
\title{Display tour path with a sage scatterplot}
\usage{
display_sage(
  axes = "center",
  half_range = NULL,
  col = "black",
  pch = 20,
  gam = 1,
  R = NULL,
  palette = "Zissou 1",
  axislablong = FALSE,
  ...
)

animate_sage(data, tour_path = grand_tour(), ...)
}
\arguments{
\item{axes}{position of the axes: center, bottomleft or off}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{col}{color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".}

\item{pch}{marker for points. Defaults to 20.}

\item{gam}{scaling of the effective dimensionality for rescaling. Defaults to 1.}

\item{R}{scale for the radial transformation.
If not set, defaults to maximum distance from origin to each row of data.}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}

\item{axislablong}{text labels only for the long axes in a projection, default FALSE}

\item{...}{other arguments passed on to \code{\link{animate}} and
\code{\link{display_sage}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a 2D tour path with a sage scatterplot that
uses a radial transformation on the projected points to re-allocate
the volume projected across the 2D plane.
}
\examples{
# Generate uniform samples in a 10d sphere using the geozoo package
sphere10 <- geozoo::sphere.solid.random(10)$points
# Columns need to be named before launching the tour
colnames(sphere10) <- paste0("x", 1:10)
# Standard grand tour display, points cluster near center
animate_xy(sphere10)
# Sage display, points are uniformly distributed across the disk
animate_sage(sphere10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/section-pursuit.r
\name{radial_bin_weight_inv}
\alias{radial_bin_weight_inv}
\title{Inverse weights for rescaling counts in radial bins.}
\usage{
radial_bin_weight_inv(r1, r2, R, p)
}
\description{
Inverse weights for rescaling counts in radial bins.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-stereo.r
\name{project3d}
\alias{project3d}
\title{Stereographic projection}
\usage{
project3d(d3, length = par("din")[1] * 25.4, z0 = 300, d = 30)
}
\arguments{
\item{d3}{3d numeric matrix giving position of points}

\item{length}{width of plot, mm}

\item{z0}{distance from eye to screen, mm}

\item{d}{half interpupilary distance, mm}
}
\description{
Math from http://dogfeathers.com/java/3dproj.html
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-frozen-geodesic.r
\name{find_frozen_path_peak}
\alias{find_frozen_path_peak}
\title{Find most highest peak along frozen geodesic.}
\usage{
find_frozen_path_peak(old, new, frozen, index, max_dist = pi/4)
}
\description{
Find most highest peak along frozen geodesic.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-slice.r
\name{display_slice}
\alias{display_slice}
\alias{animate_slice}
\title{Display tour path with a sliced scatterplot}
\usage{
display_slice(
  center = TRUE,
  axes = "center",
  half_range = NULL,
  col = "black",
  pch_slice = 20,
  pch_other = 46,
  cex_slice = 2,
  cex_other = 1,
  v_rel = NULL,
  anchor = NULL,
  anchor_nav = "off",
  edges = NULL,
  edges.col = "black",
  palette = "Zissou 1",
  axislablong = FALSE,
  ...
)

animate_slice(data, tour_path = grand_tour(), rescale = FALSE, ...)
}
\arguments{
\item{center}{if TRUE, centers projected data to (0,0).  This pins the
center of data cloud and make it easier to focus on the changing shape
rather than position.}

\item{axes}{position of the axes: center, bottomleft or off}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{col}{color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".}

\item{pch_slice}{marker for plotting points inside the slice.
Defaults to 20.}

\item{pch_other}{marker for plotting points outside the slice.
Defaults to 46.}

\item{cex_slice}{size of the points inside the slice. Defaults to 2.}

\item{cex_other}{size if the points outside the slice. Defaults to 1.}

\item{v_rel}{relative volume of the slice. If not set, suggested value
is calculated and printed to the screen.}

\item{anchor}{A vector specifying the reference point to anchor the slice.
If NULL (default) the slice will be anchored at the data center.}

\item{anchor_nav}{position of the anchor: center, topright or off}

\item{edges}{A two column integer matrix giving indices of ends of lines.}

\item{edges.col}{colour of edges to be plotted, Defaults to "black.}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}

\item{axislablong}{text labels only for the long axes in a projection, default FALSE}

\item{...}{other arguments passed on to \code{\link{animate}} and
\code{\link{display_slice}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}

\item{rescale}{Default FALSE. If TRUE, rescale all variables to range [0,1].}
}
\description{
Animate a 2D tour path with a sliced scatterplot.
}
\examples{
# Generate samples on a 3d and 5d hollow sphere using the geozoo package
sphere3 <- geozoo::sphere.hollow(3)$points
sphere5 <- geozoo::sphere.hollow(5)$points

# Columns need to be named before launching the tour
colnames(sphere3) <- c("x1", "x2", "x3")
colnames(sphere5) <- c("x1", "x2", "x3", "x4", "x5")

# Animate with the slice display using the default parameters
animate_slice(sphere3)
animate_slice(sphere5)

# Animate with off-center anchoring
anchor3 <- matrix(rep(0.7, 3), ncol=3)
anchor5 <- matrix(rep(0.3, 5), ncol=5)
animate_slice(sphere3, anchor = anchor3)
# Animate with thicker slice to capture more points in each view
animate_slice(sphere5, anchor = anchor5, v_rel = 0.02)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-grand.r
\name{grand_tour}
\alias{grand_tour}
\title{A grand tour path.}
\usage{
grand_tour(d = 2, ...)
}
\arguments{
\item{d}{target dimensionality}

\item{...}{arguments sent to the generator}
}
\description{
This method generates target bases by randomly sampling on
the space of all d-dimensional planes in p-space.
}
\details{
Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
# All animation methods use the grand tour path by default
animate_dist(flea[, 1:6])
animate_xy(flea[, 1:6])
animate_pcp(flea[, 1:6])
animate_pcp(flea[, 1:6], grand_tour(4))

# The grand tour is a function:
tour2d <- grand_tour(2)
is.function(tour2d)

# with two parameters, the previous projection and the data set
args(tour2d)
# if the previous projection is null, it will generate a starting
# basis, otherwise the argument is ignored
tour2d(NULL, mtcars)
# the data argument is just used to determine the correct dimensionality
# of the output matrix
tour2d(NULL, mtcars[, 1:2])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{areColors}
\alias{areColors}
\title{Test if all entries are colors}
\usage{
areColors(x)
}
\arguments{
\item{x}{vector}
}
\description{
Test if all entries are colors
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path-index.r
\name{paths_index}
\alias{paths_index}
\title{Compute index value for many histories.}
\usage{
paths_index(bases_list, index_f)
}
\arguments{
\item{bases_list}{list of histories produced by \code{\link{save_history}}}

\item{index_f}{index function to apply to each projection}
}
\description{
This is a convenience method that returns a data frame summarising the
index values for multiple tour paths.
}
\examples{
# The max.tries is low for satisfying CRAN checks
# Increase it for using in practice
holes1d <- guided_tour(holes(), 1, max.tries=2)
# Perform guided tour 5 times, saving results
tries <- replicate(5, save_history(flea[, 1:6], holes1d), simplify = FALSE)
# Interpolate between target bases
itries <- lapply(tries, interpolate)

paths <- paths_index(itries, holes())
head(paths)

if (require(ggplot2)) {
  ggplot(data = paths, aes(x=step, y=value, group = try)) + geom_line()
  ## ggplot(data = paths, aes(x=step, y=improvement, group = try)) + geom_line()
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-groupxy.r
\name{display_groupxy}
\alias{display_groupxy}
\alias{animate_groupxy}
\title{Display 2D tour projections displayed separately by groups}
\usage{
display_groupxy(
  centr = TRUE,
  axes = "center",
  half_range = NULL,
  col = "black",
  pch = 20,
  cex = 1,
  edges = NULL,
  edges.col = "black",
  edges.width = 1,
  group_by = NULL,
  plot_xgp = TRUE,
  palette = "Zissou 1",
  shapeset = c(15:17, 23:25),
  axislablong = FALSE,
  ...
)

animate_groupxy(data, tour_path = grand_tour(), ...)
}
\arguments{
\item{centr}{if TRUE, centers projected data to (0,0).  This pins the
center of data cloud and make it easier to focus on the changing shape
rather than position.}

\item{axes}{position of the axes: center, bottomleft or off}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{col}{color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".}

\item{pch}{shape of the point to be plotted.  Defaults to 20.}

\item{cex}{size of the point to be plotted.  Defaults to 1.}

\item{edges}{A two column integer matrix giving indices of ends of lines.}

\item{edges.col}{colour of edges to be plotted, Defaults to "black"}

\item{edges.width}{line width for edges, default 1}

\item{group_by}{variable to group by. Must have less than 25 unique values.}

\item{plot_xgp}{if TRUE, plots points from other groups in light grey}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}

\item{shapeset}{numbers corresponding to shapes in base R points, to use for mapping
categorical variable to shapes, default=c(15:17, 23:25)}

\item{axislablong}{text labels only for the long axes in a projection, default FALSE}

\item{...}{other arguments passed on to \code{\link{animate}} and
\code{\link{display_groupxy}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
This function is designed to allow comparisons across multiple groups,
 especially for examining things like two (or more) different models
 on the same data. The primary display is a scatterplot, with lines or
 contours overlaid.
}
\examples{
animate_groupxy(flea[, 1:6], col = flea$species,
  pch = flea$species, group_by = flea$species)
animate_groupxy(flea[, 1:6], col = flea$species,
  pch = flea$species, group_by = flea$species,
  plot_xgp = FALSE)
# Edges example
x <- data.frame(x1=runif(10, -1, 1), x2=runif(10, -1, 1), x3=runif(10, -1, 1))
x$cl <- factor(c(rep("A", 3), rep("B", 3), rep("C", 4)))
x.edges <- cbind(from=c(1,2, 4,5, 7,8,9), to=c(2,3, 5,6, 8,9,10))
x.edges.col <- factor(c(rep("A", 2), rep("B", 2), rep("C", 3)))
animate_groupxy(x[,1:3], col=x$cl, group_by=x$cl, edges=x.edges, edges.col=x.edges.col)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search_polish.r
\name{search_polish}
\alias{search_polish}
\title{Search very locally to find slightly better projections to polish a broader search.}
\usage{
search_polish(
  current,
  alpha = 0.5,
  index,
  tries,
  polish_max_tries = 30,
  cur_index = NA,
  n_sample = 100,
  polish_cooling = 1,
  ...
)
}
\arguments{
\item{current}{the current projection basis}

\item{alpha}{the angle used to search the target basis from the current basis}

\item{index}{index function}

\item{tries}{the counter of the outer loop of the opotimiser}

\item{polish_max_tries}{maximum number of iteration before giving up}

\item{cur_index}{the index value of the current basis}

\item{n_sample}{number of samples to generate}

\item{polish_cooling}{percentage of reduction in polish_alpha when no better basis is found}

\item{...}{other arguments being passed into the \code{search_polish()}}
}
\description{
Search very locally to find slightly better projections to polish a broader search.
}
\examples{
data(t1)
best_proj <- t1[, , dim(t1)[3]]
attr(best_proj, "data") <- NULL
best_proj <- unclass(drop(best_proj))
animate_xy(
  flea[, 1:6],
  guided_tour(holes()),
    search_f = search_polish(
         polish_max_tries = 5),
  start = best_proj
)
}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-planned.r
\name{planned_tour}
\alias{planned_tour}
\alias{planned2_tour}
\title{A planned tour path.}
\usage{
planned_tour(basis_set, cycle = FALSE)

planned2_tour(basis_set)
}
\arguments{
\item{basis_set}{the set of bases as a list of projection matrices
or a 3d array}

\item{cycle}{cycle through continuously (\code{TRUE}) or stop after
first pass (\code{FALSE})}
}
\description{
The planned tour takes you from one basis to the next in a
set order.  Once you have visited all the planned bases, you either stop
or start from the beginning once more (if \code{cycle = TRUE}).
}
\details{
Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
twod <- save_history(flea[, 1:3], max = 5)
str(twod)
animate_xy(flea[, 1:3], planned_tour(twod))
animate_xy(flea[, 1:3], planned_tour(twod, TRUE))

oned <- save_history(flea[, 1:6], grand_tour(1), max = 3)
animate_dist(flea[, 1:6], planned_tour(oned))
}
\seealso{
The \code{\link{little_tour}}, a special type of planned tour
  which cycles between all axis parallel projections.
}
\keyword{dynamic}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{Ozone measurements}
\alias{Ozone measurements}
\alias{ozone}
\title{Monthly ozone measurements over Central America}
\format{
A 24 x 24 x 72 numeric array
}
\description{
This data set is a subset of the data from the 2006 ASA Data expo challenge.
The data are monthly ozone averages on a very coarse 24 by 24 grid covering
Central America, from Jan 1995 to Dec 2000. The data is stored in a 3d area
with the first two dimensions representing latitude and longitude, and the
third representing
time.
}
\examples{

example(display_image)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.r
\name{render_anim}
\alias{render_anim}
\title{Render a set of animation frames}
\usage{
render_anim(
  data,
  vars = NULL,
  frames,
  edges = NULL,
  axis_labels = NULL,
  obs_labels = NULL,
  limits = 1,
  position = "center"
)
}
\arguments{
\item{data}{matrix, or data frame containing numeric columns,
should be standardised to have mean 0, sd 1}

\item{vars}{numeric columns of data to be projected, as a vector, eg 1:4}

\item{frames}{array of projection matrices, should be interpolated already}

\item{edges}{to and from of row id's to connect with an line}

\item{axis_labels}{labels of the axes to be displayed}

\item{obs_labels}{labels of the observations to be available for interactive mouseover}

\item{limits}{value setting the lower and upper limits of
projected data, default 1}

\item{position}{position of the axes: center (default),
left of data or off}
}
\value{
list containing indexed projected data, edges, circle and segments for axes
}
\description{
This function takes a set of frames as produced by
save_history(), and creates the projected data and axes
in for format needed to create the animation using plotly.
It will be useful for showing a tour where mouseover can
be used to identify points.
Note that for now this only works for 2D projections.
}
\examples{
data(flea)
flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
t1 <- save_history(flea_std, max=2)
t1i <- tourr::interpolate(t1, 0.1)
p <- render_anim(data=flea_std, frames=t1i)
if (require(ggplot2)) {
  pg <- ggplot() +
    geom_path(data=p$circle, aes(x=c1, y=c2, frame=frame)) +
    geom_segment(data=p$axes, aes(x=x1, y=y1, xend=x2, yend=y2, frame=frame)) +
    geom_text(data=p$axes, aes(x=x2, y=y2, frame=frame, label=axis_labels)) +
    geom_point(data=p$frames, aes(x=P1, y=P2, frame=frame, label=obs_labels)) +
    coord_equal() +
    theme_bw() +
    theme(axis.text=element_blank(),
        axis.title=element_blank(),
        axis.ticks=element_blank(),
        panel.grid=element_blank())
  if (interactive()) {
    require(plotly)
    ggplotly(pg, width=500, height=500) |>
      animation_button(label="Go") |>
      animation_slider(len=0.8, x=0.5, xanchor="center") |>
      animation_opts(easing="linear", transition=0, redraw=FALSE)
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-frozen-geodesic.r
\name{find_best_frozen_dir}
\alias{find_best_frozen_dir}
\title{Find most promising direction in frozen space.}
\usage{
find_best_frozen_dir(old, frozen, index, dist = 0.01, counter = 5)
}
\description{
Find most promising direction in frozen space.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/section-pursuit.r
\name{angular_breaks}
\alias{angular_breaks}
\title{Returns n equidistant bins between -pi and pi}
\usage{
angular_breaks(n)
}
\arguments{
\item{n}{number of bins}
}
\description{
Returns n equidistant bins between -pi and pi
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geodesic.r
\name{step_angle}
\alias{step_angle}
\title{Step along an interpolated path by angle in radians.}
\usage{
step_angle(interp, angle)
}
\arguments{
\item{interp}{interpolated path}

\item{angle}{angle, in radians}
}
\description{
Step along an interpolated path by angle in radians.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-dependence.r
\name{dependence_tour}
\alias{dependence_tour}
\title{A dependence tour path.}
\usage{
dependence_tour(pos)
}
\arguments{
\item{pos}{a numeric vector describing which variables are mapped to
which dimensions: 1 corresponds to first, 2 to second etc.}
}
\description{
The dependence tour combines a set of independent 1d tours to produce
a nd tour.  For the special case of 2d, this is known as a correlation
tour.  This tour corresponds to the multivariate method known as
generalised canonical correlation, and is used to investigate dependence
between groups of variables.
}
\details{
Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
animate_xy(flea[, 1:3], dependence_tour(c(1, 2, 2)))
animate_xy(flea[, 1:4], dependence_tour(c(1, 2, 1, 2)))
animate_pcp(flea[, 1:6], dependence_tour(c(1, 2, 3, 2, 1, 3)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geodesic.r
\name{step_fraction}
\alias{step_fraction}
\title{Step along an interpolated path by fraction of path length.}
\usage{
step_fraction(interp, fraction)
}
\arguments{
\item{interp}{interpolated path}

\item{fraction}{fraction of distance between start and end planes}
}
\description{
Step along an interpolated path by fraction of path length.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-jellyfish.R
\name{search_jellyfish}
\alias{search_jellyfish}
\alias{check_dup}
\title{A jellyfish optimiser for projection pursuit guided tour}
\usage{
search_jellyfish(current, index, tries, max.tries = 50, verbose = FALSE, ...)

check_dup(bases, min_dist)
}
\arguments{
\item{current}{starting projection, a list of basis of class "multi-bases"}

\item{index}{index function}

\item{tries}{the counter of the outer loop of the opotimiser}

\item{max.tries}{the maximum number of iteration before giving up}

\item{verbose}{whether to print out the progress messages}

\item{...}{other arguments being passed into the \code{search_jellyfish()}}

\item{bases}{a list of bases extracted from the data collection object, see examples}

\item{min_dist}{the minimum distance between two bases}
}
\description{
A jellyfish optimiser for projection pursuit guided tour
}
\examples{
library(dplyr)
res <- animate_xy(flea[, 1:6], guided_tour(lda_pp(cl = flea$species),
search_f = search_jellyfish))
bases <- res |> filter(loop == 1) |> pull(basis) |> check_dup(0.1)
animate_xy(data = flea[,1:6], tour_path = planned_tour(bases), col = flea$species)
}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/section-pursuit.r
\name{linear_breaks}
\alias{linear_breaks}
\title{Returns n equidistant bins between a and b}
\usage{
linear_breaks(n, a, b)
}
\arguments{
\item{n}{number of bins}

\item{a}{lower bound}

\item{b}{upper bound}
}
\description{
Returns n equidistant bins between a and b
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-better.r
\name{search_better_random}
\alias{search_better_random}
\title{Search for a better projection using simulated annealing}
\usage{
search_better_random(
  current,
  alpha = 0.5,
  index,
  tries,
  max.tries = Inf,
  method = "linear",
  cur_index = NA,
  t0 = 0.01,
  ...
)
}
\arguments{
\item{current}{starting projection}

\item{alpha}{the angle used to search the target basis from the current basis}

\item{index}{index function}

\item{tries}{the counter of the outer loop of the opotimiser}

\item{max.tries}{maximum number of iteration before giving up}

\item{method}{whether the nearby bases are found by a linear/ geodesic formulation}

\item{cur_index}{the index value of the current basis}

\item{t0}{initial decrease in temperature}

\item{...}{other arguments being passed into the \code{search_better_random()}}
}
\description{
Given an initial \eqn{t0}, the cooling scheme updates temperature at \deqn{T = t0 /\log(i + 1)}
The candidate basis is sampled via \deqn{B_j = (1 - \alpha) * B_i + \alpha * B} where alpha defines the neighbourhood, \eqn{B_i} is the current basis, B is a randomly generated basis
The acceptance probability is calculated as \deqn{prob = \exp{-abs(I(B_i) - I(B_j))/ T}}
For more information, see
\url{https://projecteuclid.org/download/pdf_1/euclid.ss/1177011077}
}
\examples{
animate_xy(flea[, 1:6], guided_tour(holes(), search_f = search_better_random))
}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interesting-indices.r
\name{splines2d}
\alias{splines2d}
\alias{loess2d}
\title{Spline/loess based index.}
\usage{
splines2d()

loess2d()
}
\description{
Compares the variance in residuals of a fitted
spline/loess model to the overall variance to find
functional dependence in 2D projections
of the data.
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-slice.r
\name{draw_slice_center}
\alias{draw_slice_center}
\title{Draw slice center guide with base graphics}
\usage{
draw_slice_center(anchor, rng, limits, anchor_nav, ...)
}
\description{
Draw slice center guide with base graphics
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-density2d.r
\name{display_density2d}
\alias{display_density2d}
\alias{animate_density2d}
\title{Display tour path with a density and scatterplot}
\usage{
display_density2d(
  center = TRUE,
  axes = "center",
  half_range = NULL,
  col = "black",
  pch = 20,
  cex = 1,
  contour_quartile = c(0.25, 0.5, 0.75),
  edges = NULL,
  palette = "Zissou 1",
  axislablong = FALSE,
  ...
)

animate_density2d(data, tour_path = grand_tour(), ...)
}
\arguments{
\item{center}{if TRUE, centers projected data to (0,0).  This pins the
center of data cloud and make it easier to focus on the changing shape
rather than position.}

\item{axes}{position of the axes: center, bottomleft or off}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{col}{color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".}

\item{pch}{shape of the point to be plotted.  Defaults to 20.}

\item{cex}{size of the point to be plotted.  Defaults to 1.}

\item{contour_quartile}{Vector of quartiles to plot the contours at. Defaults to 5.}

\item{edges}{A two column integer matrix giving indices of ends of lines.}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}

\item{axislablong}{text labels only for the long axes in a projection, default FALSE}

\item{...}{other arguments passed on to \code{\link{animate}} and
\code{\link{display_density2d}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a 2D tour path with density contour(s) and a scatterplot.
}
\examples{
animate_density2d(flea[, 1:6])
animate(flea[, 1:6], tour_path = grand_tour(), display = display_density2d())
animate(flea[, 1:6],
  tour_path = grand_tour(),
  display = display_density2d(axes = "bottomleft")
)
animate(flea[, 1:6],
  tour_path = grand_tour(),
  display = display_density2d(half_range = 0.5)
)
animate_density2d(flea[, 1:6], tour_path = little_tour())

animate_density2d(flea[, 1:3], tour_path = guided_tour(holes()), sphere = TRUE)
animate_density2d(flea[, 1:6], center = FALSE)

# The default axes are centered, like a biplot, but there are other options
animate_density2d(flea[, 1:6], axes = "bottomleft")
animate_density2d(flea[, 1:6], axes = "off")
animate_density2d(flea[, 1:6], dependence_tour(c(1, 2, 1, 2, 1, 2)),
  axes = "bottomleft"
)

animate_density2d(flea[, -7], col = flea$species)

# You can also draw lines
edges <- matrix(c(1:5, 2:6), ncol = 2)
animate(
  flea[, 1:6], grand_tour(),
  display_density2d(axes = "bottomleft", edges = edges)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interesting-indices.r
\name{dcor2d}
\alias{dcor2d}
\alias{dcor2d_2}
\title{Distance correlation index.}
\usage{
dcor2d()

dcor2d_2()
}
\description{
Computes the distance correlation based index on 2D projections of the data.
\code{dcor2d_2} uses the faster implementation of the distance correlation
for bivariate data, see \code{energy::dcor2d}.
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path-andrews.r
\name{plot.path_curve}
\alias{plot.path_curve}
\title{Plot history curves.}
\usage{
\method{plot}{path_curve}(x, ...)
}
\description{
The default plot method is a line plot with step on the x axis and
value on the y axis.  Each observation is drawn with a different line
line and the plot is facetted by variable.  This is rather similar in
spirit to a parallel coordinates plot or Andrews curves.
}
\details{
For alternative ways of plotting this data, see
 \code{\link{path_curves}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{find_platform}
\alias{find_platform}
\title{Find the platform
Find the platform being used by the user}
\usage{
find_platform()
}
\description{
Find the platform
Find the platform being used by the user
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{mapColors}
\alias{mapColors}
\title{Map vector of factors to color}
\usage{
mapColors(x, palette)
}
\arguments{
\item{x}{vector}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}
}
\description{
Map vector of factors to color
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-frozen-geodesic.r
\name{search_frozen_geodesic}
\alias{search_frozen_geodesic}
\title{A pseudo-derivative, line search algorithm along frozen geodesics.}
\usage{
search_frozen_geodesic(
  current,
  index,
  tries,
  max.tries = 5,
  n = 5,
  frozen,
  cur_index = NA,
  ...
)
}
\arguments{
\item{current}{starting projection}

\item{index}{index function}

\item{tries}{the counter of the outer loop of the opotimiser}

\item{max.tries}{maximum number of iteration before giving up}

\item{n}{number of random steps to take to find best direction}

\item{frozen}{a matrix of frozen variables}

\item{cur_index}{the index value of the current basis}

\item{...}{other arguments being passed into the \code{search_frozen_geodesic()}
These three functions perform a corresponding role to
\code{\link{search_geodesic}}, \code{\link{find_best_dir}} and
\code{\link{find_path_peak}} but for the frozen tour.  They work by
zero'ing out the frozen variables and travelling in that restricted
subspace.}
}
\description{
A pseudo-derivative, line search algorithm along frozen geodesics.
}
\section{To do}{
 eliminate these functions
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-trails.r
\name{display_trails}
\alias{display_trails}
\alias{animate_trails}
\title{Display tour path with trails}
\usage{
display_trails(
  center = TRUE,
  axes = "center",
  half_range = NULL,
  col = "black",
  pch = 20,
  cex = 1,
  past = 3,
  axislablong = FALSE,
  ...
)

animate_trails(data, tour_path = grand_tour(), ...)
}
\arguments{
\item{center}{if TRUE, centers projected data to (0,0).  This pins the
center of data cloud and make it easier to focus on the changing shape
rather than position.}

\item{axes}{position of the axes: center, bottomleft or off}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{col}{color to be plotted.  Defaults to "black"}

\item{pch}{shape of the point to be plotted.  Defaults to 20.}

\item{cex}{magnification of plotting text relative to default. Defaults to 1.}

\item{past}{draw line between current projection and projection \code{past}
steps ago}

\item{axislablong}{text labels only for the long axes in a projection, default FALSE}

\item{...}{other arguments passed on to \code{\link{animate}} and
\code{\link{display_xy}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a 2D tour path with a point trails
}
\examples{
animate_trails(flea[,1:6], col=flea$species)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-better.r
\name{correct_orientation}
\alias{correct_orientation}
\title{check if the current and target bases are of the same orientation}
\usage{
correct_orientation(current, target)
}
\description{
check if the current and target bases are of the same orientation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-andrews.r
\name{andrews}
\alias{andrews}
\title{Compute Andrews' curves}
\usage{
andrews(x)
}
\arguments{
\item{x}{input a new parameter}
}
\value{
a function with single argument, theta
}
\description{
This function takes a numeric vector of input, and returns a function which
allows you to compute the value of the Andrew's curve at every point along
its path from -pi to pi.
}
\examples{
a <- andrews(1:2)
a(0)
a(-pi)
grid <- seq(-pi, pi, length = 50)
a(grid)

plot(grid, andrews(1:2)(grid), type = "l")
plot(grid, andrews(runif(5))(grid), type = "l")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-stereo.r
\name{anaglyph}
\alias{anaglyph}
\title{Draw anaglyphs with base graphics.}
\usage{
anaglyph(d3, blue, red, cex = 1)
}
\arguments{
\item{d3}{3d numeric matrix giving position of points}

\item{blue}{blue colour (for right eye)}

\item{red}{red colour (for left eye)}

\item{cex}{size of the point to be plotted.  Defaults to 1.}
}
\description{
Draw anaglyphs with base graphics.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/section-pursuit.r
\name{weights_bincount_radial}
\alias{weights_bincount_radial}
\title{Computes weights for the rescaling of radial bin counts.}
\usage{
weights_bincount_radial(histo, p)
}
\arguments{
\item{histo}{input histogram (for bin definitions)}

\item{p}{number of variables}
}
\description{
Computes weights for the rescaling of radial bin counts.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-geodesic.r
\name{search_geodesic}
\alias{search_geodesic}
\title{A pseudo-derivative, line search algorithm.}
\usage{
search_geodesic(
  current,
  alpha = 1,
  index,
  tries,
  max.tries = 5,
  ...,
  n = 5,
  delta = 0.01,
  cur_index = NA
)
}
\arguments{
\item{current}{starting projection}

\item{alpha}{maximum distance to travel (currently ignored)}

\item{index}{interestingness index function}

\item{tries}{the counter of the outer loop of the opotimiser}

\item{max.tries}{maximum number of failed attempts before giving up}

\item{...}{other arguments being passed into the \code{search_geodesic()}}

\item{n}{number of random steps to take to find best direction}

\item{delta}{step size for evaluation of best direction}

\item{cur_index}{index value for starting projection, set NA if it needs to
be calculated}
}
\description{
This is a novel method for finding more interesting projections for the
guided tour.  It works by first taking a small step in \code{n} random
directions, and then picking the direction that looks most promising
(based on the height of the index function), which is effectively a gradient search.
Then it performs a linear search along the geodesic in that direction,
traveling up to half way around the sphere.
}
\details{
You should not to have call this function directly, but should supply it
to the \code{\link{guided_tour}} as a search strategy.
}
\examples{
animate_xy(flea[, 1:6], guided_tour(holes(), search_f = search_geodesic))
}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-geodesic.r
\name{find_path_peak}
\alias{find_path_peak}
\title{Find the most interesting projection along a geodesic.}
\usage{
find_path_peak(old, new, index, max_dist = pi/4, ...)
}
\arguments{
\item{old}{currention project}

\item{new}{projection that gives direction to travel in}

\item{index}{interestingness index function}

\item{max_dist}{maximum distance to travel along in radians}
}
\description{
Use \code{\link{optimize}} to find the most interesting projection amongst
all projections on a geodesic.  This method assumes that the function is
continuous with a single maximum, but seems to do ok even if there are
multiple maxima.
}
\keyword{internal}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interesting-indices.r
\name{norm_bin}
\alias{norm_bin}
\alias{norm_kol}
\title{Normality index.}
\usage{
norm_bin(nr)

norm_kol(nr)
}
\arguments{
\item{nr}{The number of rows in the target matrix}
}
\description{
Compares the similarity between the projected distribution and a normal distribution.
\itemize{
\item norm_bin: compares the count in 100 histogram bins
\item norm_kol: compares the cdf based on the Kolmogorov–Smirnov test (KS test)
}
}
\examples{
# manually compute the norm_kol index
# create the index function
set.seed(123)
index <- norm_kol(nrow(flea[, 1:3]))
# create the projection
proj <- matrix(c(1, 0, 0), nrow = 3)
# pre-process the example data
flea_s <- sphere_data(flea[, 1:3])
# produce the index value
index(flea_s \%*\% proj)
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-guided-section.r
\name{guided_section_tour}
\alias{guided_section_tour}
\title{A guided section tour path.}
\usage{
guided_section_tour(
  index_f,
  d = 2,
  alpha = 0.5,
  cooling = 0.99,
  max.tries = 25,
  max.i = Inf,
  v_rel = NULL,
  anchor = NULL,
  search_f = search_geodesic,
  ...
)
}
\arguments{
\item{index_f}{the section pursuit index function to optimise. The function
needs to take three arguments, the projected data, the vector of distances
from the current projection plane, and the slice thickness h.}

\item{d}{target dimensionality}

\item{alpha}{the initial size of the search window, in radians}

\item{cooling}{the amount the size of the search window should be adjusted
by after each step}

\item{max.tries}{the maximum number of unsuccessful attempts to find
a better projection before giving up}

\item{max.i}{the maximum index value, stop search if a larger value is found}

\item{v_rel}{relative volume of the slice. If not set, suggested value
is calculated and printed to the screen.}

\item{anchor}{A vector specifying the reference point to anchor the slice.
If NULL (default) the slice will be anchored at the data center.}

\item{search_f}{the search strategy to use}

\item{...}{arguments sent to the search_f}
}
\description{
The guided section tour is a variation of the guided tour that is
using a section pursuit index for the selection of target planes.
}
\details{
Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate_slice}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
# Generate samples on a 3d hollow sphere using the geozoo package
set.seed(12345)
sphere3 <- geozoo::sphere.hollow(3)$points
# Columns need to be named before launching the tour
colnames(sphere3) <- c("x1", "x2", "x3")
# Off-center anchoring
anchor3 <- matrix(rep(0.75, 3), ncol=3)
# Index setup
r_breaks <- linear_breaks(5, 0, 1)
a_breaks <- angular_breaks(10)
eps <- estimate_eps(nrow(sphere3), ncol(sphere3), 0.1 / 1, 5 * 10, 10, r_breaks)
idx <- slice_index(r_breaks, a_breaks, eps, bintype = "polar", power = 1, reweight = TRUE, p = 3)
# Running the guided section tour select sections showing a big hole in the center
animate_slice(sphere3, guided_section_tour(idx, v_rel = 0.1, anchor = anchor3, max.tries = 5),
  v_rel = 0.1, anchor = anchor3
)
}
\seealso{
\code{\link{slice_index}} for an example of an index functions.
\code{\link{search_geodesic}}, \code{\link{search_better}},
  \code{\link{search_better_random}} for different search strategies
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear-algebra.r
\name{normalise}
\alias{normalise}
\title{Normalise a numeric matrix.}
\usage{
normalise(x)
}
\arguments{
\item{x}{numeric matrix or vector}
}
\description{
Ensure that columns of a numeric matrix have norm 1
}
\keyword{algebra}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-little.r
\name{bases_little}
\alias{bases_little}
\title{Generate bases for the little tour}
\usage{
bases_little(p, d = 2, reorder = TRUE)
}
\arguments{
\item{p}{dimensionality of data}

\item{d}{dimensionality of target projection
bases are visited}
}
\description{
Generate bases for the little tour
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search_posse.R
\name{search_posse}
\alias{search_posse}
\title{Search for a better projection based on Poss, 1995}
\usage{
search_posse(
  current,
  alpha = 0.5,
  index,
  tries,
  max.tries = 300,
  cur_index = NA,
  ...
)
}
\arguments{
\item{current}{starting projection}

\item{alpha}{the angle used to search the target basis from the current basis}

\item{index}{index function}

\item{tries}{the counter of the outer loop of the opotimiser}

\item{max.tries}{maximum number of iteration before giving up}

\item{cur_index}{the index value of the current basis}

\item{...}{other arguments being passed into the \code{search_better()}}
}
\description{
Search for a better projection based on Poss, 1995
}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/manual_slice.r
\name{manual_slice}
\alias{manual_slice}
\title{Manually slice along a variable axis.}
\usage{
manual_slice(
  data,
  proj,
  var = 1,
  nsteps = 20,
  v_rel = 0.01,
  rescale = FALSE,
  sphere = FALSE,
  col = "black",
  half_range = NULL,
  anchor_nav = "topright",
  palette = "Zissou 1",
  ...
)
}
\arguments{
\item{data}{numeric matrix, with n rows and p columns}

\item{proj}{projection from which slices are constructed}

\item{var}{variable axis to run the center along: 1, ..., p}

\item{nsteps}{number of changes in center to make}

\item{v_rel}{relative volume of the slice. If not set, suggested value
is calculated and printed to the screen.}

\item{rescale}{Default FALSE. If TRUE, rescale all variables to range [0,1]?}

\item{sphere}{if true, sphere all variables}

\item{col}{color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{anchor_nav}{position of the anchor: center, topright or off}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}

\item{...}{other options passed to output device}
}
\description{
The manual slice tour takes the current projection, with display_slice,
and changes the slice center.
}
\examples{
# Note that you might need to use the quartz()
# on OSX to see the animation
sphere5 <- data.frame(geozoo::sphere.hollow(5)$points)
proj <- basis_random(5, 2)
manual_slice(sphere5, proj, var=3, nsteps=10, rescale=TRUE, half_range=1.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{Rat CNS}
\alias{Rat CNS}
\alias{ratcns}
\title{Rat CNS Gene Expression}
\format{
A 112 x 11 numeric array
}
\description{
Columns:
}
\details{
e11  e13  e15  e18  e21   p0   p7  p14    a class1 class2
\itemize{
  \item e11, an ebryonic timepoint from the original data with the number
     corresponding to the day
  \item e13, an ebryonic timepoint from the original data with the number
     corresponding to the day
  \item e15, an ebryonic timepoint from the original data with the number
     corresponding to the day
  \item e18, an ebryonic timepoint from the original data with the number
     corresponding to the day
  \item e21, an ebryonic timepoint from the original data with the number
    corresponding to the day
  \item p0, a postnatal timpoint from the original data with the number
    corresponding to the day
  \item p7, a postnatal timpoint from the original data with the number
    corresponding to the day
  \item p14, a postnatal timpoint from the original data with the number
    corresponding to the day
  \item a, a postnatal timpoint from the original data. It is equivalent to
    p90.
  \item class1, is the high-level class: its range is 1:4
  \item class2, breaks down the high-level classes, so its range is 1:14
}

Rows: Each case is a gene (or gene family?) And each cell is the gene
expression level for that gene at time t, averaging a few measured values
and normalizing using the maximum expression value for that gene.

Reference (available on the web at pnas.org): Large-scale temporal gene
expression mapping of central nervous system development by X. Wen, S.
Fuhrman, G. S. Michaels, D. B. Carr, S. Smith, J. L. Barker, R. Somogyi in
the Proceedings of the National Academy of Science, Vol 95, pp. 334-339,
January 1998
}
\examples{

head(ratcns)
animate_xy(ratcns[, 1:8], col = ratcns[, 10])
}
\references{
https://www.pnas.org
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{to_stop}
\alias{to_stop}
\title{Prints information on how to stop the output}
\usage{
to_stop()
}
\description{
This function prints the corresponding information on how to stop the plotting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.r
\name{render_proj}
\alias{render_proj}
\title{Render plotly animation frame}
\usage{
render_proj(
  data,
  prj,
  axis_labels = NULL,
  obs_labels = NULL,
  limits = 1,
  position = "center"
)
}
\arguments{
\item{data}{matrix, or data frame containing numeric columns,
should be standardised to have mean 0, sd 1}

\item{prj}{projection matrix}

\item{axis_labels}{of the axes to be displayed}

\item{obs_labels}{labels of the observations to be available for interactive mouseover}

\item{limits}{value setting the lower and upper limits of
projected data, default 1}

\item{position}{position of the axes: center (default),
bottomleft or off}
}
\value{
list containing projected data, circle and segments for axes
}
\description{
This function takes a projection matrix as produced by
save_history(), and draws it on the projected data
like a biplot. This will product the data objects needed
in order for the user to plot with base or ggplot2.
Note that for now this only works for 2D projections.
}
\examples{
data(flea)
flea_std <- apply(flea[,1:6], 2, function(x) (x-mean(x))/sd(x))
prj <- basis_random(ncol(flea[,1:6]), 2)
p <- render_proj(flea_std, prj)
if (require("ggplot2")) {
  ggplot() +
    geom_path(data=p$circle, aes(x=c1, y=c2)) +
    geom_segment(data=p$axes, aes(x=x1, y=y1, xend=x2, yend=y2)) +
    geom_text(data=p$axes, aes(x=x2, y=y2, label=rownames(p$axes))) +
    geom_point(data=p$data_prj, aes(x=P1, y=P2)) +
    xlim(-1,1) + ylim(-1, 1) +
    theme_bw() +
    theme(aspect.ratio=1,
       axis.text=element_blank(),
       axis.title=element_blank(),
       axis.ticks=element_blank(),
       panel.grid=element_blank())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-density.r
\name{display_dist}
\alias{display_dist}
\alias{animate_dist}
\title{1d distribution tour path animation.}
\usage{
display_dist(
  method = "density",
  center = TRUE,
  half_range = NULL,
  col = "black",
  rug = FALSE,
  palette = "Zissou 1",
  density_max = 3,
  bw = 0.2,
  scale_density = FALSE,
  ...
)

animate_dist(data, tour_path = grand_tour(1), ...)
}
\arguments{
\item{method}{display method, histogram or density plot}

\item{center}{should 1d projection be centered to have mean zero (default: TRUE).
This pins the centre of distribution to the same place, and makes it
easier to focus on the shape of the distribution.}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{col}{color to use for points, can be a vector or hexcolors or a factor.  Defaults to "black".}

\item{rug}{draw rug plot showing position of actual data points?}

\item{palette}{name of color palette for point colour, used by \code{\link{hcl.colors}}, default "Zissou 1"}

\item{density_max}{allow control of the y range for density plot}

\item{bw}{binwidth for histogram and density, between 0-1, default 0.2}

\item{scale_density}{Height of density is scaled at each projection, default FALSE}

\item{...}{other arguments passed on to \code{\link{animate}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a 1d tour path with a density plot or histogram.
}
\examples{
animate_dist(flea[, 1:6])

# Change inputs, to color by group, fix y axis, change bin width
# and scale bar height or density at each projection
animate_dist(flea[, 1:6], col=flea$species, density_max=5)
animate_dist(flea[, 1:6], col=flea$species, density_max=5, bw=0.1)
animate_dist(flea[, 1:6], col=flea$species, scale_density=TRUE)

# When the distribution is not centred, it tends to wander around in a
# distracting manner
animate_dist(flea[, 1:6], center = FALSE)

# Alternatively, you can display the distribution with a histogram
animate_dist(flea[, 1:6], method = "hist")
}
\seealso{
\code{\link{animate}} for options that apply to all animations
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpolate.r
\name{interpolate}
\alias{interpolate}
\title{Interpolate geodesically between bases.}
\usage{
interpolate(basis_set, angle = 0.05, cycle = FALSE)
}
\arguments{
\item{basis_set}{input basis set}

\item{angle}{target distance (in radians) between bases}

\item{cycle}{For \code{\link{planned_tour}} cycle through continuously (TRUE) or stop after first pass (FALSE)}
}
\description{
This function takes a set of bases and produces a tour by geodesically
interpolating between each basis
}
\examples{
t1 <- save_history(flea[, 1:6], grand_tour(1), max = 3)
dim(t1)
dim(interpolate(t1, 0.01))
dim(interpolate(t1, 0.05))
dim(interpolate(t1, 0.1))
t2 <- save_history(flea[, 1:6], grand_tour(2), max = 2)
dim(interpolate(t2, 0.05))
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{xnul}
\alias{xnul}
\alias{nul}
\title{A null function}
\usage{
nul(...)
}
\arguments{
\item{...}{all arguments to \code{...} are ignore}
}
\description{
This function does nothing, and is a useful default callback function
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-better.r
\name{basis_nearby}
\alias{basis_nearby}
\title{Generate nearby bases, e.g. for simulated annealing.}
\usage{
basis_nearby(current, alpha = 0.5, method = "linear")
}
\description{
Generate nearby bases, e.g. for simulated annealing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-idx.R
\name{display_idx}
\alias{display_idx}
\alias{animate_idx}
\title{Display a 1D linear aggregation index}
\usage{
display_idx(
  center = FALSE,
  half_range = NULL,
  abb_vars = TRUE,
  col = "red",
  cex = 3,
  panel_height_ratio = c(3, 2),
  label_x_pos = 0.7,
  label = NULL,
  label_cex = 1,
  label_col = "grey80",
  add_ref_line = TRUE,
  axis_bar_col = "#000000",
  axis_bar_lwd = 3,
  axis_label_cex_upper = 1,
  axis_label_cex_lower = 1,
  axis_bar_label_cex = 1,
  axis_bar_label_col = "#000000",
  axis_var_cex = 1,
  axis_var_col = "#000000",
  palette = "Zissou 1",
  ...
)

animate_idx(data, tour_path = grand_tour(1), ...)
}
\arguments{
\item{center}{should 1d projection be centered to have mean zero (default: TRUE).
This pins the centre of distribution to the same place, and makes it
easier to focus on the shape of the distribution.}

\item{half_range}{half range to use when calculating limits of projected.
If not set, defaults to maximum distance from origin to each row of data.}

\item{abb_vars}{logical, whether to abbreviate the variable name, if long}

\item{col}{the color used for points, can be a vector or hexcolors or a
factor, default to "red".}

\item{cex}{the size used for points, default to 0.5}

\item{panel_height_ratio}{input to the height argument in
[graphics::layout()] for the height of data and axis panel.}

\item{label_x_pos}{the x position of text label, currently labels are
positioned at a fixed x value for each observation}

\item{label}{the text label, a vector}

\item{label_cex}{the size for text labels}

\item{label_col}{the color for text labels}

\item{add_ref_line}{whether to add a horizontal reference line for each
observation, logical default to TRUE}

\item{axis_bar_col}{the color of the axis bar}

\item{axis_bar_lwd}{the width of the axis bar}

\item{axis_label_cex_upper}{the size of the axis label in the upper panel}

\item{axis_label_cex_lower}{the size of the axis label in the lower panel}

\item{axis_bar_label_cex}{the size of the axis label}

\item{axis_bar_label_col}{the color of the axis label}

\item{axis_var_cex}{the size of the variable name
to the right of the axis panel}

\item{axis_var_col}{the color of the variable name
to the right of the axis panel}

\item{palette}{name of color palette for point colour, used by
\code{\link{hcl.colors}}, default "Zissou 1"}

\item{...}{ignored}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a 1D tour path for data where individuals are ranked
by a multivariate index. Allows one to examine the sensitivity
of the ranking on the linear combination. Variables should be
scaled to be between 0-1. This is only designed to work with
a local tour, or a radial tour.
}
\examples{
data(places)
places_01 <- apply(places[1:10,1:9], 2, function(x) (x-min(x))/(max(x)-min(x)))
b <- matrix(rep(1/sqrt(9), 9), ncol=1)
places_init <- cbind(places_01, idx = as.vector(as.matrix(places_01) \%*\% b))
places_sorted <- places_init[order(places_init[,10]), 1:9]
animate_idx(places_sorted, tour_path = local_tour(b, angle=pi/8),
            label=as.character(places$stnum[1:9]),
            label_x_pos = 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{Laser measurements}
\alias{Laser measurements}
\alias{laser}
\title{Turnable laser measurements from Bellcore}
\format{
A 64 x 4 numeric array
}
\description{
This data came from an investigation of an experimental laser at Bellcore.
It was a tunable laser, in the sense that both its wavelength and power
output were controllable.
}
\details{
Rotation helped the experimental physicists to characterize the laser, which
turned out not to be a very good one, due to its unstable operating region.

This data initially came to the statistics research group when Janette
Cooper asked Paul Tukey to help her analyze the data she had collected to
describe the laser.

\itemize{
 \item ifront, current applied to the front of the laser
 \item iback, current applied to the back of the laser
 \item power, output power
 \item lambda, output wavelength
}
}
\examples{

head(laser)
animate_xy(laser[, -4])
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/anomaly-pursuit.r
\name{anomaly_index}
\alias{anomaly_index}
\title{Anomaly index.}
\usage{
anomaly_index()
}
\description{
Calculates an index that looks for the best projection of
observations that are outside a pre-determined p-D ellipse.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slicing-util.r
\name{eucl_norm_sq}
\alias{eucl_norm_sq}
\title{Calculate the squared Euclidean norm of a vector x}
\usage{
eucl_norm_sq(x)
}
\arguments{
\item{x}{numeric vector}
}
\description{
Calculate the squared Euclidean norm of a vector x
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path-index.r
\name{plot.path_index}
\alias{plot.path_index}
\title{Plot history index with ggplot2.}
\usage{
\method{plot}{path_index}(x, ...)
}
\description{
Plot history index with ggplot2.
}
\keyword{hplot}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-local.r
\name{local_tour}
\alias{local_tour}
\title{A local tour path.}
\usage{
local_tour(start, angle = pi/4)
}
\arguments{
\item{start}{initial projection matrix}

\item{angle}{distance in radians to stay within}
}
\description{
The local tour alternates between the starting position and a
nearby random projection.
}
\details{
Usually, you will not call this function directly, but will pass it to
a method that works with tour paths like \code{\link{animate}},
\code{\link{save_history}} or \code{\link{render}}.
}
\examples{
animate_xy(flea[, 1:3], local_tour(basis_init(3, 2)))
animate_xy(flea[, 1:3], local_tour(basis_init(3, 2), 0.2))
animate_xy(flea[, 1:3], local_tour(basis_random(3, 2), 0.2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geodesic.r
\name{geodesic_path}
\alias{geodesic_path}
\title{Generate geodesic path.}
\usage{
geodesic_path(current, target, frozen = NULL, ...)
}
\arguments{
\item{current}{starting projection}

\item{target}{ending projection}

\item{frozen}{matrix giving frozen variables, as described in
\code{\link{freeze}}}
}
\value{
\item{interpolate}{A function with single parameter in [0, 1] that
    returns an interpolated frame between the current and future frames.
    0 gives the current plane, 1 gives the new target frame in plane of
    current frame.}
  \item{dist}{The distance, in radians, between the current and target
    frames.}
 \item{Fa}{The current frame.}
 \item{Fz}{The new target frame.}
 \item{tau}{The principle angles between the current and target frames.}
 \item{Ga}{The current plane.}
 \item{Gz}{The target plane.}
}
\description{
Wrap basis generation method with a function that computes the
geodesic interpolation from the previous frame to the next frame, and
provides convenient access to all the information about the path.
}
\details{
Frozen variables allow us to keep certain values of the projection
fixed and generate a geodesic across the subspace generated by those
}
\examples{
a <- basis_random(4, 2)
b <- basis_random(4, 2)
path <- geodesic_path(a, b)

path$dist
all.equal(a, path$interpolate(0))
# Not true generally - a rotated into plane of b
all.equal(b, path$interpolate(1))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/history.r
\name{save_history}
\alias{save_history}
\title{Save tour history.}
\usage{
save_history(
  data,
  tour_path = grand_tour(),
  max_bases = 100,
  start = NULL,
  rescale = FALSE,
  sphere = FALSE,
  step_size = Inf,
  ...
)
}
\arguments{
\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator}

\item{max_bases}{maximum number of new bases to generate.  Some tour paths
(like the guided tour) may generate less than the maximum.}

\item{start}{starting projection, if you want to specify one}

\item{rescale}{Default FALSE. If TRUE, rescale all variables to range [0,1]?}

\item{sphere}{if true, sphere all variables}

\item{step_size}{distance between each step - defaults to \code{Inf} which
forces new basis generation at each step.}

\item{...}{additional arguments passed to tour path}
}
\description{
Save a tour path so it can later be displayed in many different ways.
}
\examples{
# You can use a saved history to replay tours with different visualisations

t1 <- save_history(flea[, 1:6], max = 3)
animate_xy(flea[, 1:6], planned_tour(t1))
## andrews_history(t1)
## andrews_history(interpolate(t1))

## t1 <- save_history(flea[, 1:6], grand_tour(4), max = 3)
## animate_pcp(flea[, 1:6], planned_tour(t1))
## animate_scatmat(flea[, 1:6], planned_tour(t1))

## t1 <- save_history(flea[, 1:6], grand_tour(1), max = 3)
## animate_dist(flea[, 1:6], planned_tour(t1))

testdata <- matrix(rnorm(100 * 3), ncol = 3)
testdata[1:50, 1] <- testdata[1:50, 1] + 10
testdata <- sphere_data(testdata)
t2 <- save_history(testdata, guided_tour(holes()),
  max = 5
)
animate_xy(testdata, planned_tour(t2))

# Or you can use saved histories to visualise the path that the tour took.
plot(path_index(interpolate(t2), holes()))

# And you can plot any individual frame using
best_prj <- matrix(t2[,,3], ncol=2)
p <- render_proj(testdata, best_prj)
# which creates a data frame with the elements
# to make the plot, see render_proj() for plotting code
# OR see draw_tour_axes() for similar code in base graphics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search-geodesic.r
\name{find_best_dir}
\alias{find_best_dir}
\title{Find the most promising direction to travel in.}
\usage{
find_best_dir(old, index, dist = 0.01, counter = 5, ...)
}
\arguments{
\item{old}{current projection}

\item{index}{interestingness index function}

\item{dist}{step size in radians, should be small}

\item{counter}{of random steps to take}
}
\description{
Starting from the current projection, pick \code{counter} random location
and take a small step towards and away from each location.  The most
promising direction has the highest value of the \code{index} function.
}
\keyword{internal}
\keyword{optimize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-stars.r
\name{display_stars}
\alias{display_stars}
\alias{animate_stars}
\title{Star glyph tour path animation.}
\usage{
display_stars(...)

animate_stars(data, tour_path = grand_tour(3), ...)
}
\arguments{
\item{...}{other arguments passed on to \code{\link{stars}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a nD tour path with star glyphs.
}
\details{
Currently, scaling doesn't seem to be computed absolutely correctly, as
centres move around as well as outside points.
}
\examples{
animate_stars(flea[1:10, 1:6])
animate_stars(flea[1:10, 1:6], grand_tour(5))
animate_stars(flea[, 1:6], grand_tour(5))
animate_stars(flea[1:10, 1:6], grand_tour(5),
  col.stars = rep("grey50", 10), radius = FALSE
)
}
\seealso{
\code{\link{animate}} for options that apply to all animations
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interesting-indices.r
\name{skewness}
\alias{skewness}
\title{Skewness index.}
\usage{
skewness()
}
\description{
Calculates the skewness index. See Cook, Buja and Cabrera (1993)
Projection pursuit indexes based on orthonormal function expansions
 for equations.
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-scatmat.r
\name{display_scatmat}
\alias{display_scatmat}
\alias{animate_scatmat}
\title{Scatterplot matrix tour path animation.}
\usage{
display_scatmat(...)

animate_scatmat(data, tour_path = grand_tour(3), ...)
}
\arguments{
\item{...}{other arguments passed on to \code{\link{animate}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Animate a nD tour path with a scatterplot matrix.
}
\details{
The lines show the observations, and the points, the values of the
projection matrix.
}
\examples{
animate_scatmat(flea[, 1:6], grand_tour(2))
animate_scatmat(flea[, 1:6], grand_tour(6))
}
\seealso{
\code{\link{animate}} for options that apply to all animations
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear-algebra.r
\name{orthonormalise}
\alias{orthonormalise}
\title{Orthonormalise using modified Gram-Schmidt process.}
\usage{
orthonormalise(x)
}
\arguments{
\item{x}{numeric matrix}
}
\description{
Orthonormalise using modified Gram-Schmidt process.
}
\keyword{algebra}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path-andrews.r
\name{path_curves}
\alias{path_curves}
\title{Draw the path that the geodesics took.}
\usage{
path_curves(history, data = attr(history, "data"))
}
\arguments{
\item{history}{list of bases produced by \code{\link{save_history}}
(or otherwise)}

\item{data}{dataset to be projected on to bases}
}
\description{
This computes the projected values of each observation at each step, and
allows you to recreate static views of the animated plots.
}
\examples{
path1d <- save_history(flea[, 1:6], grand_tour(1), 3)
path2d <- save_history(flea[, 1:6], grand_tour(2), 3)

if (require("ggplot2")) {
  plot(path_curves(path1d))
  plot(path_curves(interpolate(path1d)))

  plot(path_curves(path2d))
  plot(path_curves(interpolate(path2d)))

  # Instead of relying on the built in plot method, you might want to
  # generate your own.  Here are few examples of alternative displays:

  df <- path_curves(path2d)
  ggplot(data = df, aes(x = step, y = value, group = obs:var, colour = var)) +
    geom_line() +
    facet_wrap(~obs)

  library(tidyr)
  ggplot(
    data = pivot_wider(df,
      id_cols = c(obs, step),
      names_from = var, names_prefix = "Var",
      values_from = value
    ),
    aes(x = Var1, y = Var2)
  ) +
    geom_point() +
    facet_wrap(~step) +
    coord_equal()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interesting-indices.r
\name{cmass}
\alias{cmass}
\title{Central mass index.}
\usage{
cmass()
}
\description{
Calculates the central mass index.  See Cook and Swayne (2007)
Interactive and Dynamic Graphics for Data Analysis for equations.
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tour-frozen.r
\name{check_freezer_safe}
\alias{check_freezer_safe}
\title{Check matrix is a valid frozen matrix}
\usage{
check_freezer_safe(frozen)
}
\arguments{
\item{frozen}{matrix to check for freezability}
}
\description{
Check matrix is a valid frozen matrix
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.r
\name{render}
\alias{render}
\title{Render frames of animation to disk}
\usage{
render(
  data,
  tour_path,
  display,
  dev,
  ...,
  apf = 1/10,
  frames = 50,
  rescale = FALSE,
  sphere = FALSE,
  start = NULL
)
}
\arguments{
\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator}

\item{display}{the method used to render the projected data,
e.g. \code{\link{display_xy}}, \code{\link{display_pcp}}}

\item{dev}{name of output device to use (e.g. \code{\link{png}},
\code{\link{pdf}})}

\item{...}{other options passed to output device}

\item{apf}{angle (in radians) per frame}

\item{frames}{number of frames in output}

\item{rescale}{default FALSE. If TRUE, rescale all variables to range [0,1]}

\item{sphere}{if true, sphere all variables}

\item{start}{starting projection.  If \code{NULL}, uses path default.}
}
\description{
Render frames of animation to disk
}
\examples{
\donttest{
tmp_path <- tempdir()
render(flea[, 1:6], grand_tour(), display_xy(), "pdf",
  frames = 3,
  file.path(tmp_path, "test.pdf")
)
render(flea[, 1:6], grand_tour(), display_xy(), "png",
  frames = 3,
  file.path(tmp_path, "test-\%03d.png")
)
}
}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slicing-util.r
\name{compute_v_rel}
\alias{compute_v_rel}
\title{If not set, compute epsilon based on half_range}
\usage{
compute_v_rel(v_rel, half_range, n)
}
\description{
If not set, compute epsilon based on half_range
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.r
\name{render_gif}
\alias{render_gif}
\title{Render frames of animation to a gif file}
\usage{
render_gif(
  data,
  tour_path,
  display,
  gif_file = "animation.gif",
  ...,
  apf = 1/10,
  frames = 50,
  rescale = FALSE,
  sphere = FALSE,
  start = NULL,
  loop = TRUE
)
}
\arguments{
\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator}

\item{display}{the method used to render the projected data,
e.g. \code{\link{display_xy}}, \code{\link{display_pcp}}}

\item{gif_file}{Name of gif file (default = "animation.gif")}

\item{...}{other options passed to \code{\link{png}}}

\item{apf}{angle (in radians) per frame}

\item{frames}{number of frames in output}

\item{rescale}{default FALSE. If TRUE, rescale all variables to range [0,1]}

\item{sphere}{if true, sphere all variables}

\item{start}{starting projection.  If \code{NULL}, uses path default.}

\item{loop}{Logical for gifski to loop or not, default=TRUE}
}
\description{
Render frames of animation to a gif file
}
\examples{
\dontrun{
# gifski needs to be installed to render a gif
if (requireNamespace("gifski", quietly = TRUE)) {
  gif_file <- file.path(tempdir(), "test.gif")
  render_gif(flea[, 1:6], grand_tour(), display_xy(), gif_file)
  utils::browseURL(gif_file)
  unlink(gif_file)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/section-pursuit.r
\name{slice_binning}
\alias{slice_binning}
\title{Separately binning observations inside and outside the slice.}
\usage{
slice_binning(mat, dists, h, breaks_x, breaks_y, bintype = "square")
}
\arguments{
\item{mat}{projected data points}

\item{dists}{vector of all point distances from the projection plane.}

\item{h}{slice thickness.}

\item{breaks_x}{binning on the first variable (x or radius).}

\item{breaks_y}{binning on the second variable (y or angle).}

\item{bintype}{select polar (default) or square binning.}
}
\description{
Separately binning observations inside and outside the slice.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.r
\name{blank_plot}
\alias{blank_plot}
\title{Set up a blank plot to display data projections}
\usage{
blank_plot(...)
}
\description{
Set up a blank plot to display data projections
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear-algebra.r
\name{proj_dist}
\alias{proj_dist}
\title{Calculate the distance between two bases.}
\usage{
proj_dist(x, y)
}
\arguments{
\item{x}{projection matrix a}

\item{y}{projection matrix b}
}
\description{
Computes the Frobenius norm between two bases, in radians.  This is
equals to the Euclidean norm of the vector of principal angles between
the two subspaces.
}
\keyword{algebra}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display-stereo.r
\name{display_stereo}
\alias{display_stereo}
\alias{animate_stereo}
\title{Anaglpyh tour path animation.}
\usage{
display_stereo(blue, red, cex = 1, ...)

animate_stereo(
  data,
  tour_path = grand_tour(3),
  blue = rgb(0, 0.91, 0.89),
  red = rgb(0.98, 0.052, 0),
  ...
)
}
\arguments{
\item{blue}{blue colour (for right eye)}

\item{red}{red colour (for left eye)}

\item{cex}{size of the point to be plotted.  Defaults to 1.}

\item{...}{other arguments passed on to \code{\link{animate}}}

\item{data}{matrix, or data frame containing numeric columns}

\item{tour_path}{tour path generator, defaults to 2d grand tour}
}
\description{
Uses red-blue anaglyphs to display a 3d tour path.  You'll need some red-
blue glasses to get much out of this displays!
}
\examples{
animate_stereo(flea[, 1:6])
}
\keyword{hplot}
